//==================================================
// FILE:	DigiDocLib.c
// PROJECT: Digi Doc
// DESCRIPTION: Digi Doc functions for creating
//	and reading signed documents. 
// AUTHOR:  Veiko Sinivee, S|E|B IT Partner Estonia
//==================================================
// Copyright (C) AS Sertifitseerimiskeskus
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
// Lesser General Public License for more details.
// GNU Lesser General Public Licence is available at
// http://www.gnu.org/copyleft/lesser.html
//==========< HISTORY >=============================
//    20.09.2004       Changed createOCSPRequest() to use custom functions for adding
//                     OCSP nonce to OCSP request for digidoc formats 1.0, 1.1 and 1.2
//                     thereby guaranteing that in case of those formats only 20 bytes are sent.
//                     This is necessary to guarantee backward compatibility with OpenSSL 0.9.6
//    22.08.2004       Added DataFile content caching, function ddocGetDataFileCachedData()
//                     and modified functions readSignedDoc(), extractDataFile(), 
//                     ddocExtractDataFile(), ddocGetDataFile() and ddocCopyDataFile() to use it.
//                     Added function ddocSaxReadSignedDocFromMemory()
//                     New debugging interface ddocDebug()
//    12.08.2004       Moved all SAX xml parsing functions to DigiDocSAXParser.c
//                     and fixed some memory problems
//    03.08.2004                Added check to initializeNotaryInfoWithOCSP() to handle 22 byte ASN.1
//                              encoded nonce with OpenSSL 0.9.7d
//                              added a check on output filename for extractDataFile() function.
//                              added a check for verifyNotaryInfoCERT() function to check empty certificate array size.
//                              Added back the function verifyCertificateByOCSP()
//    14.06.2004		Changed generateDataFileXML() to replaces '&' with '&amp;'
//						int file names. Changed handleStartDataFile() acordingly.
//    20.05.2004		Added new error codes to handle OCSP response
//						status and fixed initializeNotaryInfoWithOCSP()
//    19.05.2004		Added a nullpointer test to checkFileExists()
//    12.05.2004		Fixed convertStringToTimeT() by changing
//						timestamp format pattern according to 1.3 format version
//    11.05.2004		Added error code ERR_OCSP_WRONG_URL to sendOCSPRequest
//						to indicate that the given server is not an OCSP server
//    23.03.2004        fixed setSignatureValueFromFile() that was affected
//                      by siganture caching. Fixed getConfirmation() to
//                      allowe sending not signed OCSP requests.
//    17.03.2004        fixed getSignerCode() to handle certs that
//                      have no comma in CN. Fixed NotaryInfo_new to
//                      reset szOrigContent of signature. Added function
//                      handleOCSPCertStatus() to check cert status 
//                      and modified initializeNotaryInfoWithOCSP() and
//                      checkNonceAndCertbyOCSP()
//    11.03.2004	added SignedDoc namespace in format 1.3
//					generateDataFileXML(), createSignedXMLDoc()
//					handleStartDataFile(), 
//					bugs in emptying buffers
//					unicode2ascii(), getSignerLastName()
//					Veiko Sinivee
//	  25.11.2003	added functions calculateFileLength(),
//				verifyCertificateByOCSP() and helper functions
//	  21.11.2003	Fixed SignatureValue decoding problem
//				Veiko Sinivee
//      17.11.2003      Aare Amenberg
//                      Changed error texts
//                      Changed sendOCSPRequest
//                      (added more error checkings) 
//      14.11.2003      Aare Amenberg
//                      Removed WriteOCSPResponse
//                      Changed addNotaryInfoXML by Veiko
//      11.11.2003      Aare Amenberg
//                      Added CorrectCharacters
//                      Added error 96 ERR_WRONG_URL_OR_PROXY
//      31.10.2003      Aare Amenberg
//                      Added error text 95
//                      Removed  if(single->singleExtensions) 
//                               SET_LAST_ERROR_RETURN_CODE(ERR_OCSP_NO_SINGLE_EXT);
//                      Function initializeNotaryInfoWithOCSP
//      29.10.2003      Aare Amenberg
//                      WIN32_CSP functions moved 
//                      to module DigiDocCsp
// 1.0	09.04.2002	Veiko Sinivee	
//  Supports XML format (Type: SK-XML-1.0)
//==================================================

#include <DigiDocDefs.h>

#ifndef WIN32
  #if HAVE_CONFIG_H
    #include <config.h>
  #endif
#endif // no win32

#define WIN32_PKCS 1

#ifdef WIN32
  #include <windows.h>
  #include <wincrypt.h>
#endif

#include <DigiDocLib.h>
#include <DigiDocSAXParser.h>
#include <DigiDocDebug.h>
#include <DigiDocConvert.h>
#include <DigiDocConfig.h>
#include <DigiDocCert.h>

#include <stdio.h>
#include <stdlib.h>
#include <memory.h>
#include <string.h>
#include <assert.h>
#include <ctype.h>

#ifdef WIN32
  //#include <unistd.h>
  //#include <netinet/in.h>
  //for _msize function
  #define FILESEPARATOR	"\\"
  #include <malloc.h>
  #include <direct.h>  
#else
  #define FILESEPARATOR	"/"
  #define DIGI_DOC_LIB
  #include <unistd.h>
  #define _mkdir mkdir
  #define _rmdir rmdir
  #define _unlink unlink 
  #define _tzset tzset
  #define _getcwd getcwd
  #ifdef __FreeBSD__
    #define _timezone tzone
    long int tzone = -7200;	/* default for Estonia, but see initDigiDocLib() */
    #define _daylight daylight
    int daylight = 0;		/* default, but see initDigiDocLib() */
  #else
    #define _timezone timezone
    #define _daylight daylight
  #endif
#endif

#include <fcntl.h>
#include <time.h>


#include <openssl/sha.h>
#include <openssl/rsa.h>
#include <openssl/evp.h>
#include <openssl/objects.h>
#include <openssl/x509.h>
#include <openssl/x509v3.h>
#include <openssl/err.h>
#include <openssl/pem.h>
#include <openssl/ssl.h>
#include <openssl/ocsp.h>
#include <openssl/pkcs12.h>
#include <openssl/rand.h>

#include <libxml/globals.h>
#include <libxml/xmlerror.h>
#include <libxml/parser.h>
#include <libxml/parserInternals.h> /* only for xmlNewInputFromFile() */
#include <libxml/tree.h>
#include <libxml/debugXML.h>
#include <libxml/xmlmemory.h>
#include <libxml/c14n.h>

#define FILE_BUFSIZE	1024*16

//==========< global constants >====================

char g_szLibName[] = "CDigiDoc";
char g_szLibVer[] = "2.2.11";
//char g_szLibVer[] = VERSION;
char* g_szSupportedVersions[] = { "1.0", "1.1", "1.2", "1.3" };

char* g_arrFormats[] = { SK_XML_1_NAME, 0 };

FormatAndVer g_supportedFormatsAndVersions[] = {
	{ SK_XML_1_NAME, SK_XML_1_VER },
	{ DIGIDOC_XML_1_1_NAME, DIGIDOC_XML_1_1_VER },
	{ DIGIDOC_XML_1_1_NAME, DIGIDOC_XML_1_2_VER },
	{ DIGIDOC_XML_1_1_NAME, DIGIDOC_XML_1_3_VER },
	{ NULL, NULL } // list end marker
};

// static buffer for name & version of the program using this library
#define GUI_VERSION_LEN   100
char g_szGUIVersion[GUI_VERSION_LEN];

//==========< global variables >====================



// forward deklaratsioon
int notary2ocspresp(const SignedDoc* pSigDoc, const NotaryInfo* pNotInfo, X509* notCert, OCSP_RESPONSE** pResp);
int notary2ocspBasResp(const SignedDoc* pSigDoc, const NotaryInfo* pNotInfo, X509* notCert, OCSP_BASICRESP** pBasResp);
int calculateOcspBasicResponseDigest(OCSP_BASICRESP* pBsResp, byte* digBuf, int* digLen);
int generateDataFileXML(SignedDoc* pSigDoc, DataFile* pDataFile, 
						const char* szDataFile, BIO* bOutFile);
char* canonicalizeXML(char* source, int len);
void swapBytes(byte* src, int len);
int getTempBIO(BIO** pBio);
int createOCSPRequest(SignedDoc* pSigDoc, OCSP_REQUEST **req, 
		      X509 *cert, X509 *pCA, byte* nonce, int nlen);
EXP_OPTION int signOCSPRequestPKCS12(OCSP_REQUEST *req, const char* filename, const char* passwd);

int verifyOCSPResponse(OCSP_RESPONSE* pResp, 
				    const X509** caCerts, const char *CApath, 
				    const X509* notCert);
int checkNonceAndCertbyOCSP(OCSP_RESPONSE* resp, X509* cert, byte* nonce1, int nonceLen);

int verifyResp(OCSP_BASICRESP* bs, const char *CA2file, const char *CA1file,
			   const char *CApath, const char* notCertFile);

extern int writeCertToXMLFile(BIO* bout, X509* cert);

//==========< utility functions >====================


//--------------------------------------------------
// Converts a filename according to platform rules
// dest - destination buffer
// destlen - destination buffer length
// src - source filename
// returns error code or ERR_OK
//--------------------------------------------------
int ddocConvertFileName(char* dest, int destlen, const char* src)
{
  int err = ERR_OK;
  
  RETURN_IF_NULL_PARAM(dest);
  RETURN_IF_NULL_PARAM(src);

  *dest = 0; // initialize
#ifdef WIN32
  err = utf82oem(src, dest, destlen);
#else
  SET_LAST_ERROR_IF_NOT(strlen(src) < destlen, ERR_BUF_LEN);
  strcpy(dest, src);
#endif
  return err;
}

#ifdef WIN32
  CRITICAL_SECTION cs_ddocLocaltime;
#endif

//--------------------------------------------------
// Converts timestamp (time_t) value to a struct
// tm value. Caller must provide address of tm struct.
// This function is used because loacltime() is not
// thread-safe and win32 has no equvalent of localtime_r().
// pTime - time_t value address
// pTmStruct - struct tm address
// bLocal - 1=localtime_r, 0=gmtime_r
// returns error code or ERR_OK
//--------------------------------------------------
EXP_OPTION int ddocLocalTime(time_t* pTime, struct tm* pTmStruct, int bLocal)
{
#ifdef WIN32
  struct tm *pTm;
#endif
  RETURN_IF_NULL_PARAM(pTime);
  RETURN_IF_NULL_PARAM(pTmStruct);

#ifdef WIN32
  EnterCriticalSection(&cs_ddocLocaltime);
  if(bLocal)
    pTm = localtime(pTime);
  else
    pTm = gmtime(pTime);
  *pTmStruct = *pTm;
  LeaveCriticalSection(&cs_ddocLocaltime);
#else
  if(bLocal)
    localtime_r(pTime, pTmStruct);
  else
    gmtime_r(pTime, pTmStruct);
#endif
  return ERR_OK;
}


//==========< admin functions >====================

//--------------------------------------------------
// returns the library name
//--------------------------------------------------
EXP_OPTION const char* getLibName()
{
	return g_szLibName;
}

//--------------------------------------------------
// returns the library version
//--------------------------------------------------
EXP_OPTION const char* getLibVersion()
{
	return g_szLibVer;
}

//--------------------------------------------------
// returns an array of supported formats terminated by NULL
//--------------------------------------------------
EXP_OPTION const char** getSupportedFormats()
{
	return (const char**)g_arrFormats;
}

//--------------------------------------------------
// returns an array of supported formats  and versions terminated by NULL
//--------------------------------------------------
EXP_OPTION FormatAndVer* getSupportedFormatsAndVersions()
{
	return g_supportedFormatsAndVersions;
}

//--------------------------------------------------
// Cheks a combination of format and version for validity
// format - digidoc format name
// version - version name
// returns error code or ERR_OK
//--------------------------------------------------
int ddocCheckFormatAndVer(const char* format, const char* version)
{
  int i;

  for(i = 0; i < SUPPORTED_VERSION_COUNT; i++)
    if(!strcmp(g_supportedFormatsAndVersions[i].szFormat, format) &&
       !strcmp(g_supportedFormatsAndVersions[i].szVersion, version))
      return ERR_OK;
  return ERR_UNSUPPORTED_FORMAT;
}

//--------------------------------------------------
// initializes DigiDoc library
//--------------------------------------------------
EXP_OPTION void initDigiDocLib() 
{
  memset(g_szGUIVersion, 0, sizeof(g_szGUIVersion));
#ifdef __FreeBSD__
  struct tm local_tm;
  time_t t;
  time(&t);
  ddocLocalTime(&t, &local_tm, 1);	// Extract timezone info from struct tm
  _timezone = - local_tm.tm_gmtoff;
  _daylight = local_tm.tm_isdst;
#endif
  ERR_load_ERR_strings();
  ERR_load_crypto_strings();
  ERR_clear_error();
  OpenSSL_add_all_algorithms();
#ifdef WIN32 
  InitializeCriticalSection(&cs_ddocErrors);
  InitializeCriticalSection(&cs_ddocLocaltime);
#endif
}

//--------------------------------------------------
// returns the GUI version
//--------------------------------------------------
EXP_OPTION const char* getGUIVersion()
{
	return g_szGUIVersion;
}

//--------------------------------------------------
// sets the GUI version
//--------------------------------------------------
EXP_OPTION void setGUIVersion(const char* szVer)
{
  strncpy(g_szGUIVersion, szVer, sizeof(g_szGUIVersion)-1);
}

//--------------------------------------------------
// cleanup of DigiDoc library
//--------------------------------------------------
EXP_OPTION void finalizeDigiDocLib()
{
	ERR_free_strings();
	EVP_cleanup();
}

//==========< general crypto fucntions >============


//--------------------------------------------------
// Checks and prints errors
//--------------------------------------------------
EXP_OPTION long checkErrors() 
{
	char buf[200], *p;
	long e=0;

	while((e = ERR_peek_error()) > ERR_LIB_NONE) {
		e = ERR_get_error();
		p = ERR_error_string( e,buf);
		// VS: keep silent !!!
		printf("ERROR: %ld - %s \n", e, p);
	}
	ERR_clear_error();
	return e;
}




//--------------------------------------------------
// Calculates files SHA1 digest
// szFileName - file name
// nDigestType - digest type. Supports only SHA1 (0)
// pDigestBuf - buffer to store the digest
// nDigestLen - buffer size, must be at least 20
//				will be updated by actual digest length
// lFileLen - pointer to a buffer where to store the file length
// returns error code or ERR_OK for success
//--------------------------------------------------
EXP_OPTION int calculateFileDigest(const char* szFileName, int nDigestType,
				   byte* pDigestBuf, int* nDigestLen, long* lFileLen)
{
  int err = ERR_OK;
  SHA_CTX ctx;
  byte buf[FILE_BUFSIZE];
  int i;
  FILE *f = NULL;

  RETURN_IF_NULL_PARAM(szFileName);
  RETURN_IF_NULL_PARAM(pDigestBuf);
  RETURN_IF_NULL_PARAM(nDigestLen);
  RETURN_IF_NULL_PARAM(lFileLen);

  memset(pDigestBuf, 0, *nDigestLen);
  if(nDigestType == DIGEST_SHA1) {
    if(*nDigestLen >= SHA_DIGEST_LENGTH) {
      *nDigestLen = SHA_DIGEST_LENGTH;
      if((f = fopen(szFileName,"rb")) != NULL) {
	//byte *data,*temp_data;
	/**/
	SHA1_Init(&ctx);
	*lFileLen = 0;
	for (;;) {
	  i = fread(buf, sizeof(char), FILE_BUFSIZE, f);
	  if (i <= 0) break;
	  *lFileLen += i;
	  SHA1_Update(&ctx, buf, (unsigned long)i);
	}
	SHA1_Final(pDigestBuf,&ctx);
	fclose(f);
      } // if - fopen
      else
	err = ERR_FILE_READ;
    } 
    else
      err = ERR_DIGEST_LEN;
  }
  else
    err = ERR_UNSUPPORTED_DIGEST;
  if (err != ERR_OK) SET_LAST_ERROR(err);
  return err;
}


//--------------------------------------------------
// Calculates file size
// szFileName - file name
// lFileLen - pointer to a buffer where to store the file length
// returns error code or ERR_OK for success
//--------------------------------------------------
EXP_OPTION int calculateFileSize(const char* szFileName, long* lFileLen)
{
	FILE* hFile = 0;

	RETURN_IF_NULL_PARAM(szFileName);
	RETURN_IF_NULL_PARAM(lFileLen);
	if((hFile = fopen(szFileName,"rb")) != NULL) {
		fseek(hFile, 0, SEEK_END);
		*lFileLen = ftell(hFile);
		fclose(hFile);
	} // if - fopen
	else
		SET_LAST_ERROR_RETURN_CODE(ERR_FILE_READ);
	return ERR_OK;
}


//--------------------------------------------------
// Calculates files SHA1-RSA signature
// szFileName - file name
// nDigestType - digest type. Supports only SHA1 (0)
// pSigBuf - buffer to store the signature
// nSigLen - buffer size, must be at least 128
//			will be updated by actual signature length
// keyfile - name of the private key file
// passwd - private key password
// returns error code or ERR_OK for success
//--------------------------------------------------
EXP_OPTION int calculateFileSignature(const char* szFileName, int nDigestType,
							byte* pSigBuf, int* nSigLen,
							const char *keyfile, const char* passwd)
{
  int err = ERR_OK;
  EVP_MD_CTX  ctx;
  byte buf[FILE_BUFSIZE];
  int i;
  FILE *f = NULL;
  EVP_PKEY* pkey = NULL;

  RETURN_IF_NULL_PARAM(szFileName);
  RETURN_IF_NULL_PARAM(pSigBuf);
  RETURN_IF_NULL_PARAM(nSigLen);
  RETURN_IF_NULL_PARAM(keyfile);
  RETURN_IF_NULL_PARAM(passwd);

  memset(pSigBuf, 0, *nSigLen);
  if(nDigestType == DIGEST_SHA1) {
    if(*nSigLen >= SIGNATURE_LEN) {
      if((err = ReadPrivateKey(&pkey, keyfile, passwd, FILE_FORMAT_PEM)) == ERR_OK) {
	if((f = fopen(szFileName,"rb")) != NULL) {					
	  EVP_SignInit(&ctx, EVP_sha1());
	  for (;;) {
	    i = fread(buf, sizeof(char), FILE_BUFSIZE, f);
	    if (i <= 0) break;
	    EVP_SignUpdate (&ctx, buf, (unsigned long)i);
	  }
	  err = EVP_SignFinal(&ctx, pSigBuf, (unsigned int*)nSigLen, pkey);
	  if(err == ERR_LIB_NONE)
	    err = ERR_OK;
	  fclose(f);
	  EVP_PKEY_free(pkey);					
	} // if - fopen
	else
	  err = ERR_FILE_READ;
      }
      else
	err = ERR_PRIVKEY_READ;
    } 
    else
      err = ERR_SIGNATURE_LEN;
  }
  else
    err = ERR_UNSUPPORTED_DIGEST;
  if (err != ERR_OK) SET_LAST_ERROR(err);
  return err;
}


//--------------------------------------------------
// Calculates input datas SHA1-RSA signature
// data - input data
// dlen - input data length
// nDigestType - digest type. Supports only SHA1 (0)
// pSigBuf - buffer to store the signature
// nSigLen - buffer size, must be at least 128
//			will be updated by actual signature length
// keyfile - name of the private key file
// passwd - private key password
// returns error code or ERR_OK for success
//--------------------------------------------------
EXP_OPTION int signData(const byte* data, int dlen, byte* pSigBuf, int* nSigLen,
			 int nDigestType, const char *keyfile, const char* passwd)
{
	int err = ERR_OK;
	EVP_MD_CTX  ctx;
	EVP_PKEY* pkey;

	RETURN_IF_NULL_PARAM(data);
	RETURN_IF_NULL_PARAM(pSigBuf);
	RETURN_IF_NULL_PARAM(nSigLen);
	RETURN_IF_NULL_PARAM(keyfile);
	RETURN_IF_NULL_PARAM(passwd);

	memset(pSigBuf, 0, *nSigLen);
	if(nDigestType == DIGEST_SHA1) {
		if(*nSigLen >= SIGNATURE_LEN) {
			if((err = ReadPrivateKey(&pkey, keyfile, passwd, FILE_FORMAT_PEM)) == ERR_OK) {
				EVP_SignInit(&ctx, EVP_sha1());
				EVP_SignUpdate (&ctx, data, (unsigned long)dlen);
				err = EVP_SignFinal(&ctx, pSigBuf, (unsigned int*)nSigLen, pkey);
				if(err == ERR_LIB_NONE)
					err = ERR_OK;
				EVP_PKEY_free(pkey);				
			}
			else
				err = ERR_PRIVKEY_READ;
		} 
		else
			err = ERR_SIGNATURE_LEN;
	}
	else
		err = ERR_UNSUPPORTED_DIGEST;

	if (err != ERR_OK) SET_LAST_ERROR(err);
	return err;
}

//--------------------------------------------------
// Calculates files SHA1 digest
// szFileName - file name
// nDigestType - digest type. Supports only SHA1 (0)
// pDigestBuf - buffer to store the digest
// nDigestLen - buffer size, must be at least 20
//				will be updated by actual digest length
// returns error code or ERR_OK for success
//--------------------------------------------------
EXP_OPTION int calculateDigest(const byte* data, int nDataLen, int nDigestType,
					byte* pDigestBuf, int* nDigestLen)
{
  int err = ERR_OK;

  RETURN_IF_NULL_PARAM(data);
  RETURN_IF_NULL_PARAM(pDigestBuf);
  RETURN_IF_NULL_PARAM(nDigestLen);
  memset(pDigestBuf, 0, *nDigestLen);
  if(nDigestType == DIGEST_SHA1) {
    if(*nDigestLen >= SHA_DIGEST_LENGTH) {
      *nDigestLen = SHA_DIGEST_LENGTH;
      SHA1(data, nDataLen, pDigestBuf);
    } 
    else
      err = ERR_DIGEST_LEN;
  }
  else
    err = ERR_UNSUPPORTED_DIGEST;
  if (err != ERR_OK) SET_LAST_ERROR(err);
  return err;
}


//--------------------------------------------------
// Verifies files SHA1-RSA signature
// szFileName - file name
// nDigestType - digest type. Supports only SHA1 (0)
// pSigBuf - buffer to store the signature
// nSigLen - buffer size, must be at least 128
//			will be updated by actual signature length
// certfile - name of the certificate file
// returns error code or ERR_OK for success
//--------------------------------------------------
EXP_OPTION int verifyFileSignature(const char* szFileName, int nDigestType,
						byte* pSigBuf, int nSigLen,
						const char *certfile)
{
  int err = ERR_OK;
  EVP_MD_CTX  ctx;
  unsigned char buf[FILE_BUFSIZE];
  int i;
  FILE *f;
  EVP_PKEY* pkey = NULL;

  RETURN_IF_NULL_PARAM(szFileName);
  RETURN_IF_NULL_PARAM(pSigBuf);
  RETURN_IF_NULL_PARAM(certfile);
  
  if(nDigestType == DIGEST_SHA1) {
    if((err = ReadPublicKey(&pkey, certfile)) == ERR_OK) {
      if((f = fopen(szFileName,"rb")) != NULL) {
	EVP_VerifyInit(&ctx, EVP_sha1());
	for (;;) {
	  i = fread(buf, sizeof(char), FILE_BUFSIZE, f);
	  if (i <= 0) break;
	  EVP_VerifyUpdate (&ctx, buf, (unsigned long)i);
	}
	err = EVP_VerifyFinal(&ctx, pSigBuf, nSigLen, pkey);
	if(err == ERR_LIB_NONE)
	  err = ERR_OK;
	fclose(f);
	EVP_PKEY_free(pkey);								
      } // if - fopen
      else
	err = ERR_FILE_READ;
    }
    else
      err = ERR_CERT_READ;
  }
  else
    err = ERR_UNSUPPORTED_DIGEST;
  
  if (err != ERR_OK) SET_LAST_ERROR(err);
  return err;
}


//--------------------------------------------------
// Verifies files SHA1-RSA signature
// szData - input data
// dataLen - input data length
// nDigestType - digest type
// pSigBuf - buffer to store the signature
// nSigLen - buffer size, must be at least 128
//			will be updated by actual signature length
// cert - certificate data
// returns error code or ERR_OK for success
//--------------------------------------------------
EXP_OPTION int verifySignature(const char* szData, unsigned long dataLen, int nDigestType,
					byte* pSigBuf, int nSigLen, X509* cert)
{
	int err = ERR_OK;
	EVP_MD_CTX  ctx;
	EVP_PKEY* pkey = NULL;

	RETURN_IF_NULL_PARAM(szData);
	RETURN_IF_NULL_PARAM(pSigBuf);
	RETURN_IF_NULL_PARAM(cert);

	if(nDigestType == DIGEST_SHA1) {
		if((err = GetPublicKey(&pkey, cert)) == ERR_OK) {
			checkErrors();
			EVP_VerifyInit(&ctx, EVP_sha1());
			checkErrors();
			EVP_VerifyUpdate (&ctx, szData, dataLen);
			checkErrors();
			err = EVP_VerifyFinal(&ctx, pSigBuf, nSigLen, pkey);
			if(err == ERR_LIB_NONE)
				err = ERR_OK;
			checkErrors();
			EVP_PKEY_free(pkey);
			checkErrors();
		}
		else
			err = ERR_CERT_READ;
	}
	else
		err = ERR_UNSUPPORTED_DIGEST;	

	if (err != ERR_OK) SET_LAST_ERROR(err);
	return err;
}

//--------------------------------------------------
// Verifies files SHA1-RSA signature (EstID specific!!!)
// digest - digest data
// dataLen - digest data length
// nDigestType - digest type
// pSigBuf - buffer to store the signature
// nSigLen - buffer size, must be at least 128
//			will be updated by actual signature length
// cert - certificate data
// returns error code or ERR_OK for success
//--------------------------------------------------
EXP_OPTION int verifyEstIDSignature(const byte* digest, int digestLen, int nDigestType,
					byte* pSigBuf, int nSigLen, X509* cert)
{
  int err = ERR_OK;
  EVP_PKEY* pkey = 0;
  byte buf2[DIGEST_LEN+2], buf3[200], buf4[50], buf5[50];
  int l2 = 0, l1;

  RETURN_IF_NULL_PARAM(digest);
  RETURN_IF_NULL_PARAM(pSigBuf);
  RETURN_IF_NULL_PARAM(cert);
  if(nDigestType == DIGEST_SHA1) {
    if((err = GetPublicKey(&pkey, cert)) == ERR_OK) {
      l2 = sizeof(buf3);
      memset(buf3, 0, sizeof(buf3));
      ERR_clear_error();
      //swapBytes(pSigBuf, nSigLen);
      l2 = RSA_public_decrypt(nSigLen, pSigBuf, buf3, pkey->pkey.rsa, RSA_PKCS1_PADDING); //RSA_PKCS1_PADDING); //RSA_NO_PADDING);
      ddocDebug(3, "verifyEstIDSignature", "decryted sig-hash len: %d", l2);

      memset(buf2, 0, DIGEST_LEN);
      if(l2 > DIGEST_LEN)
	memcpy(buf2, buf3 + l2 - DIGEST_LEN, DIGEST_LEN);
      else
	memcpy(buf2, buf3, DIGEST_LEN);
      checkErrors();
      err = compareByteArrays(digest, digestLen, buf2, DIGEST_LEN);
      //debug
      l1 = sizeof(buf4);
      encode((const byte*)digest, digestLen, (byte*)buf4, &l1);
      l1 = sizeof(buf5);
      encode((const byte*)buf2, DIGEST_LEN, (byte*)buf5, &l1);
      ddocDebug(3, "verifyEstIDSignature", "comp-hash: %s sig-hash: %s, err: %d", buf4, buf5, err);

      EVP_PKEY_free(pkey);
      checkErrors();
    }
    else
      err = ERR_CERT_READ;
  }
  else
    err = ERR_UNSUPPORTED_DIGEST;	

  if (err != ERR_OK) SET_LAST_ERROR(err);
  return err;
}


//==========< Base64 functions >================


//============================================================
// Encode a byte array in Base64 format.
// raw - input data
// rawlen - length of input data
// buf - destination buffer
// buflen - destination buffer length (will be modified by catual number of bytes)
//============================================================
EXP_OPTION void encode(const byte* raw, int rawlen, byte* buf, int* buflen) 
{
	EVP_ENCODE_CTX ectx;

	RETURN_VOID_IF_NULL(raw);
	RETURN_VOID_IF_NULL(buf);
	RETURN_VOID_IF_NULL(buflen);
	
	memset(buf, 0, *buflen);
	EVP_EncodeInit(&ectx);
	EVP_EncodeUpdate(&ectx, buf, buflen, (byte*)raw, rawlen);
	EVP_EncodeFinal(&ectx, (unsigned char*)strchr((const char*)buf, 0), buflen); 
	*buflen = strlen((const char*)buf);
	while(buf[*buflen-1] == '\n' || buf[*buflen-1] == '\r')
		buf[*buflen-1] = 0;
	*buflen = strlen((const char*)buf);
}


byte* breakToLinesOf64(byte* raw, int rawlen)
{
	int i, l1, j;
	byte* p = raw;
	
	// if it's not divide in lines
	l1 = rawlen;
	j = l1 + (l1/64) * 3;
	p = (byte*)malloc(j);
	if (!p) return NULL;
	memset(p, 0, j);
	for(i = j = 0; i < l1; i += 64) {
		strncpy(strchr((const char*)p,0), (const char*)raw+i, ((i+64<l1) ? 64 : l1-i));
		if(i + 64 < l1)
			strcat((char*)p, "\n");
	}
	l1 = strlen((const char*)p);
	return p;
}

//============================================================
// Decodes input data in Base64 format.
// raw - input data
// rawlen - length of input data
// buf - destination buffer
// buflen - destination buffer length (will be modified by catual number of bytes)
//============================================================
EXP_OPTION void decode(const byte* raw, int rawlen, byte* buf, int* buflen) 
{
	EVP_ENCODE_CTX ectx;
	int l1 = 0;
	byte* p;

	RETURN_VOID_IF_NULL(raw);
	RETURN_VOID_IF_NULL(buf);
	RETURN_VOID_IF_NULL(buflen);

	memset(buf, 0, *buflen);
	*buflen = 0;
	EVP_DecodeInit(&ectx);
	if((!strstr((const char*)raw, "\n") ||
            !strstr((const char*)raw, "\r")) && 
            strlen((const char*)raw) > 64) {
		p = breakToLinesOf64((byte*)raw, rawlen);
		l1 = strlen((const char*)p);
		EVP_DecodeUpdate(&ectx, (unsigned char*)buf, &l1, 
			(unsigned char*)p, 
			strlen((const char*)p));
		*buflen += l1;
		free(p);
	}
	else
		EVP_DecodeUpdate(&ectx, buf, buflen, (byte*)raw, rawlen);
	EVP_DecodeFinal(&ectx, buf, &l1);
	*buflen += l1;
}




//==========< conversion fucntions >================

//--------------------------------------------------
// Decodes an ASN1 generalized time
// tm - ASN1 generalized time
// y - year
// M - month
// d - day of month
// h - hour
// m - minute
// s - second
// returns error code or ERR_OK
//--------------------------------------------------
int decodeGeneralizedTime(ASN1_GENERALIZEDTIME *tm, 
			   int* y, int* M, int* d, 
			   int* h, int* m, int* s)
{
	char *v = NULL;
	int gmt=0;
	int i, err = ERR_OK;

	RETURN_IF_NULL_PARAM(tm);
	RETURN_IF_NULL_PARAM(y);
	RETURN_IF_NULL_PARAM(M);
	RETURN_IF_NULL_PARAM(d);
	RETURN_IF_NULL_PARAM(h);
	RETURN_IF_NULL_PARAM(m);
	RETURN_IF_NULL_PARAM(s);

	i=tm->length;
	v=(char *)tm->data;

	if (i < 12) 
		err = ERR_TIMESTAMP_DECODE;
	if (v[i-1] == 'Z') gmt=1;
	for (i=0; i<12; i++)
		if ((v[i] > '9') || (v[i] < '0')) 
			err = ERR_TIMESTAMP_DECODE;
	*y= (v[0]-'0')*1000+(v[1]-'0')*100 + (v[2]-'0')*10+(v[3]-'0');
	*M= (v[4]-'0')*10+(v[5]-'0');
	if ((*M > 12) || (*M < 1)) 
		err = ERR_TIMESTAMP_DECODE;
	*d= (v[6]-'0')*10+(v[7]-'0');
	*h= (v[8]-'0')*10+(v[9]-'0');
	*m=  (v[10]-'0')*10+(v[11]-'0');
	if (	(v[12] >= '0') && (v[12] <= '9') &&
		(v[13] >= '0') && (v[13] <= '9'))
		*s=  (v[12]-'0')*10+(v[13]-'0');

	if (err != ERR_OK) SET_LAST_ERROR(err);
	return err;
}

//--------------------------------------------------
// Decodes an ASN1 UTC time
// tm - ASN1 generalized time
// y - year
// M - month
// d - day of month
// h - hour
// m - minute
// s - second
// returns error code or ERR_OK
//--------------------------------------------------
int decodeUTCTime(ASN1_UTCTIME *tm, 
		  int* y, int* M, int* d, 
		  int* h, int* m, int* s)
{
	char *v;
	int gmt=0;
	int i, err = 0;

	RETURN_IF_NULL_PARAM(tm);
	RETURN_IF_NULL_PARAM(y);
	RETURN_IF_NULL_PARAM(M);
	RETURN_IF_NULL_PARAM(d);
	RETURN_IF_NULL_PARAM(h);
	RETURN_IF_NULL_PARAM(m);
	RETURN_IF_NULL_PARAM(s);

	i=tm->length;
        v=(char *)tm->data;
	
	if (i < 10) 
		err = ERR_TIMESTAMP_DECODE;
	if (v[i-1] == 'Z') gmt=1;
	for (i=0; i<10; i++)
		if ((v[i] > '9') || (v[i] < '0')) 
			err = ERR_TIMESTAMP_DECODE;
	*y= (v[0]-'0')*10+(v[1]-'0');
	if (*y < 50) *y+=100;
	*M= (v[2]-'0')*10+(v[3]-'0');
	if ((*M > 12) || (*M < 1)) 
		err = ERR_TIMESTAMP_DECODE;
	*d= (v[4]-'0')*10+(v[5]-'0');
	*h= (v[6]-'0')*10+(v[7]-'0');
	*m=  (v[8]-'0')*10+(v[9]-'0');
	if (	(v[10] >= '0') && (v[10] <= '9') &&
		(v[11] >= '0') && (v[11] <= '9'))
		*s=  (v[10]-'0')*10+(v[11]-'0');

	if (err != ERR_OK) SET_LAST_ERROR(err);
	return err;
}

//==========================================================
// converts ASN1 time to time_t
//==========================================================
int asn1time2time_t(ASN1_TIME* tm, time_t* pT)
{
	int err = ERR_OK;
	struct tm tm1;
	int dmz=0;
	
	RETURN_IF_NULL_PARAM(tm);
	RETURN_IF_NULL_PARAM(pT);	
	_tzset();
	memset(&tm1, 0, sizeof(tm1));
	if(tm->type == V_ASN1_UTCTIME)
		err = decodeUTCTime(tm, &tm1.tm_year, &tm1.tm_mon, 
		&tm1.tm_mday, &tm1.tm_hour, &tm1.tm_min, &tm1.tm_sec);
	if(tm->type == V_ASN1_GENERALIZEDTIME)
		err = decodeGeneralizedTime(tm, &tm1.tm_year, &tm1.tm_mon, 
		&tm1.tm_mday, &tm1.tm_hour, &tm1.tm_min, &tm1.tm_sec);
	//tm1.tm_year += 70;
	tm1.tm_mon -= 1;
	tm1.tm_isdst = _daylight;
	(*pT) = mktime(&tm1);
	if(_daylight!=0) {
		if(_timezone<0){
			dmz = (_timezone / 3600) - _daylight;
		}else{
			dmz = (_timezone / 3600) + _daylight;
		}
	}else{
		dmz=_timezone / 3600;
	}
	(*pT) = (*pT) - (dmz * 3600);	
	
	if (err != ERR_OK) SET_LAST_ERROR(err);
	return err;
}


//==========================================================
// converts ASN1 time to string
//==========================================================
int asn1time2strYear(const SignedDoc* pSigDoc, ASN1_TIME* tm, char* buf, int year)
{
  int err = ERR_OK;
  struct tm tm1, tm2;
  time_t t2;
  int dmz=0;
	
  RETURN_IF_NULL_PARAM(tm);
  RETURN_IF_NULL_PARAM(buf);	
  //RETURN_IF_NULL_PARAM(pSigDoc);	
  _tzset();
  memset(&tm1, 0, sizeof(tm1));
  if(tm->type == V_ASN1_UTCTIME)
    err = decodeUTCTime(tm, &tm1.tm_year, &tm1.tm_mon, 
			&tm1.tm_mday, &tm1.tm_hour, &tm1.tm_min, &tm1.tm_sec);
  if(tm->type == V_ASN1_GENERALIZEDTIME)
    err = decodeGeneralizedTime(tm, &tm1.tm_year, &tm1.tm_mon, 
				&tm1.tm_mday, &tm1.tm_hour, &tm1.tm_min, &tm1.tm_sec);

  // V 1.76 - in format 1.3 we use CCYY-MM-DDTHH:MM:SSZ
  if(!pSigDoc || (pSigDoc && !strcmp(pSigDoc->szFormatVer, DIGIDOC_XML_1_3_VER))) {
    sprintf(buf, "%04d-%02d-%02dT%02d:%02d:%02dZ",
	    year + tm1.tm_year, tm1.tm_mon, tm1.tm_mday, 
	    tm1.tm_hour, tm1.tm_min, tm1.tm_sec);
  } else
    // in version 1.0 we use format CCYY.MM.DDTHH:MM:SS-TZ
    if(pSigDoc && !strcmp(pSigDoc->szFormatVer, SK_XML_1_VER)) {
      tm1.tm_year -= 1900 - year;
      tm1.tm_mon -= 1;
      tm1.tm_isdst = _daylight;
      t2 = mktime(&tm1);
      if (err == ERR_OK) {
	if (_daylight!=0) {
	  if (_timezone<0) {
	    dmz = (_timezone / 3600) - _daylight;
	  } else {
	    dmz = (_timezone / 3600) + _daylight;
	  }
	} else {
	  dmz=_timezone / 3600;
	}
	t2 -= (dmz * 3600);
	ddocLocalTime(&t2, &tm2, 1);
	sprintf(buf, "%04d.%02d.%02dT%02d:%02d:%02d%+03d:00",
		tm2.tm_year + 1900, tm2.tm_mon + 1, 
		tm2.tm_mday, tm2.tm_hour, tm2.tm_min, 
		tm2.tm_sec, dmz * -1);
	ddocDebug(5, "asn1time2strYear", 
		  "ASN1 time: %s, tz: %d, string: %s\n", tm->data, dmz * -1, buf);		
      }
    } else { // in version 1.1 we use format CCYY.MM.DDTHH:MM:SSZ and allways UTC time
      sprintf(buf, "%04d.%02d.%02dT%02d:%02d:%02dZ",
	      year + tm1.tm_year, tm1.tm_mon, tm1.tm_mday, 
	      tm1.tm_hour, tm1.tm_min, tm1.tm_sec);
    }
  
  if (err != ERR_OK) SET_LAST_ERROR(err);
  return err;
}

//==========================================================
// converts ASN1 time to string
//==========================================================
int asn1time2str(const SignedDoc* pSigDoc, ASN1_TIME* tm, char* buf)
{
  return asn1time2strYear(pSigDoc, tm, buf, 0);
}

//===================================================================
// converts string to timestamp
// IN  const char* szTimestamp  - timestamp string
// OUT Timestamp* pTimestamp
//===================================================================
EXP_OPTION int convertStringToTimestamp(const SignedDoc* pSigDoc, const char* szTimestamp, Timestamp* pTimestamp)
{
	RETURN_IF_NULL_PARAM(szTimestamp);
    RETURN_IF_NULL_PARAM(pTimestamp);
    RETURN_IF_NULL_PARAM(pSigDoc);
        // in version 1.3 we use format CCYY-MM-DDTHH:MM:SS-TZ
	if(!strcmp(pSigDoc->szFormatVer, DIGIDOC_XML_1_3_VER)) {
		sscanf(szTimestamp, "%04d-%02d-%04dT%02d:%02d:%02dZ",
			&(pTimestamp->year), &(pTimestamp->mon), &(pTimestamp->day), 
			&(pTimestamp->hour), &(pTimestamp->min), &(pTimestamp->sec));
		pTimestamp->tz = 0;
	} else
	// in version 1.0 we use format CCYY.MM.DDTHH:MM:SS-TZ
	if(!strcmp(pSigDoc->szFormatVer, SK_XML_1_VER)) {
		sscanf(szTimestamp, "%04d.%02d.%04dT%02d:%02d:%02d%3d:00",
			&(pTimestamp->year), &(pTimestamp->mon), &(pTimestamp->day), 
			&(pTimestamp->hour), &(pTimestamp->min), &(pTimestamp->sec), &(pTimestamp->tz));
	} else { // in version 1.1 we use format CCYY.MM.DDTHH:MM:SSZ and allways UTC time
		sscanf(szTimestamp, "%04d.%02d.%04dT%02d:%02d:%02dZ",
			&(pTimestamp->year), &(pTimestamp->mon), &(pTimestamp->day), 
			&(pTimestamp->hour), &(pTimestamp->min), &(pTimestamp->sec));
		pTimestamp->tz = 0;
	}
	return ERR_OK;
}

//===================================================================
// converts timestamp string to time_t value
// IN  const char* szTimestamp  - timestamp string
// OUT Timestamp* pTimestamp
//===================================================================
EXP_OPTION time_t convertStringToTimeT(const SignedDoc* pSigDoc, const char* szTimestamp)
{
	struct tm tm1;
	int tz, dmz = 0;
	time_t t2;

		
	memset(&tm1, 0, sizeof(tm1));
	tzset();
	t2 = 0;
	// in version 1.0 we use format CCYY.MM.DDTHH:MM:SS-TZ
	if(!strcmp(pSigDoc->szFormatVer, SK_XML_1_VER)) {
		sscanf(szTimestamp, "%04d.%02d.%02dT%02d:%02d:%02d+%03d:00", 
			&(tm1.tm_year), &(tm1.tm_mon), &(tm1.tm_mday), 
			&(tm1.tm_hour) , &(tm1.tm_min), &(tm1.tm_sec), &tz);
	} else if(!strcmp(pSigDoc->szFormatVer, DIGIDOC_XML_1_1_VER) ||
			  !strcmp(pSigDoc->szFormatVer, DIGIDOC_XML_1_2_VER)) {  // in version 1.1 we use format CCYY.MM.DDTHH:MM:SSZ and allways UTC time
		sscanf(szTimestamp, "%04d.%02d.%02dT%02d:%02d:%02dZ", 
			&(tm1.tm_year), &(tm1.tm_mon), &(tm1.tm_mday), 
			&(tm1.tm_hour), &(tm1.tm_min), &(tm1.tm_sec));
	} else {  // in version 1.3 we use format CCYY-MM-DDTHH:MM:SSZ and allways UTC time
		sscanf(szTimestamp, "%04d-%02d-%02dT%02d:%02d:%02dZ", 
			&(tm1.tm_year), &(tm1.tm_mon), &(tm1.tm_mday), 
			&(tm1.tm_hour), &(tm1.tm_min), &(tm1.tm_sec));
	}
	tm1.tm_year -= 1900;
	tm1.tm_mon -= 1;
	tm1.tm_isdst = daylight;				
	t2 = mktime(&tm1);
	if(_daylight != 0) {
		if(_timezone < 0)
			dmz = (_timezone / 3600) - daylight;
		else
			dmz = (_timezone / 3600) + daylight;
	}
	else
		dmz = _timezone / 3600;
	t2 -= (dmz * 3600);
	return t2;
}


//===================================================================
// converts string to timestamp
// IN  const char* szTimestamp  - timestamp string
// OUT Timestamp* pTimestamp
//===================================================================
EXP_OPTION int convertTimestampToString(const SignedDoc* pSigDoc, const Timestamp* pTimestamp, char* szTimestamp)
{
  RETURN_IF_NULL_PARAM(szTimestamp);
  RETURN_IF_NULL_PARAM(pTimestamp);
  //RETURN_IF_NULL_PARAM(pSigDoc); // if null then latest format is used

  // in version 1.3 we use format CCYY-MM-DDTHH:MM:SS-TZ
  if(!pSigDoc || (pSigDoc && !strcmp(pSigDoc->szFormatVer, DIGIDOC_XML_1_3_VER))) {
    sprintf(szTimestamp, "%04d-%02d-%02dT%02d:%02d:%02dZ",pTimestamp->year,
	    pTimestamp->mon, pTimestamp->day, pTimestamp->hour , pTimestamp->min,
	    pTimestamp->sec);
  } else	// in version 1.0 we use format CCYY.MM.DDTHH:MM:SS-TZ
    if(!strcmp(pSigDoc->szFormatVer, SK_XML_1_VER)) {
      sprintf(szTimestamp, "%04d.%02d.%02dT%02d:%02d:%02d%+03d:00",pTimestamp->year,
	      pTimestamp->mon, pTimestamp->day, pTimestamp->hour , pTimestamp->min,
	      pTimestamp->sec, pTimestamp->tz);
    } else {  // in version 1.1 we use format CCYY.MM.DDTHH:MM:SSZ and allways UTC time
      sprintf(szTimestamp, "%04d.%02d.%02dT%02d:%02d:%02dZ",pTimestamp->year,
	      pTimestamp->mon, pTimestamp->day, pTimestamp->hour , pTimestamp->min,
	      pTimestamp->sec);
    }
  return ERR_OK;
}

// converts date integers to timestamp object
EXP_OPTION int Timestamp_new(Timestamp **ppTimestamp, int year,int month,int day,int hour,int minute,int second,int timezone){
	Timestamp* pTimestamp;
	pTimestamp = (Timestamp*)malloc(sizeof(Timestamp));
	RETURN_IF_BAD_ALLOC(pTimestamp);
	memset(pTimestamp, 0, sizeof(Timestamp));
	pTimestamp->year=year;
	pTimestamp->mon=month;
	pTimestamp->day=day;
	pTimestamp->hour=hour;
	pTimestamp->min=minute;
	pTimestamp->sec=second;
	pTimestamp->tz=timezone;
	*ppTimestamp = pTimestamp;
	return ERR_OK;
}

//======================================================================
// frees timestamp object
//======================================================================
EXP_OPTION void Timestamp_free(Timestamp* pTimestamp){
	free(pTimestamp);
}


//==============================================================================
// converts ASN1_GENERALIZEDTOME object to string
//==============================================================================
int str2asn1time(const SignedDoc* pSigDoc, const char* str, ASN1_GENERALIZEDTIME* asn1tm)
{
  char buf[50];
  int rc, dmz = 0;
  struct tm tm1, tm2;
  time_t t2;

	
  RETURN_IF_NULL_PARAM(str);
  RETURN_IF_NULL_PARAM(asn1tm);
  _tzset();
  memset(&tm1, 0, sizeof(tm1));
  // in version 1.0 we use format CCYY.MM.DDTHH:MM:SS-TZ
  if(!strcmp(pSigDoc->szFormatVer, SK_XML_1_VER)) {
    sscanf(str, "%04d.%02d.%04dT%02d:%02d:%02d%3d:00",
	   &(tm1.tm_year), &(tm1.tm_mon), &(tm1.tm_mday), 
	   &(tm1.tm_hour), &(tm1.tm_min), &(tm1.tm_sec), &dmz);
    tm1.tm_year -= 1900;
    tm1.tm_mon -= 1;
    tm1.tm_isdst = _daylight;
    t2 = mktime(&tm1);
    t2 -= (dmz * 3600);
    ddocLocalTime(&t2, &tm2, 1);
    sprintf(buf,"%04d%02d%02d%02d%02d%02dZ", 
	    tm2.tm_year + 1900, tm2.tm_mon + 1, tm2.tm_mday, 
	    tm2.tm_hour, tm2.tm_min, tm2.tm_sec);
  } else { // in version 1.1 we use format CCYY.MM.DDTHH:MM:SSZ and allways UTC time
    sscanf(str, "%04d.%02d.%04dT%02d:%02d:%02dZ",
	   &(tm1.tm_year), &(tm1.tm_mon), &(tm1.tm_mday), 
	   &(tm1.tm_hour), &(tm1.tm_min), &(tm1.tm_sec));
    sprintf(buf,"%04d%02d%02d%02d%02d%02dZ", 
	    tm1.tm_year, tm1.tm_mon, tm1.tm_mday, 
	    tm1.tm_hour, tm1.tm_min, tm1.tm_sec);
  }
  rc = ASN1_GENERALIZEDTIME_set_string(asn1tm, buf);
  //printf("String: %s, tz: %d, ASN1 time: %s\n", str, dmz, asn1tm->data);
  return ERR_OK;
}


//==========< signed doc fucntions >================

//============================================================
// Sets a string element of a struct to a new value
// dest - element pointer
// value - new value
// valLen - value length (use -1 for null terminated strings)
//============================================================
int setString(char** dest, const char* value, int valLen) 
{
  RETURN_IF_NULL_PARAM(dest);
  RETURN_IF_NULL_PARAM(value);

  if(*dest) {
    free(*dest);
    *dest = NULL;
  }
  if(valLen == -1) {
    *dest = (char*)malloc(strlen(value)+1);
    RETURN_IF_BAD_ALLOC(*dest);
    strcpy(*dest, value);
  } else {
    *dest = (char*)malloc(valLen);
    RETURN_IF_BAD_ALLOC(*dest);
    memcpy(*dest, value, valLen);
  }	
  return ERR_OK;
}

//============================================================
// Get the filename part of full file name
//============================================================
EXP_OPTION const char* getSimpleFileName(const char* szFileName)
{
	const char *p;

	RETURN_OBJ_IF_NULL(szFileName, NULL);
	
	for(p = szFileName + strlen(szFileName) -1; 
		(p > szFileName) && (*p != '\\') && (*p != '/');
		p--);
	if((*p == '\\') || (*p == '/'))
		p++;
	return p;
}

//============================================================
// Get the absolute filename with path
//============================================================
EXP_OPTION int getFullFileName(const char* szFileName, char* szDest, int len)
{
  int err = ERR_OK;
  memset(szDest, 0, len);
  // if this is a filename with path then we are ready
  if(strchr(szFileName, '/') || strchr(szFileName, '\\')) {
    strcpy(szDest, szFileName);
  } else { // local filename, must prepend directory
    _getcwd(szDest, len);
#ifdef WIN32
    strcat(szDest, "\\");
#else
    strcat(szDest, "/");
#endif
    strcat(szDest, szFileName);
  }
  return err;
}


//============================================================
// Get the path part of full file name
//============================================================
EXP_OPTION int getFileNamePath(const char* szFileName, char* szPath)
{
	int nFound = 0, i, err = ERR_OK;

	RETURN_IF_NULL_PARAM(szFileName);
	RETURN_IF_NULL_PARAM(szPath);
	strcpy(szPath, szFileName);
	for(i = strlen(szPath); i >= 0; i--) {
	if(szPath[i] == '\\') {
		  szPath[i+1] = 0;
		  nFound = 1;
		  break;
	  }
	}
	if(!nFound)
		szPath[0] = 0;
	return err;
}

//============================================================
// Allocates a new SignedDoc element and initializes it
// format - format name
// version - format version
//============================================================
EXP_OPTION int SignedDoc_new(SignedDoc **pSignedDoc, const char* format, const char* version)
{
	SignedDoc* pSigDoc = NULL;
	
	RETURN_IF_NULL_PARAM(format);
	RETURN_IF_NULL_PARAM(version);
	
	pSigDoc = (SignedDoc*)malloc(sizeof(SignedDoc));
	RETURN_IF_BAD_ALLOC(pSigDoc);
	memset(pSigDoc, 0, sizeof(SignedDoc));
	if((!strcmp(format, SK_XML_1_NAME) && !strcmp(version, SK_XML_1_VER)) ||
		(!strcmp(format, DIGIDOC_XML_1_1_NAME) && 
		(!strcmp(version, DIGIDOC_XML_1_1_VER) || 
		 !strcmp(version, DIGIDOC_XML_1_2_VER) ||
		 // Ver 1.76 - add support for format 1.3
                 !strcmp(version, DIGIDOC_XML_1_3_VER)))) {
		setString(&(pSigDoc->szFormat), format, -1);
		setString(&(pSigDoc->szFormatVer), version, -1);
	}
	else
		SET_LAST_ERROR_RETURN_CODE(ERR_UNSUPPORTED_VERSION);
 	*pSignedDoc = pSigDoc;
	return ERR_OK;
}


//============================================================
// Frees the memory of SignedDoc element
// pSigDoc - element to free
//============================================================
EXP_OPTION void SignedDoc_free(SignedDoc* pSigDoc) 
{
  int i;

  RETURN_VOID_IF_NULL(pSigDoc);
  for(i = 0; (pSigDoc->pDataFiles != NULL) && 
	(i < pSigDoc->nDataFiles); i++) 
    DataFile_free(pSigDoc->pDataFiles[i]);
  if(pSigDoc->pDataFiles)
    free(pSigDoc->pDataFiles);
  for(i = 0; (pSigDoc->pSignatures != NULL) && 
	(i < pSigDoc->nSignatures); i++)
    SignatureInfo_free(pSigDoc->pSignatures[i]);
  if(pSigDoc->pSignatures)
    free(pSigDoc->pSignatures);
  for(i = 0; (pSigDoc->pNotaries != NULL) && 
	(i < pSigDoc->nNotaries); i++) 
    NotaryInfo_free(pSigDoc->pNotaries[i]);
  if(pSigDoc->pNotaries)
    free(pSigDoc->pNotaries);
  if(pSigDoc->szFormat)
    free(pSigDoc->szFormat);
  if(pSigDoc->szFormatVer)
    free(pSigDoc->szFormatVer);
  if(pSigDoc)
    free(pSigDoc);
}


//============================================================
// Returns the number of data files
// pSigDoc - signed doc pointer
//============================================================
EXP_OPTION int getCountOfDataFiles(const SignedDoc* pSigDoc)
{
	RETURN_OBJ_IF_NULL(pSigDoc, 0);

	return pSigDoc->nDataFiles;
}

//============================================================
// Returns the next free data file id
// pSigDoc - signed doc pointer
//============================================================
EXP_OPTION int getNextDataFileId(const SignedDoc* pSigDoc)
{
	int id = 0, n, i;

	RETURN_OBJ_IF_NULL(pSigDoc, 0);

	for(i = 0; i < pSigDoc->nDataFiles; i++) {
		DataFile* pDataFile = pSigDoc->pDataFiles[i];
		RETURN_OBJ_IF_NULL(pDataFile, 0);
    		RETURN_OBJ_IF_NULL(pDataFile->szId, 0);
		SET_LAST_ERROR_RETURN_IF_NOT(strlen(pDataFile->szId) > 1, ERR_EMPTY_STRING, 0);
		n = atoi(pDataFile->szId+1);
		if(id <= n)
			id = n+1;
	}
	return id;
}

//============================================================
// Returns the desired DataFile object
// pSigDoc - signed doc pointer
// nIdx - DataFile index (starting with 0)
//============================================================
EXP_OPTION DataFile* getDataFile(const SignedDoc* pSigDoc, int nIdx)
{
	RETURN_OBJ_IF_NULL(pSigDoc, NULL);
	SET_LAST_ERROR_RETURN_IF_NOT(nIdx < pSigDoc->nDataFiles, ERR_BAD_DATAFILE_INDEX, NULL);
	RETURN_OBJ_IF_NULL(pSigDoc->pDataFiles[nIdx], 0);
	return pSigDoc->pDataFiles[nIdx];
}

//============================================================
// Returns the last DataFile object
// pSigDoc - signed doc pointer
//============================================================
EXP_OPTION DataFile* ddocGetLastDataFile(const SignedDoc* pSigDoc)
{
  int nIdx;
  RETURN_OBJ_IF_NULL(pSigDoc, NULL);
  nIdx = pSigDoc->nDataFiles - 1;
  SET_LAST_ERROR_RETURN_IF_NOT(nIdx >= 0, ERR_BAD_DATAFILE_INDEX, NULL);
  RETURN_OBJ_IF_NULL(pSigDoc->pDataFiles[nIdx], 0);
  return pSigDoc->pDataFiles[nIdx];
}

//============================================================
// Returns the DataFile object with the given id
// pSigDoc - signed doc pointer
// id - DataFile id
//============================================================
EXP_OPTION DataFile* getDataFileWithId(const SignedDoc* pSigDoc, const char* id)
{
	DataFile* pDataFile = NULL;
	int i;
	//AA 12.11.2003
	RETURN_OBJ_IF_NULL(id, NULL);
	RETURN_OBJ_IF_NULL(pSigDoc, NULL);
	ddocDebug(4, "getDataFileWithId", "id: \'%s\', files: %d", id, pSigDoc->nDataFiles);
	for(i = 0; i < pSigDoc->nDataFiles; i++) {
	  ddocDebug(4, "getDataFileWithId", "pos: %d - \'%s\', files: %d", i, pSigDoc->pDataFiles[i]->szId, pSigDoc->nDataFiles);
		RETURN_OBJ_IF_NULL(pSigDoc->pDataFiles[i], NULL);
		RETURN_OBJ_IF_NULL(pSigDoc->pDataFiles[i]->szId, NULL);
		if(!strcmp(pSigDoc->pDataFiles[i]->szId, id)) {
			pDataFile = pSigDoc->pDataFiles[i];
			break;
		}
	}
	return pDataFile;
}


//============================================================
// Adds a new DataFile element to  a SignedDoc element and initializes it
// pSigDoc - signed document
// id - data file id
// filename - filename
// contentType - DETATCHED, EMBEDDED or EMBEDDED_BASE64
// mime - mime type
// size - file size
// digType - digestType
// digest - file digest (SHA1)
// digLen - digest length
//============================================================
EXP_OPTION int DataFile_new(DataFile **newDataFile, 
			    SignedDoc* pSigDoc, const char* id,
			    const char* filename, const char* contentType, 
			    const char* mime, long size,
			    const byte* digest, int digLen,
			    const char* digType, const char* szCharset)
{
  char buf1[300], *p;
  int nId = 0, i;
  DataFile **pDataFiles;
  DataFile *pDataFile;
  FILE* hFile;

  RETURN_IF_NULL_PARAM(newDataFile);
  RETURN_IF_NULL_PARAM(pSigDoc);
  ddocDebug(4, "DataFile_new", "SigDoc ver: %s, file: %s, contentType: %s, mimeType: %s", 
	    (pSigDoc ? pSigDoc->szFormatVer : "NULL"), (filename ? filename : "NULL"), contentType, mime);
  if(filename) {
    p = (char*)malloc(strlen(filename) * 5);
    if(p) {
      memset(p, 0, strlen(filename) * 5);
      for(i = 0; i < (int)strlen(filename); i++)
	sprintf(strchr(p, 0), " %02x", (unsigned char)filename[i]);
      ddocDebug(4, "DataFile_new", "FileName: %s - %s", p, filename);
      free(p);
      p = 0;
    }
  }
  //clearErrors();
  if(!id)
    nId = getNextDataFileId(pSigDoc);
  
  if(pSigDoc->nDataFiles == 0) {
    SET_LAST_ERROR_RETURN_IF_NOT(!pSigDoc->pDataFiles, ERR_BAD_DATAFILE_COUNT, 0);
    pSigDoc->nDataFiles = 1;
  }
  else
    pSigDoc->nDataFiles++;	
  pDataFiles = (DataFile**)malloc((pSigDoc->nDataFiles) * sizeof(void *));
  
  RETURN_IF_BAD_ALLOC(pDataFiles);
  for(i = 0; i < pSigDoc->nDataFiles-1; i++)
    pDataFiles[i] = pSigDoc->pDataFiles[i];
  pDataFile = (DataFile*)malloc(sizeof(DataFile));
  RETURN_IF_BAD_ALLOC(pDataFile);
  memset(pDataFile, 0, sizeof(DataFile));
  pDataFiles[pSigDoc->nDataFiles-1] = pDataFile;
  if(pSigDoc->pDataFiles)
    free(pSigDoc->pDataFiles);
  pSigDoc->pDataFiles = pDataFiles;
  if(id) {
    setString(&(pDataFile->szId), id, -1);
  } else {		
    sprintf(buf1, "D%d", nId);
    setString(&(pDataFile->szId), buf1, -1);
  }
  if(szCharset)
    setString(&(pDataFile->szCharset), szCharset, -1);
  else
    setString(&(pDataFile->szCharset), CHARSET_ISO_8859_1, -1);
  if(filename) {
    // in versions 1.0, 1.1 and 1.2 we used wrong encoding for OEM windows charset
    setString(&(pDataFile->szFileName), filename, -1);
    if(!strcmp(contentType, CONTENT_EMBEDDED)) {
      if((hFile = fopen(pDataFile->szFileName, "rt")) != NULL) {
	fgets(buf1, sizeof(buf1), hFile);
	if(strstr(buf1, "<?xml")) 
	  SET_LAST_ERROR(ERR_BAD_DATAFILE_XML);
	fclose(hFile);
      }
    }
  }
  if(mime)
    setString(&(pDataFile->szMimeType), mime, -1);
  if(contentType)
    setString(&(pDataFile->szContentType), contentType, -1);
  pDataFile->nSize = size;
  if(digType && strlen(digType))
    setString(&(pDataFile->szDigestType), digType, -1);
  if(digest && digLen) {
    pDataFile->szDigest = 0;
    setString((char**)&(pDataFile->szDigest), 
	      (const char*)digest, digLen);
    pDataFile->nDigestLen = digLen;
  }
  *newDataFile = pDataFile;
  return ERR_OK;
}


//============================================================
// Removes this DataFile from signed doc and frees it's memory
// pSigDoc - signed doc object
// id - DataFile id to be removed
//============================================================
EXP_OPTION int DataFile_delete(SignedDoc* pSigDoc, const char* id)
{
	int err = ERR_OK, n, i, j;
	DataFile* pDataFile = NULL;
	DataFile** pDataFiles = NULL;

	RETURN_IF_NULL_PARAM(pSigDoc);
	if(pSigDoc->nSignatures > 0)
		SET_LAST_ERROR_RETURN_CODE(ERR_MODIFY_SIGNED_DOC);	
	if((pDataFile = getDataFileWithId(pSigDoc, id)) != 0) {
		n = pSigDoc->nDataFiles - 1;
		if(n > 0) {
			pDataFiles = (DataFile**)malloc(n * sizeof(void*));
			RETURN_IF_BAD_ALLOC(pDataFiles);
			for(i = j = 0; i < pSigDoc->nDataFiles; i++) {
				if(strcmp(pSigDoc->pDataFiles[i]->szId, id)) 
					pDataFiles[j++] = pSigDoc->pDataFiles[i];
				else
					DataFile_free(pSigDoc->pDataFiles[i]);
			}
			free(pSigDoc->pDataFiles);
			pSigDoc->pDataFiles = pDataFiles;
			pSigDoc->nDataFiles = n;

		} else {
			for(i = 0; i < pSigDoc->nDataFiles; i++)
				DataFile_free(pSigDoc->pDataFiles[i]);
			free(pSigDoc->pDataFiles);
			pSigDoc->pDataFiles = NULL;
			pSigDoc->nDataFiles = 0;
		}

	}
	else
		err = ERR_BAD_DATAFILE_INDEX;

	if (err != ERR_OK) SET_LAST_ERROR(err);
	return err;
}

//============================================================
// Retrieve and convert cached DataFile data if possible
// pSigDoc - signed document object
// szDocId - datafile id
// ppBuf - address of buffer pointer
// pLen - address of lenght of bytes
//============================================================
int ddocGetDataFileCachedData(SignedDoc* pSigDoc, const char* szDocId, void** ppBuf, long* pLen)
{
  DataFile* pDf;

  //RETURN_IF_NULL_PARAM(pSigDoc); // if null then don't check for cached data (old logic)
  RETURN_IF_NULL_PARAM(szDocId);
  RETURN_IF_NULL_PARAM(ppBuf);
  RETURN_IF_NULL_PARAM(pLen);
  *ppBuf = 0;
  *pLen = 0;
  if(pSigDoc) {
    pDf = getDataFileWithId(pSigDoc, szDocId);
    if(pDf && pDf->szContent) { // gotcha!
		int len;
		// base64 content, allocate exact length and initialize
		if(!strcmp(pDf->szContentType, CONTENT_EMBEDDED_BASE64)) {
			*ppBuf = malloc(pDf->nContentLen);
			*pLen = pDf->nContentLen;
			RETURN_IF_BAD_ALLOC(*ppBuf);
			memset(*ppBuf, 0, *pLen);
			len = (int)*pLen;
			decode((const byte*)pDf->szContent, pDf->nContentLen, (byte*)*ppBuf, &len);
			*pLen = len;
		} 
		// simple text content. Make it zero terminated string
		else if(!strcmp(pDf->szContentType, CONTENT_EMBEDDED)) {
			*ppBuf = malloc(pDf->nContentLen+1);
			RETURN_IF_BAD_ALLOC(*ppBuf);
			*pLen = pDf->nContentLen;
			memcpy(*ppBuf, pDf->szContent, pDf->nContentLen);
			((char*)(*ppBuf))[*pLen] = 0;
		}
    }
  }
  return ERR_OK;
}

//============================================================
// Retrieve and convert DataFile Filename atribute and convert
// to proper UTF-8 if necessary.
// pSigDoc - signed document object
// szDocId - datafile id
// ppBuf - address of buffer pointer. Caller must free the buffer
// pLen - address of lenght of bytes. Will be changed.
//============================================================
int ddocGetDataFileFilename(SignedDoc* pSigDoc, const char* szDocId, void** ppBuf, int* pLen)
{
  DataFile* pDf;

  RETURN_IF_NULL_PARAM(pSigDoc);
  RETURN_IF_NULL_PARAM(szDocId);
  RETURN_IF_NULL_PARAM(ppBuf);
  RETURN_IF_NULL_PARAM(pLen);
  *ppBuf = 0;
  *pLen = 0;
  if(pSigDoc) {
    pDf = getDataFileWithId(pSigDoc, szDocId);
    if(pDf && pDf->szFileName) { // gotcha!
      // duplicate filename first.
      *ppBuf = (char*)strdup(pDf->szFileName);
      RETURN_IF_BAD_ALLOC(*ppBuf);
      *pLen = strlen((const char*)*ppBuf);
      // in version 1.2 and earlier we had bad UTF-8 for some chars
      // check and fix it for newer clients
      if(!strcmp(pSigDoc->szFormatVer, SK_XML_1_VER) ||
	     !strcmp(pSigDoc->szFormatVer, DIGIDOC_XML_1_1_VER) ||
	     !strcmp(pSigDoc->szFormatVer, DIGIDOC_XML_1_2_VER)) {
        convWinToFName(pDf->szFileName, (char*)*ppBuf);
      }
    }
  }
  return ERR_OK;
}

//--------------------------------------------------
// Checks if the size of this DataFile is less than
// max size for memory cache and if so caches the data.
// pDf - DataFile object
// maxLen - max cacheable DataFile size
// value - character values read from file
// len - length of chars ???
//--------------------------------------------------
void ddocAppendDataFileData(DataFile* pDf, int maxLen, void* data, int len)
{
  ddocDebug(5, "ddocAppendDataFileData", "append: %d, max: %d", len, maxLen);
  if(pDf && pDf->nSize < maxLen) {
    ddocDebug(6, "ddocAppendDataFileData", "DF: %s, size: %d, max: %d", pDf->szId, pDf->nSize, maxLen);
    pDf->szContent = (byte*)realloc(pDf->szContent, pDf->nContentLen + len + 1);
    if(!pDf->szContent) {
      ddocDebug(1, "ddocAppendDataFileData", "Error reallocing from: %d to %d", 
		pDf->nContentLen, pDf->nContentLen + len + 1);
      SET_LAST_ERROR(ERR_BAD_ALLOC);
      return;
    }
    memset(pDf->szContent + pDf->nContentLen, 0, len+1);
    memcpy(pDf->szContent + pDf->nContentLen, data, len);
    pDf->nContentLen += len;
  }
}


//============================================================
// cleanup DataFile memory
// pDataFile - data file object to be cleaned up
//============================================================
EXP_OPTION void DataFile_free(DataFile* pDataFile)
{
	int i=0;

	RETURN_VOID_IF_NULL(pDataFile);
	if(pDataFile->szId)
		free(pDataFile->szId);
	if(pDataFile->szFileName)
		free(pDataFile->szFileName);
	if(pDataFile->szMimeType)
		free(pDataFile->szMimeType);
	if(pDataFile->szDigestType)
		free(pDataFile->szDigestType);
	if(pDataFile->szDigest)
		free(pDataFile->szDigest);
	if(pDataFile->szDetachedDigest)
		free(pDataFile->szDetachedDigest);
	if(pDataFile->szContentType)
		free(pDataFile->szContentType);
	for(i = 0; i < pDataFile->nAttributes; i++) {
		free(pDataFile->pAttNames[i]);
		free(pDataFile->pAttValues[i]);
	}
	if(pDataFile->szCharset)
		free(pDataFile->szCharset);
	if(pDataFile->pAttNames)
		free(pDataFile->pAttNames);
	if(pDataFile->pAttValues)
		free(pDataFile->pAttValues);
	if(pDataFile->szContent)
	  free(pDataFile->szContent);
	free(pDataFile);
}

//============================================================
// Removes this NotaryInfo from signed doc and frees it's memory
// pSigDoc - signed doc object
// id - notary id to be removed
//============================================================
EXP_OPTION int NotaryInfo_delete(SignedDoc* pSigDoc, const char* id)
{
	int n, i, j, err = ERR_OK;
	NotaryInfo* pNotary;
	NotaryInfo** pNotaries;

	RETURN_IF_NULL_PARAM(pSigDoc);
	RETURN_IF_NULL_PARAM(id);
	if((pNotary = getNotaryWithId(pSigDoc, id)) != NULL) {
		n = pSigDoc->nNotaries - 1;
		if(n > 0) {
			pNotaries = (NotaryInfo**)malloc(n * sizeof(void*));
			RETURN_IF_BAD_ALLOC(pNotaries);
			for(i = j = 0; i < pSigDoc->nNotaries; i++) {
				if(strcmp(pSigDoc->pNotaries[i]->szId, id)) 
					pNotaries[j++] = pSigDoc->pNotaries[i];					
				else
					NotaryInfo_free(pSigDoc->pNotaries[i]);
			}
			free(pSigDoc->pNotaries);
			pSigDoc->pNotaries = pNotaries;
			pSigDoc->nNotaries = n;

		} else {
			for(i = 0; i < pSigDoc->nNotaries; i++)
				NotaryInfo_free(pSigDoc->pNotaries[i]);
			free(pSigDoc->pNotaries);
			pSigDoc->pNotaries = NULL;
			pSigDoc->nNotaries = 0;
		}
	}
	else
		err = ERR_BAD_NOTARY_ID;

	if (err != ERR_OK) SET_LAST_ERROR(err);
	return err;
}

//============================================================
// Returns number of DataFile attributes
// pDataFile - data file
//============================================================
EXP_OPTION int getCountOfDataFileAttributes(const DataFile* pDataFile)
{
	RETURN_OBJ_IF_NULL(pDataFile, -1);
	return pDataFile->nAttributes;
}


//============================================================
// Adds an attribute to data file
// pDataFile - data file
// name - attribute name
// value - attribute value
//============================================================
// FIXME : Badly in need for a rewrite - memory leaks, when something fails.
int addDataFileAttribute(DataFile* pDataFile, const char* name, const char* value)
{
	char	**pp;
	int i;

	RETURN_IF_NULL_PARAM(pDataFile);
	RETURN_IF_NULL_PARAM(name);
	RETURN_IF_NULL_PARAM(value);
	if(pDataFile->nAttributes == 0) {
		RETURN_IF_NOT(!pDataFile->pAttNames, ERR_BAD_ATTR_COUNT);
		RETURN_IF_NOT(!pDataFile->pAttValues, ERR_BAD_ATTR_COUNT);
		pDataFile->nAttributes = 1;
	}
	else
		pDataFile->nAttributes++;
	// set name
	pp = (char**)malloc((pDataFile->nAttributes) * sizeof(char *));
	RETURN_IF_BAD_ALLOC(pp);
	for(i = 0; i < pDataFile->nAttributes-1; i++)
		pp[i] = pDataFile->pAttNames[i];
	if(pDataFile->pAttNames)
		free(pDataFile->pAttNames);
	pDataFile->pAttNames = pp;
	pp[pDataFile->nAttributes-1] = NULL;
	setString(&(pp[pDataFile->nAttributes-1]), name, -1);
	// set value
	pp = (char**)malloc((pDataFile->nAttributes) * sizeof(char *));
	RETURN_IF_BAD_ALLOC(pp);
	for(i = 0; i < pDataFile->nAttributes-1; i++)
		pp[i] = pDataFile->pAttValues[i];
	if(pDataFile->pAttValues)
		free(pDataFile->pAttValues);
	pDataFile->pAttValues = pp;
	pp[pDataFile->nAttributes-1] = NULL;
	setString(&(pp[pDataFile->nAttributes-1]), value, -1);
	return ERR_OK;
}


//============================================================
// Gets an attribute of a data file
// pDataFile - data file
// name - buffer for attribute name pointer
// value - buffer for attribute value pointer
//============================================================
EXP_OPTION int getDataFileAttribute(DataFile* pDataFile, int idx, char** name, char** value)
{

	RETURN_IF_NULL_PARAM(pDataFile);
	RETURN_IF_NOT(idx < pDataFile->nAttributes, ERR_BAD_ATTR_INDEX);
	RETURN_IF_NULL_PARAM(name);
	RETURN_IF_NULL_PARAM(value);
	*name = pDataFile->pAttNames[idx];
	*value = pDataFile->pAttValues[idx];
	return ERR_OK;
}

//============================================================
// Creates a temporary file and returns a BIO object for it
// pBio - address of BIO pointer
//============================================================
/*int getTempBIO(BIO** pBio)
{
	int err = ERR_OK;
	FILE* hFile;

	RETURN_IF_NULL_PARAM(pBio);
	hFile = tmpfile();	
	RETURN_IF_NOT(hFile, ERR_FILE_WRITE);
	*pBio = BIO_new_fp(hFile, BIO_CLOSE);
	RETURN_IF_NOT(*pBio, ERR_FILE_WRITE);
	return err;
}*/


//============================================================
// Gets a new temporary filename
// buf - filename buffer
//============================================================
int getTempFileName(char* szFileName)
{
	char tbuf[200];
#ifdef WIN32
	char* pFileName = 0;
#endif
	RETURN_IF_NULL_PARAM(szFileName);
#ifdef WIN32
	GetTempPath(sizeof(tbuf), tbuf);
	pFileName = _tempnam(tbuf, "ddoc");
	strcpy(szFileName, pFileName);
	free(pFileName);
#else
	strcpy(tbuf, "/tmp/ddocXXXXXX");
	mkstemp(tbuf);
	strcpy(szFileName, tbuf);
#endif
	return ERR_OK;
}


//============================================================
// Calculates the file size and digest
// pSigDoc - signed document
// id - data file id
// filename - filename
// digType - digestType (code)
//============================================================
EXP_OPTION int calculateDataFileSizeAndDigest(SignedDoc* pSigDoc, const char* id,
							const char* filename, int digType)
{
  int err = ERR_OK, len1;
  char buf1[DIGEST_LEN+2], *fbuf = 0, *cbuf = 0, buf2[50];
  char fnameBuf[200];
  long len2;
  DataFile* pDataFile;
  BIO* bout;
  //FILE* hFile;

  RETURN_IF_NULL_PARAM(pSigDoc);
  pDataFile = getDataFileWithId(pSigDoc, id);
  RETURN_IF_NULL(pDataFile);
  if(digType == DIGEST_SHA1) {
    // in version 1.0 we use DigestType and DigestValue
    // attrributes of DataFile element and calculate the digest
    // over the original content
    if(!strcmp(pSigDoc->szFormatVer, SK_XML_1_VER)) {
      // if content type is HASHCODE then we are finished because
      // we are supposed to believe what the sender has stored here
      if(strcmp(pDataFile->szContentType, CONTENT_HASHCODE)) {
      len1 = sizeof(buf1);
      err = calculateFileDigest(filename, digType,
				(byte*)buf1, &len1, &len2);
      RETURN_IF_NOT(err == ERR_OK, err);
      setString((char**)&(pDataFile->szDigest), buf1, len1);
      pDataFile->nDigestLen = len1;
      pDataFile->nSize = len2;
      }
    } 
    // in version 1.1 we don't use those attributes
    // and we calculate the digest over the whole
    // DataFile element including the tags
    else {
      ddocConvertFileName(fnameBuf, sizeof(fnameBuf), pDataFile->szFileName);
      if(!strcmp(pDataFile->szContentType, CONTENT_DETATCHED)) {
	len1 = sizeof(buf1);
	err = calculateFileDigest(fnameBuf, digType,
				  (byte*)buf1, &len1, &len2);
	RETURN_IF_NOT(err == ERR_OK, err);
	setString((char**)&(pDataFile->szDigestType), DIGEST_SHA1_NAME, -1);
	setString((char**)&(pDataFile->szDetachedDigest), buf1, len1);
	pDataFile->nDetachedDigestLen = len1;
	pDataFile->nSize = len2;
      } else
	err = calculateFileSize(fnameBuf, &pDataFile->nSize);
      ddocDebug(4, "calculateDataFileSizeAndDigest", "File: %s size: %d", fnameBuf, pDataFile->nSize);
      bout = BIO_new(BIO_s_mem());
      RETURN_IF_NOT(bout, ERR_NULL_POINTER);
      err = generateDataFileXML(pSigDoc, pDataFile, fnameBuf, bout);
      if (err != ERR_OK) {
	BIO_free(bout);
	SET_LAST_ERROR_RETURN_CODE(err);
      }
      BIO_get_mem_data(bout, &fbuf);
      len2 = BIO_number_written(bout);
      ddocDebug(5, "calculateDataFileSizeAndDigest", "CanInput: %s size: %d", fbuf, len2);
      cbuf = canonicalizeXML(fbuf, len2);
      BIO_free(bout);
      RETURN_IF_NOT(cbuf, ERR_NULL_POINTER);
      // now calculate the digest
      len1 = sizeof(buf1);
      len2 = strlen(cbuf);
      SHA1((const unsigned char*)cbuf, len2, (unsigned char*)buf1);
      ddocDebug(5, "calculateDataFileSizeAndDigest", "CANONICAL XML: \'%s\'", cbuf);
      /*if((hFile = fopen("c:\\veiko\\DigiDocLib\\Debug\\df1.xml", "wb")) != NULL) {
	fputs(cbuf, hFile);
	fclose(hFile);
	}*/
      if (cbuf)
	free(cbuf);
      setString((char**)&(pDataFile->szDigest), buf1, DIGEST_LEN);
      len1 = sizeof(buf2);
      encode((const byte*)buf1, DIGEST_LEN, (byte*)buf2, &len1);
      ddocDebug(5, "calculateDataFileSizeAndDigest", "DF digest: %s", buf2);
      pDataFile->nDigestLen = DIGEST_LEN;						
    }
    setString(&(pDataFile->szDigestType), DIGEST_SHA1_NAME, -1);
  } 
  else
    SET_LAST_ERROR_RETURN_CODE(ERR_UNSUPPORTED_DIGEST);
  return ERR_OK;
}


//============================================================
// Returns the number of signatures
// pSigDoc - signed doc pointer
//============================================================
EXP_OPTION int getCountOfSignatures(const SignedDoc* pSigDoc)
{
	RETURN_OBJ_IF_NULL(pSigDoc, -1);
	return pSigDoc->nSignatures;
}

//============================================================
// Returns the next free signature id
// pSigDoc - signed doc pointer
//============================================================
EXP_OPTION int getNextSignatureId(const SignedDoc* pSigDoc)
{
	int id = 0, n, i;

	RETURN_OBJ_IF_NULL(pSigDoc, -1);
	for(i = 0; i < pSigDoc->nSignatures; i++) {
		SignatureInfo* pSignature = pSigDoc->pSignatures[i];
		RETURN_OBJ_IF_NULL(pSignature, -1);
		RETURN_OBJ_IF_NULL(pSignature->szId, -1);
    		SET_LAST_ERROR_RETURN_IF_NOT(strlen(pSignature->szId) > 1, ERR_EMPTY_STRING, -1);
		n = atoi(pSignature->szId+1);
		if(id <= n)
			id = n+1;
	}
	return id;
}

//============================================================
// Returns the desired SignatureInfo object
// pSigDoc - signed doc pointer
// nIdx - SignatureInfo index (starting with 0)
//============================================================
EXP_OPTION SignatureInfo* getSignature(const SignedDoc* pSigDoc, int nIdx)
{
	RETURN_OBJ_IF_NULL(pSigDoc, NULL);
	SET_LAST_ERROR_RETURN_IF_NOT(nIdx < pSigDoc->nSignatures, ERR_BAD_SIG_INDEX, NULL);
	RETURN_OBJ_IF_NULL(pSigDoc->pSignatures[nIdx], NULL);
	return pSigDoc->pSignatures[nIdx];
}

//============================================================
// Returns the last SignatureInfo object
// pSigDoc - signed doc pointer
//============================================================
EXP_OPTION SignatureInfo* ddocGetLastSignature(const SignedDoc* pSigDoc)
{
  int nIdx;
  RETURN_OBJ_IF_NULL(pSigDoc, NULL);
  nIdx = pSigDoc->nSignatures - 1;
  SET_LAST_ERROR_RETURN_IF_NOT(nIdx < pSigDoc->nSignatures, ERR_BAD_SIG_INDEX, NULL);
  RETURN_OBJ_IF_NULL(pSigDoc->pSignatures && pSigDoc->pSignatures[nIdx], NULL);
  return pSigDoc->pSignatures[nIdx];
}

//============================================================
// Returns the SignatureInfo object with the given id
// pSigDoc - signed doc pointer
// id - SignatureInfo id
//============================================================
EXP_OPTION SignatureInfo* getSignatureWithId(const SignedDoc* pSigDoc, const char* id)
{
  SignatureInfo* pSignature = NULL;
  int i;

  RETURN_OBJ_IF_NULL(pSigDoc, NULL);
  for(i = 0; i < pSigDoc->nSignatures; i++) {
    RETURN_OBJ_IF_NULL(pSigDoc->pSignatures[i], NULL);
    RETURN_OBJ_IF_NULL(pSigDoc->pSignatures[i]->szId, NULL);
    if(!strcmp(pSigDoc->pSignatures[i]->szId, id)) {
      pSignature = pSigDoc->pSignatures[i];
      break;
    }
  }
  return pSignature;
}





//============================================================
// Adds a new SignedInfo element to a SignedDoc element and initializes it
// id - signature id (use NULL for default)
// return the newly created structure 
//============================================================
// FIXME : memory leaks possible..
EXP_OPTION int SignatureInfo_new(SignatureInfo **newSignatureInfo, SignedDoc* pSigDoc, const char* id)
{
	int i, nId = 0;
	SignatureInfo** pSignatures = NULL;
	SignatureInfo* pSigInfo = NULL;
	char buf[100];


	RETURN_IF_NULL_PARAM(pSigDoc);

	//clearErrors();

	if(!id)
		nId = getNextSignatureId(pSigDoc);
	if(pSigDoc->nSignatures == 0) {
		RETURN_IF_NOT(!pSigDoc->pSignatures, ERR_BAD_SIG_COUNT);
		pSigDoc->nSignatures = 1;
	}
	else
		pSigDoc->nSignatures++;
	pSignatures = (SignatureInfo**)malloc((pSigDoc->nSignatures) * sizeof(void *));
	RETURN_IF_BAD_ALLOC(pSignatures);
	for(i = 0; i < pSigDoc->nSignatures-1; i++)
		pSignatures[i] = pSigDoc->pSignatures[i];
	pSigInfo = (SignatureInfo*)malloc(sizeof(SignatureInfo)); // MEMLEAK: ???
	RETURN_IF_BAD_ALLOC(pSigInfo);
	memset(pSigInfo, 0, sizeof(SignatureInfo));
	pSignatures[pSigDoc->nSignatures-1] = pSigInfo;
	if(pSigDoc->pSignatures)
		free(pSigDoc->pSignatures);
	pSigDoc->pSignatures = pSignatures;
	if(id) {
		setString(&(pSigInfo->szId), id, -1);
	} else {
		sprintf(buf, "S%d", nId);
		setString(&(pSigInfo->szId), buf, -1);
	}
	// create timestamp
	createTimestamp(pSigDoc, buf);
	setString(&(pSigInfo->szTimeStamp), buf, -1);
	*newSignatureInfo = pSigInfo;
	return ERR_OK;
}




//============================================================
// Sets the signature production place info
// pSigInfo - signature info object
// city - city name
// state - state or province name
// zip - postal code
// country - country name
//============================================================
EXP_OPTION int setSignatureProductionPlace(SignatureInfo* pSigInfo,
                                const char* city, const char* state,
                                const char* zip, const char* country)
{
  char* p = 0;
  RETURN_IF_NULL_PARAM(pSigInfo);
  if(city) {
    p = escape2xmlsym(city);
    if(p) {
      setString(&(pSigInfo->sigProdPlace.szCity), p, -1);
      free(p);
    }
  }
  if(state) {
    p = escape2xmlsym(state);
    if(p) {
      setString(&(pSigInfo->sigProdPlace.szStateOrProvince), p, -1);
      free(p);
    }
  }
  if(zip) {
    p = escape2xmlsym(zip);
    if(p) {
      setString(&(pSigInfo->sigProdPlace.szPostalCode), p, -1);
      free(p);
    }
  }
  if(country) {
    p = escape2xmlsym(country);
    if(p) {
      setString(&(pSigInfo->sigProdPlace.szCountryName), country, -1);
      free(p);
    }
  }
  return ERR_OK;
}



//============================================================
// Adds a signer role 
// pSigInfo - signature info object
// nCertified - certified role? (1/0)
// role - role data
// rLen - role data length
// encode - 1=encode it with Base64, 0=use as is
//============================================================
// FIXME : memory leaks possible...
EXP_OPTION int addSignerRole(SignatureInfo* pSigInfo, int nCertified, 
				   const char* role, int rLen, int enc)
{
  int n, i;
  char **p = NULL, *b = NULL, *p1;

  RETURN_IF_NULL_PARAM(pSigInfo);
  //if(!enc && role && (strchr(role, '<') || strchr(role, '>')))
  //  SET_LAST_ERROR_RETURN(ERR_INVALID_CONTENT, ERR_INVALID_CONTENT);
  
  if(nCertified) {
    n = pSigInfo->signerRole.nCertifiedRoles + 1;
    p = (char**)malloc(n * sizeof(void*));
    RETURN_IF_BAD_ALLOC(p);
    if(pSigInfo->signerRole.nCertifiedRoles) {
      RETURN_IF_NULL(pSigInfo->signerRole.pCertifiedRoles);
      for(i = 0; i < pSigInfo->signerRole.nCertifiedRoles; i++)
	p[i] = pSigInfo->signerRole.pCertifiedRoles[i];
      free(pSigInfo->signerRole.pCertifiedRoles);
    }
    p[n-1] = 0;
    if(enc) {
      b = (char*)malloc(rLen * 2);
      RETURN_IF_BAD_ALLOC(b);
      i = sizeof(b);
      encode((const byte*)role, rLen, (byte*)b, &i);
      b[i] = 0;
      setString(&(p[n-1]), b, i);
    }
    else
      setString(&(p[n-1]), role, rLen);
    pSigInfo->signerRole.pCertifiedRoles = p;
    pSigInfo->signerRole.nCertifiedRoles = n;
  } else {
    n = pSigInfo->signerRole.nClaimedRoles + 1;
    p = (char**)malloc(n * sizeof(void*));
    RETURN_IF_BAD_ALLOC(p);
    if(pSigInfo->signerRole.nClaimedRoles) {
      RETURN_IF_NULL(pSigInfo->signerRole.pClaimedRoles);
      for(i = 0; i < pSigInfo->signerRole.nClaimedRoles; i++)
	p[i] = pSigInfo->signerRole.pClaimedRoles[i];
      free(pSigInfo->signerRole.pClaimedRoles);
    }
    p[n-1] = 0;
    if(enc) {
      b = (char*)malloc(rLen * 2);
      RETURN_IF_BAD_ALLOC(b);
      i = sizeof(b);
      encode((const byte*)role, rLen, (byte*)b, &i);
      b[i] = 0;
      setString(&(p[n-1]), b, i);
    }
    else {
      p1 = escape2xmlsym(role);
      if(p1) {
	setString(&(p[n-1]), p1, -1);
	free(p1);
      }
    }
    pSigInfo->signerRole.pClaimedRoles = p;
    pSigInfo->signerRole.nClaimedRoles = n;
  }
  return ERR_OK;
}

//============================================================
// Returns the number of signer roles
// pSigInfo - signature info object
// nCertified - certified role? (1/0)
//============================================================
EXP_OPTION int getCountOfSignerRoles(SignatureInfo* pSigInfo, int nCertified)
{
	RETURN_OBJ_IF_NULL(pSigInfo, -1);
	if(nCertified)
		return pSigInfo->signerRole.nCertifiedRoles;
	else
		return pSigInfo->signerRole.nClaimedRoles;
}

//============================================================
// Returns the desired signer role
// pSigInfo - signature info object
// nCertified - certified role? (1/0)
//============================================================
EXP_OPTION const char* getSignerRole(SignatureInfo* pSigInfo, int nCertified, int nIdx)
{
	RETURN_OBJ_IF_NULL(pSigInfo, 0);
	if(nCertified) {
		SET_LAST_ERROR_RETURN_IF_NOT(nIdx < pSigInfo->signerRole.nCertifiedRoles, ERR_BAD_ROLE_INDEX, 0);
		RETURN_OBJ_IF_NULL(pSigInfo->signerRole.pCertifiedRoles[nIdx], 0);
		return pSigInfo->signerRole.pCertifiedRoles[nIdx];
	} else {
		SET_LAST_ERROR_RETURN_IF_NOT(nIdx < pSigInfo->signerRole.nClaimedRoles, ERR_BAD_ROLE_INDEX, 0);
		RETURN_OBJ_IF_NULL(pSigInfo->signerRole.pClaimedRoles[nIdx], 0);
		return pSigInfo->signerRole.pClaimedRoles[nIdx];
	}
}


//============================================================
// Removes this SignatureInfo from signed doc and frees it's memory
// pSigDoc - signed doc object
// id - signature id to be removed
//============================================================
EXP_OPTION int SignatureInfo_delete(SignedDoc* pSigDoc, const char* id)
{
	int n, i, j, err = ERR_OK;
	SignatureInfo* pSignature;
	SignatureInfo** pSignatures;
	NotaryInfo* pNotInfo;

	RETURN_IF_NULL_PARAM(pSigDoc);
	// VS: it useless to generate an error here
	// just delete the corresponding notary
	pNotInfo = getNotaryWithSigId(pSigDoc, id);
	if(pNotInfo)
		NotaryInfo_delete(pSigDoc, pNotInfo->szId);
	if((pSignature = getSignatureWithId(pSigDoc, id)) != NULL) {
		n = pSigDoc->nSignatures - 1;
		if(n > 0) {
			pSignatures = (SignatureInfo**)malloc(n * sizeof(void*));
			RETURN_IF_BAD_ALLOC(pSignatures);
			for(i = j = 0; i < pSigDoc->nSignatures; i++) {
				if(strcmp(pSigDoc->pSignatures[i]->szId, id)) 
					pSignatures[j++] = pSigDoc->pSignatures[i];					
				else
					SignatureInfo_free(pSigDoc->pSignatures[i]);
			}
			free(pSigDoc->pSignatures);
			pSigDoc->pSignatures = pSignatures;
			pSigDoc->nSignatures = n;

		} else {
			for(i = 0; i < pSigDoc->nSignatures; i++)
				SignatureInfo_free(pSigDoc->pSignatures[i]);
			free(pSigDoc->pSignatures);
			pSigDoc->pSignatures = NULL;
			pSigDoc->nSignatures = 0;
		}
	}
	else
		err = ERR_BAD_SIG_INDEX;
	if (err != ERR_OK) SET_LAST_ERROR(err);
	return err;
}


//============================================================
// cleanup SignatureInfo memory
// pSigInfo - object to be cleaned up
//============================================================
EXP_OPTION void SignatureInfo_free(SignatureInfo* pSigInfo)
{
	int i;

	RETURN_VOID_IF_NULL(pSigInfo);
	if(pSigInfo->szId)
		free(pSigInfo->szId);
	if(pSigInfo->szTimeStamp)
		free(pSigInfo->szTimeStamp);
	if(pSigInfo->szSigPropDigest)
		free(pSigInfo->szSigPropDigest);
	if(pSigInfo->szSigPropRealDigest)
		free(pSigInfo->szSigPropRealDigest);
	if(pSigInfo->szSigInfoRealDigest)
		free(pSigInfo->szSigInfoRealDigest);
	if(pSigInfo->szSigType)
		free(pSigInfo->szSigType);
	if(pSigInfo->szSigValue)
		free(pSigInfo->szSigValue);
	if(pSigInfo->pX509Cert)
		X509_free((X509*)pSigInfo->pX509Cert);
	if(pSigInfo->szCertDigest)
		free(pSigInfo->szCertDigest);
	for(i = 0; i < pSigInfo->nDocs; i++) {
		DocInfo_free(pSigInfo->pDocs[i]);		
	}
	if(pSigInfo->pDocs)
		free(pSigInfo->pDocs);
	// signature production place
	if(pSigInfo->sigProdPlace.szCity)
		free(pSigInfo->sigProdPlace.szCity);
	if(pSigInfo->sigProdPlace.szStateOrProvince)
		free(pSigInfo->sigProdPlace.szStateOrProvince);
	if(pSigInfo->sigProdPlace.szPostalCode)
		free(pSigInfo->sigProdPlace.szPostalCode);
	if(pSigInfo->sigProdPlace.szCountryName)
		free(pSigInfo->sigProdPlace.szCountryName);
	// signer role
	for(i = 0; i < pSigInfo->signerRole.nClaimedRoles; i++) {
		if(pSigInfo->signerRole.pClaimedRoles[i])
		    free(pSigInfo->signerRole.pClaimedRoles[i]);    
	}
	if(pSigInfo->signerRole.pClaimedRoles)
		free(pSigInfo->signerRole.pClaimedRoles);
	for(i = 0; i < pSigInfo->signerRole.nCertifiedRoles; i++) {
		if (pSigInfo->signerRole.pCertifiedRoles[i])
			free(pSigInfo->signerRole.pCertifiedRoles[i]);    
	}
	if(pSigInfo->signerRole.pCertifiedRoles)
		free(pSigInfo->signerRole.pCertifiedRoles);
	if(pSigInfo->szOrigContent)
		free(pSigInfo->szOrigContent);
	if(pSigInfo->szIssuerSerial)
		free(pSigInfo->szIssuerSerial);
	free(pSigInfo);
}


//============================================================
// Adds a new DocInfo element to a SignatureInfo element and initializes it
// pSigInfo - signature info object
// docId - document id (use NULL for default)
// digType - digest type
// digest - documents digest
// digLen - digest length
// mime - mime type
// mimeDig - mime digest
// mimeDigLen - mime digest length
// return the newly created structure 
//============================================================
EXP_OPTION int addDocInfo(DocInfo **newDocInfo, SignatureInfo* pSigInfo, const char* docId,
			   const char* digType, const byte* digest,
			   int digLen, const byte* mimeDig, int mimeDigLen)
{
  DocInfo** pDocInfos = NULL;
  DocInfo* pDocInfo = NULL;
  int i;

  RETURN_IF_NULL_PARAM(pSigInfo);
  if(pSigInfo->nDocs == 0) {
    RETURN_IF_NOT(!pSigInfo->pDocs, ERR_BAD_DOCINFO_COUNT);
    pSigInfo->nDocs = 1;
  }
  else
    pSigInfo->nDocs++;
  pDocInfos = (DocInfo**)malloc((pSigInfo->nDocs) * sizeof(void *));
  RETURN_IF_BAD_ALLOC(pDocInfos);
  for(i = 0; i < pSigInfo->nDocs-1; i++)
    pDocInfos[i] = pSigInfo->pDocs[i];
  pDocInfo = (DocInfo*)malloc(sizeof(DocInfo)); // MEMLEAK: ???
  if (!pDocInfo) {
    free(pDocInfos);
    SET_LAST_ERROR_RETURN_CODE(ERR_BAD_ALLOC);
  }
  memset(pDocInfo, 0, sizeof(DocInfo));
  pDocInfos[pSigInfo->nDocs-1] = pDocInfo;
  pSigInfo->pDocs = pDocInfos;
  if(docId) 
    setString(&(pDocInfo->szDocId), docId, -1);
  if(digType)
    setString(&(pDocInfo->szDigestType), digType, -1);
  if(digest) {
    setString((char**)&(pDocInfo->szDigest), (const char*)digest, digLen);
    pDocInfo->nDigestLen = digLen;
  }
  if(mimeDig && strlen((const char*)mimeDig)) {
    setString((char**)&(pDocInfo->szMimeDigest), (const char*)mimeDig, mimeDigLen);
    pDocInfo->nMimeDigestLen = mimeDigLen;
  }
  *newDocInfo = pDocInfo;
  return ERR_OK;
}

//============================================================
// cleanup DocInfo memory
// pDocInfo - object to be cleaned up
//============================================================
EXP_OPTION void DocInfo_free(DocInfo* pDocInfo)
{
	RETURN_VOID_IF_NULL(pDocInfo);
	//assert(pDocInfo);
	if(pDocInfo->szDocId)
		free(pDocInfo->szDocId);
	if(pDocInfo->szDigestType)
		free(pDocInfo->szDigestType);
	if(pDocInfo->szDigest)
		free(pDocInfo->szDigest);
	if(pDocInfo->szMimeDigest)
		free(pDocInfo->szMimeDigest);
	// free the object itself
	free(pDocInfo);
}

//============================================================
// Returns number of DocInfos 
// pSigInfo - signature info pointer
//============================================================
EXP_OPTION int getCountOfDocInfos(const SignatureInfo* pSigInfo)
{
	RETURN_OBJ_IF_NULL(pSigInfo, -1);
	return pSigInfo->nDocs;
}

//============================================================
// Returns the desired DocInfo
// pSigInfo - signature info pointer
// idx - DocInfo index
//============================================================
EXP_OPTION DocInfo* getDocInfo(const SignatureInfo* pSigInfo, int idx)
{
  ddocDebug(3, "getDocInfo", "Idx: %d, Docs: %d", idx, pSigInfo->nDocs);
	RETURN_OBJ_IF_NULL(pSigInfo, NULL);
	SET_LAST_ERROR_RETURN_IF_NOT(idx < pSigInfo->nDocs, ERR_BAD_DOCINFO_INDEX, NULL);
	RETURN_OBJ_IF_NULL(pSigInfo->pDocs, NULL);
	RETURN_OBJ_IF_NULL(pSigInfo->pDocs[idx], NULL);
	return pSigInfo->pDocs[idx];
}

//============================================================
// Returns the last DocInfo
// pSigInfo - signature info pointer
// idx - DocInfo index
//============================================================
EXP_OPTION DocInfo* ddocGetLastDocInfo(const SignatureInfo* pSigInfo)
{
  int idx;

  RETURN_OBJ_IF_NULL(pSigInfo, NULL);
  idx = pSigInfo->nDocs - 1;
  SET_LAST_ERROR_RETURN_IF_NOT(idx < pSigInfo->nDocs, ERR_BAD_DOCINFO_INDEX, NULL);
  RETURN_OBJ_IF_NULL(pSigInfo->pDocs, NULL);
  RETURN_OBJ_IF_NULL(pSigInfo->pDocs[idx], NULL);
  return pSigInfo->pDocs[idx];
}


//============================================================
// Returns the DocInfo object with the given id
// pSigInfo - signature info pointer
// id - SignatureInfo id
//============================================================
EXP_OPTION DocInfo* getDocInfoWithId(const SignatureInfo* pSigInfo, const char* id)
{
	DocInfo* pDocInfo = NULL;
	int i;

	RETURN_OBJ_IF_NULL(pSigInfo, NULL);
	for(i = 0; i < pSigInfo->nDocs; i++) {
		RETURN_OBJ_IF_NULL(pSigInfo->pDocs[i], NULL);
		RETURN_OBJ_IF_NULL(pSigInfo->pDocs[i]->szDocId, NULL);
		if(!strcmp(pSigInfo->pDocs[i]->szDocId, id)) {
			pDocInfo = pSigInfo->pDocs[i];
			break;
		}
	}
	return pDocInfo;
}

//============================================================
// Sets the DocInfo objects document digest and digest type
// pDocInfo - document info pointer
// digest - digest value
// digLen - digest value length
// digType - digest type
//============================================================
EXP_OPTION void setDocInfoDigest(DocInfo* pDocInfo, const byte* digest, 
					  int digLen, const char* digType)
{
	RETURN_VOID_IF_NULL(pDocInfo);
	if(digType)
		setString(&(pDocInfo->szDigestType), digType, -1);
	if(digest) {
		setString((char**)&(pDocInfo->szDigest), (const char*)digest, digLen);
		pDocInfo->nDigestLen = digLen;
	}
}

//============================================================
// Sets the DocInfo objects mime digest and mime type
// pDocInfo - document info pointer
// mimeDig - mime digest value
// mimeDigLen - mime digest value length
//============================================================
EXP_OPTION void setDocInfoMimeDigest(DocInfo* pDocInfo, const byte* mimeDig, 
					  int mimeDigLen)
{
	RETURN_VOID_IF_NULL(pDocInfo);
  	if(mimeDig) {
		setString((char**)&(pDocInfo->szMimeDigest), (const char*)mimeDig, mimeDigLen);
		pDocInfo->nMimeDigestLen = mimeDigLen;
	}
}


//============================================================
// Adds all DocInfo elements in this file to a SignatureInfo element
// pSigInfo - signature info object
// pSigDoc - signed document
//============================================================
EXP_OPTION int addAllDocInfos(SignedDoc* pSigDoc, SignatureInfo* pSigInfo)
{
	int i, c;
	int len = 0;
	//Added by AA 28/10/2003 - not defined len value
	DataFile *pDf = NULL;
	DocInfo  *pDocInfo = NULL;
	byte buf[DIGEST_LEN+2];

	RETURN_IF_NULL_PARAM(pSigDoc);
	RETURN_IF_NULL_PARAM(pSigInfo);
	c = getCountOfDataFiles(pSigDoc);
	for(i = 0; i < c; i++ ) {
		pDf = getDataFile(pSigDoc, i);
		RETURN_IF_NULL(pDf);
		buf[0] = 0;
		// in version 1.0 we use mime digest
		if(!strcmp(pSigDoc->szFormatVer, SK_XML_1_VER)) {
			len = sizeof(buf);
			calculateDigest((const byte*)pDf->szMimeType, strlen(pDf->szMimeType), 
				DIGEST_SHA1, buf, &len);
		}
		// in 1.1 we don't use mime digest
		addDocInfo(&pDocInfo, pSigInfo, pDf->szId,
			   pDf->szDigestType, pDf->szDigest,
			   pDf->nDigestLen, buf, len);
	}
	return ERR_OK;
}

//============================================================
// Calculates and stores a signature for this SignatureInfo object
// pSigInfo - signature info object
// nSigType - signature type code
// keyfile - RSA key file
// passwd - key password
// certfile - certificate file
//============================================================
EXP_OPTION int calculateSigInfoSignature(const SignedDoc* pSigDoc, SignatureInfo* pSigInfo, int nSigType, 
			const char* keyfile, const char* passwd, const char* certfile)
{
	int err = ERR_OK;
	char buf2[SIGNATURE_LEN], *buf1 = NULL;
	int l2, l1;

	RETURN_IF_NULL_PARAM(pSigInfo);

	clearErrors();

  	if(nSigType == SIGNATURE_RSA) {
		setString(&(pSigInfo->szSigType), SIGN_RSA_NAME, -1);
		RETURN_IF_NULL(pSigInfo->szSigType);
		err = ReadCertificate((X509**)&(pSigInfo->pX509Cert), certfile);
		RETURN_IF_NOT(err == ERR_OK, ERR_CERT_READ);
		// Signed properties digest
		buf1 = createXMLSignedProperties(pSigDoc, pSigInfo, 0);
		RETURN_IF_NULL(buf1);
		l1 = strlen(buf1);
		l2 = sizeof(buf2);
		calculateDigest((const byte*)buf1, l1, DIGEST_SHA1, (byte*)buf2, &l2);
		free(buf1);
		setString((char**)&(pSigInfo->szSigPropDigest), buf2, l2);
		pSigInfo->nSigPropDigestLen = l2;		
		// SignedInfo digest
		buf1 = createXMLSignedInfo(pSigDoc, pSigInfo);
		RETURN_IF_NULL(buf1);
		l2 = sizeof(buf2);
		err = signData((const byte*)buf1, strlen(buf1), (byte*)buf2, &l2, DIGEST_SHA1,
				keyfile, passwd);
		free(buf1);
		setString((char**)&(pSigInfo->szSigValue), buf2, l2);
		pSigInfo->nSigLen = (short)l2;
	} else
		err = ERR_UNSUPPORTED_SIGNATURE;

	if (err != ERR_OK) SET_LAST_ERROR(err);
	return err;
}


//============================================================
// Decodes binary (DER) OCSP_RESPONSE data and returns a OCSP_RESPONSE object
// ppResp - pointer to a buffer to receive newly allocated OCSP_RESPONSE pointer
// data - (DER) OCSP_RESPONSE data
// len - length of data in bytes
//============================================================
EXP_OPTION int ddocDecodeOCSPResponseData(OCSP_RESPONSE **ppResp, const byte* data, int len)
{
  BIO* b1 = 0;

  // check input params
  RETURN_IF_NULL_PARAM(data);
  RETURN_IF_NULL_PARAM(ppResp);
  // mark as not read yet
  *ppResp = 0;
  // create BIO
  b1 = BIO_new_mem_buf((void*)data, len);
  RETURN_IF_NOT(b1, ERR_NULL_POINTER);
  // decode OCSP
  *ppResp = d2i_OCSP_RESPONSE_bio(b1, NULL);
  BIO_free(b1);
  ddocDebug(3, "ddocDecodeOCSPResponseData", "Decoding %d bytes DER data - OCSP_RESPONSE %s", len, (*ppResp ? "OK" : "ERROR"));
  RETURN_IF_NOT(*ppResp, ERR_OCSP_UNKNOWN_TYPE);
  return ERR_OK;
}

//============================================================
// Decodes base64 (PEM) OCSP_RESPONSE data and returns a OCSP_RESPONSE object
// ppResp - pointer to a buffer to receive newly allocated OCSP_RESPONSE pointer
// data - (PEM) OCSP_RESPONSE data
// len - length of data in bytes
//============================================================
EXP_OPTION int ddocDecodeOCSPResponsePEMData(OCSP_RESPONSE **ppResp, const byte* data, int len)
{
  byte* p1 = 0;
  int l1 = 0, err = ERR_OK;

  // check input params
  RETURN_IF_NULL_PARAM(data);
  RETURN_IF_NULL_PARAM(ppResp);
  // mark as not read yet
  *ppResp = 0;
  // allocate memory for decoding
  l1 = len; // should be enough as it shrinks
  p1 = (byte*)malloc(l1);
  RETURN_IF_BAD_ALLOC(p1);
  memset(p1, 0, l1);
  // decode base64 data
  decode((const byte*)data, len, p1, &l1);
  // decode OCSP
  err = ddocDecodeOCSPResponseData(ppResp, p1, l1);
  // cleanup
  if(p1)
	  free(p1);
  ddocDebug(3, "ddocDecodeOCSPResponsePEMData", "Decoding %d bytes PEM data - OSCP_RESPONSE %s", len, (*ppResp ? "OK" : "ERROR"));
  return err;
}


//============================================================
// Sets the signatures certificate and calculates
// certificate digest & serial number
// pSigInfo - signature info object
// cert - certficate 
//============================================================
EXP_OPTION int setSignatureCert(SignatureInfo* pSigInfo, X509* cert)
{
  int err = ERR_OK, l1;
  byte buf1[100];

  RETURN_IF_NULL_PARAM(pSigInfo);
  RETURN_IF_NULL_PARAM(cert);

  RETURN_IF_NOT(err == ERR_OK, err);
  l1 = sizeof(buf1);
  RETURN_IF_NOT(X509_digest(cert, EVP_sha1(), buf1, (unsigned int*)&l1), ERR_X509_DIGEST);
  if(pSigInfo->pX509Cert)
	  X509_free((X509*)pSigInfo->pX509Cert);
  pSigInfo->pX509Cert = cert;  
  setString((char**)&(pSigInfo->szCertDigest), (const char*)buf1, l1);
  pSigInfo->nCertDigestLen = l1;
  err = ReadCertSerialNumber((char*)buf1, sizeof(buf1), cert);
  setString((char**)&(pSigInfo->szIssuerSerial), (const char*)buf1, -1);
  return ERR_OK;
}

//============================================================
// Sets the signatures certificate and calculates
// certificate digest & serial number
// pSigInfo - signature info object
// certFile - certficate file in PEM
//============================================================
EXP_OPTION int setSignatureCertFile(SignatureInfo* pSigInfo, const char* certFile)
{
  X509 *cert = NULL;
  int err = ERR_OK, l1;
  byte buf1[100];

  RETURN_IF_NULL_PARAM(pSigInfo);
  RETURN_IF_NULL_PARAM(certFile);

  err = ReadCertificate(&cert, certFile);
  RETURN_IF_NOT(err == ERR_OK, err);
  if (err != ERR_OK) {
    X509_free(cert);
    SET_LAST_ERROR_RETURN_CODE(err);
  }
  l1 = sizeof(buf1);
  RETURN_IF_NOT(X509_digest(cert, EVP_sha1(), buf1, (unsigned int*)&l1), ERR_X509_DIGEST);

  if(pSigInfo->pX509Cert)
    X509_free((X509*)pSigInfo->pX509Cert);
  pSigInfo->pX509Cert = cert;
  setString((char**)&(pSigInfo->szCertDigest), (const char*)buf1, l1);
  pSigInfo->nCertDigestLen = l1;
  err = ReadCertSerialNumber((char*)buf1, sizeof(buf1), cert);
  setString((char**)&(pSigInfo->szIssuerSerial), (const char*)buf1, -1);
  return err;
}


//============================================================
// Calculates <SignedProperties> digest
// pSigInfo - signature info object
//============================================================
EXP_OPTION int calculateSignedPropertiesDigest(SignedDoc* pSigDoc, SignatureInfo* pSigInfo)
{
  int err = ERR_OK, l1, l2;
  byte buf1[DIGEST_LEN+2], *buf2 = 0;

  RETURN_IF_NULL_PARAM(pSigInfo);
  buf2 = (byte*)createXMLSignedProperties(pSigDoc, pSigInfo, 0);
  RETURN_IF_NULL(buf2);
  //dumpInFile("sigprop-sig1.txt", buf2);
  l2 = strlen((const char*)buf2);
  l1 = sizeof(buf1);
  ddocDebug(5, "calculateSignedPropertiesDigest", "Hash over: \n---\n%s\n---\n", buf1);
  calculateDigest(buf2, l2, DIGEST_SHA1, buf1, &l1);
  free(buf2);
  setString((char**)&(pSigInfo->szSigPropDigest), (const char*)buf1, l1);
  pSigInfo->nSigPropDigestLen = l1;
  return err;
}

//============================================================
// Calculates <SignedInfo> digest
// pSigInfo - signature info object
// digBuf - buffer for digest value
// digLen - address of buffer length. Must be initialized
// with buf max len, will be changed to actual length
//============================================================
EXP_OPTION int calculateSignedInfoDigest(SignedDoc* pSigDoc, SignatureInfo* pSigInfo, byte* digBuf, int* digLen)
{
  int err = ERR_OK, l2;
  byte *buf2 = NULL;
  FILE* hFile = 0;

  RETURN_IF_NULL_PARAM(pSigInfo);
  RETURN_IF_NULL_PARAM(digBuf);
  buf2 = (byte*)createXMLSignedInfo(pSigDoc, pSigInfo);
  RETURN_IF_NULL(buf2);
  l2 = strlen((const char*)buf2);
  calculateDigest(buf2, l2, DIGEST_SHA1, digBuf, digLen);
  free(buf2);
  return err;
}

//============================================================
// Sets the signature value from a file that contains
// the base64 encoded signature value
// pSigInfo - signature info object
// szSigFile - filename
//============================================================
EXP_OPTION int setSignatureValueFromFile(SignatureInfo* pSigInfo, char* szSigFile)
{
  int err = ERR_OK, i, j, slen;
  FILE* hFile = 0;
  byte buf[FILE_BUFSIZE], sbuf[300];
  EVP_ENCODE_CTX ectx;

  RETURN_IF_NULL_PARAM(pSigInfo);
  RETURN_IF_NULL_PARAM(szSigFile);

  if((hFile = fopen(szSigFile, "rb")) != NULL) {
    EVP_DecodeInit(&ectx);
    slen = 0;
    memset(sbuf, 0, sizeof(sbuf));
    for(;;) {
      i = fread(buf, sizeof(char), FILE_BUFSIZE, hFile);
      if(i <= 0) break;
      j = 0;
      EVP_DecodeUpdate(&ectx, sbuf+j, &j, (byte*)buf, i);
      slen += j;
    }
    EVP_DecodeFinal(&ectx, sbuf+j, &j); 
    slen += j;
    if(slen == SIGNATURE_LEN) {
      setString(&(pSigInfo->szSigType), SIGN_RSA_NAME, -1);
      setString((char**)&(pSigInfo->szSigValue), (const char*)sbuf, slen);
      pSigInfo->nSigLen = (short)slen;
      if(pSigInfo->szOrigContent) {
  	free(pSigInfo->szOrigContent);
  	((SignatureInfo*)pSigInfo)->szOrigContent = NULL;
      }
      pSigInfo->nOrigContentLen = 0;
    }
    else
      err = ERR_SIGNATURE;
    fclose(hFile);
  }
  else
    err = ERR_FILE_READ;
  if (err != ERR_OK) SET_LAST_ERROR(err);
  return err;
}

//============================================================
// Sets the signature value 
// pSigInfo - signature info object
// szSignature - signature value
// sigLen - signature length
//============================================================
EXP_OPTION int setSignatureValue(SignatureInfo* pSigInfo, byte* szSignature, int sigLen)
{
	RETURN_IF_NULL_PARAM(pSigInfo);
	RETURN_IF_NULL_PARAM(szSignature);
	
	//clearErrors();

	RETURN_IF_NOT(sigLen == SIGNATURE_LEN, ERR_SIGNATURE);
	setString(&(pSigInfo->szSigType), SIGN_RSA_NAME, -1);
	setString((char**)&(pSigInfo->szSigValue), (const char*)szSignature, sigLen);
	pSigInfo->nSigLen = (short)sigLen;
	if(pSigInfo->szOrigContent) {
	  free(pSigInfo->szOrigContent);
	  ((SignatureInfo*)pSigInfo)->szOrigContent = NULL;
	}
	pSigInfo->nOrigContentLen = 0;
	return ERR_OK;
}

//============================================================
// Compares two byte arrays and returns 0 for OK
// dig1 - byte array 1
// len1 - byte array 1 length
// dig2 - byte array 2
// len2 - byte array 2 length
//============================================================
EXP_OPTION int compareByteArrays(const byte* dig1, int len1, const byte* dig2, int len2)
{
	int err = ERR_OK, i;

	RETURN_IF_NULL_PARAM(dig1);
	RETURN_IF_NULL_PARAM(dig2);
	RETURN_IF_NOT(len1 == len2, ERR_COMPARE);	
	for(i = 0; i < len1; i++) {
		if(dig1[i] != dig2[i]) {
			err = ERR_COMPARE;
			break;
		}
	}
	return err;
}

//============================================================
// Verifies the digest of the given doc in this signature
// pSigDoc - signed doc data
// pSigInfo - signature info object
// filename - file name for not embedded files
// szDataFile - name of the digidoc file
//============================================================
// FIXME : Hard to understand the logic
EXP_OPTION int verifySigDocDigest(const SignedDoc* pSigDoc, const SignatureInfo* pSigInfo, 
				 const DocInfo* pDocInfo, const char* szFileName, const char* szDataFile)
{
  int err = ERR_OK;
  int l1, l2;
  long l;
  DataFile *pDf = NULL ;
  byte buf1[DIGEST_LEN+2], buf2[50], buf3[50], buf4[50];
  char *attNames = NULL, *attValues = NULL, *pTmp1 = NULL, *pTmp2 = NULL;
  //FILE *hFile;

  RETURN_IF_NULL_PARAM(pSigInfo);
  RETURN_IF_NULL_PARAM(pSigDoc);
  RETURN_IF_NULL_PARAM(pDocInfo);
  pDf = getDataFileWithId(pSigDoc, pDocInfo->szDocId);
  RETURN_IF_NULL(pDf);
  RETURN_IF_NULL(pDf->szContentType);
  RETURN_IF_NULL(pDf->szDigestType);
  RETURN_IF_NULL(pDocInfo->szDigestType);
  // verify detached file signature
  ddocDebug(3, "verifySigDocDigest", "SigDoc: %s DF: %s len1: %d len2: %d, ctype: %s", 
	    pSigDoc->szFormatVer, pDf->szId, pDf->nDigestLen, pDocInfo->nDigestLen, pDf->szContentType);
  if(!strcmp(pDf->szContentType, CONTENT_HASHCODE)) {
    ddocDebug(3, "verifySigDocDigest", "DF: %s ContentType: %s len1: %d len2: %d, type1: %s type2: %s", 
	      pDf->szId, pDf->szContentType, 
	      pDf->nDetachedDigestLen, pDocInfo->nDigestLen, 
	      pDocInfo->szDigestType, pDf->szDigestType);
    if(strcmp(pDocInfo->szDigestType, pDf->szDigestType))
      err = ERR_DOC_DIGEST;
    else
      err = compareByteArrays(pDocInfo->szDigest, pDocInfo->nDigestLen, 
			      pDf->szDetachedDigest, pDf->nDetachedDigestLen);
    if(err != ERR_OK)
      err = ERR_DOC_DIGEST;
  }
  if(!strcmp(pDf->szContentType, CONTENT_DETATCHED)) {
    if(szFileName) {
      l1 = sizeof(buf1);
      err = calculateFileDigest(szFileName, 
				DIGEST_SHA1, buf1, &l1, &l);
      RETURN_IF_NOT(err == ERR_OK, err);
      err = compareByteArrays(buf1, l1, 
			      pDf->szDetachedDigest, pDf->nDetachedDigestLen);
      RETURN_IF_NOT(err == ERR_OK, err);
    } else
      SET_LAST_ERROR_RETURN_CODE(ERR_DOC_DIGEST);
  }
  // the new digest calculation on the fly doesn't
  // work for old 1.0 files
  if(!strcmp(pDf->szContentType, CONTENT_EMBEDDED) &&
     !strcmp(pSigDoc->szFormatVer, SK_XML_1_VER)) {
    attNames = "Id";
    attValues = pDf->szId;
    err = readTagContents(&pTmp2, szDataFile, "DataFile", 1, 
			  (const char**)&attNames, (const char**)&attValues, 0);
    if(err == ERR_OK) {
      pTmp1 = pTmp2;
      //skip leading newlines
      while(*pTmp1 && *pTmp1 != '<') pTmp1++;
      /*sprintf(buf2, "/tmp/%s-data.txt", pDf->szId);
      if((hFile = fopen(buf2, "wb")) != NULL) {
	fputs(pTmp1, hFile);
	fclose(hFile);
	}*/
      l1 = sizeof(buf1);
      err = calculateDigest((const byte*)pTmp1, strlen(pTmp1),
			    DIGEST_SHA1, (byte*)buf1, &l1);	
      if(err == ERR_OK) {
	pDf->szDigest = 0;
	setString((char**)&(pDf->szDigest), (const char*)buf1, DIGEST_LEN);
	l2 = sizeof(buf3);
	encode((const byte*)pDf->szDigest, pDf->nDigestLen, (byte*)buf3, &l2);
	ddocDebug(3, "verifySigDocDigest", "DF: %s calculated digest: %s", 
	      pDf->szId, buf3);
	pDf->nDigestLen = DIGEST_LEN;
      }
      free(pTmp2);
    }
  }
  if(!strcmp(pDf->szContentType, CONTENT_EMBEDDED) ||
     !strcmp(pDf->szContentType, CONTENT_EMBEDDED_BASE64)) {
    buf2[0] = buf3[0] = buf4[0] = 0;
    l2 = sizeof(buf2);
    if(pDocInfo->szDigest)
      bin2hex((const byte*)pDocInfo->szDigest, pDocInfo->nDigestLen, (char*)buf2, &l2);
    l2 = sizeof(buf3);
    if(pDf->szDigest)
      bin2hex((const byte*)pDf->szDigest, pDf->nDigestLen, (char*)buf3, &l2);
    l2 = sizeof(buf4);
    if(pDf->szWrongDigest)
      bin2hex((const byte*)pDf->szWrongDigest, pDf->nWrongDigestLen, (char*)buf4, &l2);

    ddocDebug(3, "verifySigDocDigest", "DF: %s len1: %d len2: %d, type1: %s type2: %s, digest1: %s digest2: %s digest3: %s", 
	      pDf->szId, pDf->nDigestLen, pDocInfo->nDigestLen, pDocInfo->szDigestType, pDf->szDigestType, buf2, buf3, buf4);
    if(strcmp(pDocInfo->szDigestType, pDf->szDigestType))
      err = ERR_DOC_DIGEST;
    else
      err = compareByteArrays(pDocInfo->szDigest, pDocInfo->nDigestLen, 
			      pDf->szDigest, pDf->nDigestLen);
    if(err) { // check also the wrong digest
      err = compareByteArrays(pDocInfo->szDigest, pDocInfo->nDigestLen, 
			      pDf->szWrongDigest, pDf->nWrongDigestLen);
    }
    if(err != ERR_OK)
      err = ERR_DOC_DIGEST;
  }
  if (err != ERR_OK) SET_LAST_ERROR(err);
  ddocDebug(3, "verifySigDocDigest", "SigDoc DF: %s err: %d", 
	   pDf->szId, err);
  
  return err;
}

//============================================================
// Verifies the mime digest of the given doc in this signature
// pSigDoc - signed doc data
// pSigInfo - signature info object
// filename - file name for not embedded files
//============================================================
EXP_OPTION int verifySigDocMimeDigest(const SignedDoc* pSigDoc, const SignatureInfo* pSigInfo, 
				      const DocInfo* pDocInfo, const char* szFileName)
{
	int err = ERR_OK;
	int l1;
	DataFile* pDf;
	byte buf1[DIGEST_LEN+2];

	RETURN_IF_NULL_PARAM(pSigInfo);
	RETURN_IF_NULL_PARAM(pSigDoc);
	RETURN_IF_NULL_PARAM(pDocInfo);
	pDf = getDataFileWithId(pSigDoc, pDocInfo->szDocId);
	RETURN_IF_NULL(pDf);
	// we check mime digest only in ver 1.0
	if(!strcmp(pSigDoc->szFormatVer, SK_XML_1_VER)) {
		l1 = sizeof(buf1);
		err = calculateDigest((const byte*)pDf->szMimeType, strlen(pDf->szMimeType), 
				DIGEST_SHA1, buf1, &l1);
		RETURN_IF_NOT(err == ERR_OK, err);
		err = compareByteArrays(pDocInfo->szMimeDigest, pDocInfo->nMimeDigestLen, 
				buf1, l1);
		if(err != ERR_OK)
			err = ERR_MIME_DIGEST;
	}
	if (err != ERR_OK) SET_LAST_ERROR(err);
	return err;
}


//============================================================
// Verifies the SignedProperties digest
// pSigInfo - signature info object

// from original file and use it for hash function.
// This is usefull if the file has been generated by
// another library and possibly formats these elements
// differently.
//============================================================
// FIXME : pSigDoc and szDataFile not used at all.
EXP_OPTION int verifySigDocSigPropDigest(const SignedDoc* pSigDoc, const SignatureInfo* pSigInfo, const char* szDataFile)
{
	int err = ERR_OK;

	RETURN_IF_NULL_PARAM(pSigInfo);
	err = compareByteArrays(pSigInfo->szSigPropDigest, pSigInfo->nSigPropDigestLen, 
				pSigInfo->szSigPropRealDigest, pSigInfo->nSigPropRealDigestLen);	
	RETURN_IF_NOT(err == ERR_OK, ERR_SIGPROP_DIGEST);
	return ERR_OK;
}

//============================================================
// Verifies the certificates signed attributes
// pSigInfo - signature info object
//============================================================
EXP_OPTION int verifySigCert(const SignatureInfo* pSigInfo)
{
	int err = ERR_OK;
	int l1, l2;
	char szOtherSerial[100];
	byte buf1[DIGEST_LEN+2], buf2[DIGEST_LEN*2], buf3[DIGEST_LEN*2];

	RETURN_IF_NULL_PARAM(pSigInfo);
	RETURN_IF_NULL_PARAM(pSigInfo->pX509Cert);
	l1 = sizeof(buf1);
	RETURN_IF_NOT(X509_digest((const X509*)pSigInfo->pX509Cert, EVP_sha1(), buf1, (unsigned int*)&l1), ERR_X509_DIGEST); 
	// debug
	memset(buf2, 0, sizeof(buf2));
	memset(buf3, 0, sizeof(buf3));
	l2 = sizeof(buf2);
	encode((const byte*)pSigInfo->szCertDigest, pSigInfo->nCertDigestLen, (byte*)buf2, &l2);
	l2 = sizeof(buf3);
	encode((const byte*)buf1, l1, (byte*)buf3, &l2);

	ddocDebug(3, "verifySigCert", "SIG: %s cdig1: %d - %s cdig2: %d - %s", 
		  pSigInfo->szId, pSigInfo->nCertDigestLen, buf2, l1, buf3);
	err = compareByteArrays(pSigInfo->szCertDigest, pSigInfo->nCertDigestLen, 
				buf1, l1);	
	RETURN_IF_NOT(err == ERR_OK, ERR_WRONG_CERT);
	err = ReadCertSerialNumber(szOtherSerial, sizeof(szOtherSerial), (X509*)pSigInfo->pX509Cert);
	RETURN_IF_NOT(((err == ERR_OK) && !strcmp(pSigInfo->szIssuerSerial, szOtherSerial)), ERR_WRONG_CERT);
	return err;
}


//============================================================
// Verifies this signature
// pSigDoc - signed doc data
// pSigInfo - signature info object
// signerCA - direct signer CA certs filename
// szDateFile - name of the digidoc file
// bUseCA - use CA certs or not 1/0
// pPaths - file name / path pairs for detatched files
// nPaths - number of file name / path pairs
// from original file and use it for hash function.
// This is usefull if the file has been generated by
// another library and possibly formats these elements
// differently.
//============================================================
EXP_OPTION int verifySignatureInfo(const SignedDoc* pSigDoc, const SignatureInfo* pSigInfo, 
				   const char* signerCA, const char* szDataFile, int bUseCA,
				   FilePath* pPaths, int nPaths)
{
  int err = ERR_OK;
  int j, k, i;
  char *pFileName = NULL;
  X509* cert = NULL;
  DocInfo* pDocInfo = NULL;
  DataFile* pDf = NULL;


  RETURN_IF_NULL_PARAM(pSigInfo);
  clearErrors();
  err = verifyEstIDSignature((const byte*)pSigInfo->szSigInfoRealDigest, 
			     (int)pSigInfo->nSigInfoRealDigestLen, DIGEST_SHA1,
			     pSigInfo->szSigValue, pSigInfo->nSigLen, (X509*)pSigInfo->pX509Cert);
  if(err != ERR_OK)
    err = ERR_SIGNATURE;
  if(err == ERR_OK) {
    k = getCountOfDocInfos(pSigInfo);
    ddocDebug(4, "verifySignatureInfo", "DFs: %d", k);
    for(j = 0; (err == ERR_OK) && (j < k); j++) {
      pDocInfo = getDocInfo(pSigInfo, j);
      RETURN_IF_NULL(pDocInfo);
      ddocDebug(4, "verifySignatureInfo", "DocInfo: %s", pDocInfo->szDocId);      
      pDf = getDataFileWithId(pSigDoc, pDocInfo->szDocId);
      RETURN_IF_NULL(pDf);
      ddocDebug(4, "verifySignatureInfo", "DF: %s", pDf->szId);      
      // find data file path
      for(i = 0; i < nPaths; i++) {
	if(!strcmp(pDf->szFileName, pPaths->szFileName)) {
	  pFileName = pPaths->szPath;
	  break;
	}
      }
      err = verifySigDocDigest(pSigDoc, pSigInfo, pDocInfo, pFileName, szDataFile);
      ddocDebug(4, "verifySignatureInfo", "DF: %s verify: %d", pDf->szId, err);  
      if(err == ERR_OK)
	err = verifySigDocMimeDigest(pSigDoc, pSigInfo, pDocInfo, NULL);			
    }
  }
  if(err == ERR_OK)
    err = verifySigDocSigPropDigest(pSigDoc, pSigInfo, szDataFile);
  if(err == ERR_OK)
    err = verifySigCert(pSigInfo);
  if(err == ERR_OK) {
    cert = getSignCertData(pSigInfo);
    // VS: ver 2.2.4 - removed this check as OCSP check is sufficient
    //if(err == ERR_OK) 
    //  err = isCertValid(cert, convertStringToTimeT(pSigDoc, pSigInfo->szTimeStamp));		
    if(bUseCA && (err == ERR_OK))
      err = isCertSignedBy(cert, signerCA);
  }
  if ( err != ERR_OK) SET_LAST_ERROR(err);
  return err;
}


//============================================================
// Verifies the whole document, but returns on first error
// Use the functions defined earlier to verify all contents
// step by step.
// pSigDoc - signed doc data
// 
//============================================================
EXP_OPTION int verifySigDoc(const SignedDoc* pSigDoc, const char* signerCA, 
				 const char** caFiles, const char* caPath, const char* notCert, 
			     const char* szDataFile, int bUseCA, FilePath* pPaths, int nPaths)

{
	SignatureInfo* pSigInfo;
	NotaryInfo* pNotInfo;
	int i, d, err = ERR_OK;
	
	RETURN_IF_NULL_PARAM(pSigDoc);

	d = getCountOfSignatures(pSigDoc);
	for(i = 0; i < d; i++) {
		pSigInfo = getSignature(pSigDoc, i);
		err = verifySignatureInfo(pSigDoc, pSigInfo, signerCA, 
			szDataFile, bUseCA, pPaths, nPaths);
		RETURN_IF_NOT(err == ERR_OK, err);
	}
	d = getCountOfNotaryInfos(pSigDoc);
	for(i = 0; i < d; i++) {
		pNotInfo = getNotaryInfo(pSigDoc, i);
		err = verifyNotaryInfo(pSigDoc, pNotInfo, caFiles, caPath, notCert);
		RETURN_IF_NOT(err == ERR_OK, err);
	}
	return err;
}


//============================================================
// Reads in an OCSP Response file in DER format
// szFileName - OCSP response file name
//============================================================
int ReadOCSPResponse(OCSP_RESPONSE **newOCSP_RESPONSE, const char* szFileName)
{
  BIO *bin = NULL;
  OCSP_RESPONSE *resp = NULL;
  int err = ERR_OK;
	
  ddocDebug(4, "ReadOCSPResponse", "File: %s", szFileName);
  RETURN_IF_NULL_PARAM(newOCSP_RESPONSE);
  RETURN_IF_NULL_PARAM(szFileName);

  if((bin = BIO_new_file(szFileName, "rb")) != NULL) {
    ddocDebug(4, "ReadOCSPResponse", "File opened");
    resp = d2i_OCSP_RESPONSE_bio(bin, NULL);
    BIO_free(bin);
    if (resp == NULL) {
      err = ERR_OCSP_WRONG_VERSION;
    }
  } else {
    ddocDebug(4, "ReadOCSPResponse", "Cannot read file:%s", szFileName);
    err =ERR_FILE_READ;
  }
  if (err != ERR_OK) SET_LAST_ERROR(err);
  *newOCSP_RESPONSE = resp;
  return err;
}

//============================================================
// Writes an OCSP Response to a file in DER format
// szFileName - OCSP response file name
// resp - OCSP response object
//============================================================
int WriteOCSPResponse(const char* szFileName, const OCSP_RESPONSE* resp)
{
	BIO* bout = 0;

	RETURN_IF_NULL_PARAM(szFileName);
	RETURN_IF_NULL_PARAM(resp);
	if((bout = BIO_new_file(szFileName, "wb")) != NULL) {
#if OPENSSL_VERSION_NUMBER > 0x00908000
                ASN1_i2d_bio((int (*)(void*, unsigned char**))i2d_OCSP_RESPONSE, bout, (unsigned char*)resp);
#else
		ASN1_i2d_bio((int (*)())i2d_OCSP_RESPONSE, bout, (unsigned char*)resp);
#endif
		//i2d_OCSP_RESPONSE_bio((unsigned char*)bout, resp);
		BIO_free(bout);
	} else 
		SET_LAST_ERROR_RETURN_CODE(ERR_FILE_WRITE);
	return ERR_OK;	
}

//============================================================
// Reads in an OCSP Request file in DER format
// szFileName - OCSP Request file name
//============================================================
int ReadOCSPRequest(OCSP_REQUEST **newOCSP_REQUEST, const char* szFileName)
{
	BIO *bin = NULL;
        OCSP_REQUEST *req = NULL;
        int err = ERR_OK;

        RETURN_IF_NULL_PARAM(*newOCSP_REQUEST);
        RETURN_IF_NULL_PARAM(szFileName);

        if((bin = BIO_new_file(szFileName, "rb")) != NULL) {
                req = d2i_OCSP_REQUEST_bio(bin, NULL);
                BIO_free(bin);
                if (req == NULL) {
                        err = ERR_OCSP_WRONG_VERSION;
                }
        } else
                err =ERR_FILE_READ;

        if (err != ERR_OK) SET_LAST_ERROR(err);
        *newOCSP_REQUEST = req;
        return err;
}

//============================================================
// Writes an OCSP Request to a file in DER format
// szFileName - OCSP Request file name
// resp - OCSP Request object
//============================================================
int WriteOCSPRequest(const char* szFileName, const OCSP_REQUEST* req)
{
	BIO* bout = 0;
	
	if((bout = BIO_new_file(szFileName, "wb")) != NULL) {
#if OPENSSL_VERSION_NUMBER > 0x00908000
                ASN1_i2d_bio((int (*)(void*, unsigned char**))i2d_OCSP_RESPONSE, bout, (unsigned char*)req);
#else
                ASN1_i2d_bio((int (*)())i2d_OCSP_RESPONSE, bout, (unsigned char*)req);
#endif
		//i2d_OCSP_REQUEST_bio(bout, req);
		BIO_free(bout);
	} else
		SET_LAST_ERROR_RETURN_CODE(ERR_FILE_WRITE);
	return ERR_OK;
}


//============================================================
// Returns the number of notary infos
// pSigDoc - signed doc pointer
//============================================================
EXP_OPTION int getCountOfNotaryInfos(const SignedDoc* pSigDoc)
{
	RETURN_OBJ_IF_NULL(pSigDoc, -1);
	return pSigDoc->nNotaries;
}

//============================================================
// Returns the next free notary id
// pSigDoc - signed doc pointer
//============================================================
EXP_OPTION int getNextNotaryId(const SignedDoc* pSigDoc)
{
  int id = 0, n, i;

  RETURN_OBJ_IF_NULL(pSigDoc, -1);
  for(i = 0; i < pSigDoc->nNotaries; i++) {
    NotaryInfo* pNotary = pSigDoc->pNotaries[i];
    RETURN_OBJ_IF_NULL(pNotary, -1);
    RETURN_OBJ_IF_NULL(pNotary->szId, -1);
    SET_LAST_ERROR_RETURN_IF_NOT(strlen(pNotary->szId) > 1, ERR_EMPTY_STRING, -1);
    n = atoi(pNotary->szId+1);
    if(id <= n)
      id = n+1;
  }
  return id;
}

//============================================================
// Returns the desired NotaryInfo object
// pSigDoc - signed doc pointer
// nIdx - NotaryInfo index (starting with 0)
//============================================================
EXP_OPTION NotaryInfo* getNotaryInfo(const SignedDoc* pSigDoc, int nIdx)
{
  RETURN_OBJ_IF_NULL(pSigDoc, NULL);
  SET_LAST_ERROR_RETURN_IF_NOT(nIdx < pSigDoc->nNotaries, ERR_BAD_NOTARY_INDEX, NULL);
  RETURN_OBJ_IF_NULL(pSigDoc->pNotaries[nIdx], NULL);
  return pSigDoc->pNotaries[nIdx];
}

//============================================================
// Returns the last NotaryInfo object
// pSigDoc - signed doc pointer
// nIdx - NotaryInfo index (starting with 0)
//============================================================
EXP_OPTION NotaryInfo* ddocGetLastNotaryInfo(const SignedDoc* pSigDoc)
{
  int nIdx;
  RETURN_OBJ_IF_NULL(pSigDoc, NULL);
  nIdx = pSigDoc->nNotaries - 1;
  SET_LAST_ERROR_RETURN_IF_NOT(nIdx < pSigDoc->nNotaries, ERR_BAD_NOTARY_INDEX, NULL);
  RETURN_OBJ_IF_NULL(pSigDoc->pNotaries[nIdx], NULL);
  return pSigDoc->pNotaries[nIdx];
}

//============================================================
// Returns the NotaryInfo object with the given id
// pSigDoc - signed doc pointer
// id - NotaryInfo id
//============================================================
EXP_OPTION NotaryInfo* getNotaryWithId(const SignedDoc* pSigDoc, const char* id)
{
  NotaryInfo* pNotary = NULL;
  int i;
  
  RETURN_OBJ_IF_NULL(pSigDoc, NULL);
  for(i = 0; i < pSigDoc->nNotaries; i++) {
    RETURN_OBJ_IF_NULL(pSigDoc->pNotaries[i], NULL);
    RETURN_OBJ_IF_NULL(pSigDoc->pNotaries[i]->szId, NULL);
    if(!strcmp(pSigDoc->pNotaries[i]->szId, id)) {
      pNotary = pSigDoc->pNotaries[i];
      break;
    }
  }
  return pNotary;
}

//============================================================
// Returns the NotaryInfo object that corresponds to the given signature
// pSigDoc - signed doc pointer
// id - NotaryInfo id
//============================================================
NotaryInfo* getNotaryWithSigId(const SignedDoc* pSigDoc, const char* sigId)
{
  NotaryInfo* pNotary = NULL;
  int i;

  RETURN_OBJ_IF_NULL(pSigDoc, NULL);
  for(i = 0; i < pSigDoc->nNotaries; i++) {
    RETURN_OBJ_IF_NULL(pSigDoc->pNotaries[i], NULL);
    RETURN_OBJ_IF_NULL(pSigDoc->pNotaries[i]->szSigId, NULL);
    if(!strcmp(pSigDoc->pNotaries[i]->szSigId, sigId)) {
      pNotary = pSigDoc->pNotaries[i];
      break;
    }
  }
  return pNotary;
}

//============================================================
// Returns the NotaryInfo object that corresponds to the given signature
// ore creates a new one
// pSigDoc - signed doc pointer
// id - SignatureInfo id
//============================================================
NotaryInfo* getOrCreateNotaryWithSigId(SignedDoc* pSigDoc, const char* sigId)
{
  NotaryInfo* pNotary = 0;
  SignatureInfo* pSigInfo = 0;
  RETURN_OBJ_IF_NULL(pSigDoc, NULL);
  RETURN_OBJ_IF_NULL(sigId, NULL);
  pNotary = getNotaryWithSigId(pSigDoc, sigId);
  if(!pNotary) {
    pSigInfo = getSignatureWithId(pSigDoc, sigId);
    RETURN_OBJ_IF_NULL(pSigInfo, NULL);
    NotaryInfo_new(&pNotary, pSigDoc, pSigInfo);
  }
  return pNotary;
}



//============================================================
// Adds a new Notary element to a SignedDoc element and 
// initializes it partly
// pSigDoc - signed doc data
// pSigInfo - signature object to be verified by this notary
// return the newly created structure
//============================================================
EXP_OPTION int NotaryInfo_new(NotaryInfo **newNotaryInfo, SignedDoc* pSigDoc, const SignatureInfo* pSigInfo)
{
  int i, n;
  NotaryInfo** pNotaries = NULL;
  NotaryInfo* pNotary = NULL;
  char buf[10];

  RETURN_IF_NULL_PARAM(pSigDoc);
  RETURN_IF_NULL_PARAM(pSigInfo);

  if ((pSigDoc->nNotaries == 0) && pSigDoc->pNotaries) {
    SET_LAST_ERROR_RETURN_CODE(ERR_BAD_NOTARY_COUNT);
  };
  // get next notary id
  if (pSigInfo) {
    strcpy(buf, pSigInfo->szId);
    buf[0] = 'N';
    // make sure we reset the szOrigContent on this signature
    // otherwise the new confirmation will not be written to file!
    if(pSigInfo->szOrigContent) {
      free(pSigInfo->szOrigContent);
      ((SignatureInfo*)pSigInfo)->szOrigContent = NULL;
    }
  } else {
    n = getNextNotaryId(pSigDoc);
    RETURN_IF_NOT(n >= 0, ERR_BAD_NOTARY_ID);
    sprintf(buf, "N%d", n);
  }
  // memory management
  pNotaries = (NotaryInfo**)malloc((pSigDoc->nNotaries + 1) * sizeof(void *));
  RETURN_IF_BAD_ALLOC(pNotaries);
  for(i = 0; i < pSigDoc->nNotaries; i++)
    pNotaries[i] = pSigDoc->pNotaries[i];
  pNotary = (NotaryInfo*)malloc(sizeof(NotaryInfo));  // MEMLEAK:  ???
  if (!pNotary) {
    free(pNotaries);
    RETURN_IF_BAD_ALLOC(pNotary);
  }
  memset(pNotary, 0, sizeof(NotaryInfo));
  pNotaries[pSigDoc->nNotaries] = pNotary;
  if(pSigDoc->pNotaries)
    free(pSigDoc->pNotaries);
  pSigDoc->pNotaries = pNotaries;
  pSigDoc->nNotaries++;
  // set id	
  setString(&(pNotary->szId), buf, -1);
  // set target
  setString(&(pNotary->szSigId), pSigInfo->szId, -1);
  // version
  setString(&(pNotary->szNotType), SK_NOT_VERSION, -1);
  
  *newNotaryInfo = pNotary;
  return ERR_OK;
}

//============================================================
// cleanup NotaryInfo memory
// pNotary - object to be cleaned up
//============================================================
EXP_OPTION void NotaryInfo_free(NotaryInfo* pNotary)
{
	RETURN_VOID_IF_NULL(pNotary);
	//assert(pNotary);
	if(pNotary->szId)
		free(pNotary->szId);
	if(pNotary->szSigId)
		free(pNotary->szSigId);
	if(pNotary->szNotType)
		free(pNotary->szNotType);
	if(pNotary->timeProduced)
		free(pNotary->timeProduced);
	if(pNotary->szRespIdType)
		free(pNotary->szRespIdType);
	if(pNotary->szRespId)
		free(pNotary->szRespId);
	if(pNotary->thisUpdate)
		free(pNotary->thisUpdate);
	if(pNotary->nextUpdate)
		free(pNotary->nextUpdate);
	if(pNotary->szDigestType)
		free(pNotary->szDigestType);
	if(pNotary->szIssuerNameHash)
		free(pNotary->szIssuerNameHash);
	if(pNotary->szIssuerKeyHash)
		free(pNotary->szIssuerKeyHash);
	if(pNotary->szUserSign)
		free(pNotary->szUserSign);
	if(pNotary->szSigType)
		free(pNotary->szSigType);
	if(pNotary->szSigValue)
		free(pNotary->szSigValue);
	if(pNotary->pX509Cert)
		X509_free((X509*)pNotary->pX509Cert);
	if(pNotary->szCertDigest)
		free(pNotary->szCertDigest);
	if(pNotary->szOcspDigest)
		free(pNotary->szOcspDigest);
	if(pNotary->szIssuerSerial)
		free(pNotary->szIssuerSerial);

	// OCSP response
	if(pNotary->pOcspResponse)
		OCSP_RESPONSE_free((OCSP_RESPONSE*)pNotary->pOcspResponse);
	// free the object itself
	free(pNotary);
}

//============================================================
// Adds a certificate to Notary and initializes Notary
// pNotary - Notary info
// cert - responders certificate
// return error code
//============================================================
int addNotaryInfoCert(NotaryInfo *pNotary, X509 *cert)
{
	int err = ERR_OK, n;
	char buf[100];

	ddocDebug(3, "addNotaryInfoCert", "adding cert: %s to notary: %s", (cert ? "OK" : "NULL"), pNotary->szId);
	RETURN_IF_NULL_PARAM(cert);
	n = sizeof(buf);
	RETURN_IF_NOT(X509_digest(cert, EVP_sha1(), (unsigned char*)buf, (unsigned int*)&n), ERR_X509_DIGEST);
	RETURN_IF_NOT(err == ERR_OK, err);
	pNotary->pX509Cert = cert;
	setString((char**)&(pNotary->szCertDigest), buf, n);
	pNotary->nCertDigestLen = n;
	err = ReadCertSerialNumber(buf, sizeof(buf), cert);
	setString((char**)&(pNotary->szIssuerSerial), buf, -1);
	ddocDebug(3, "addNotaryInfoCert", "cert: %s to notary: %s", (pNotary->pX509Cert ? "OK" : "NULL"), pNotary->szId, err);
	return err;
}

//============================================================
// Checks OCSP certificate status and handles errors
// status - status code
// return error code
//============================================================
int handleOCSPCertStatus(int status)
{
  int err = ERR_OK;
  switch(status) {
  case V_OCSP_CERTSTATUS_GOOD: // cert is ok, do nothing
    break;
  case V_OCSP_CERTSTATUS_REVOKED: // cert has been revoked
    err = ERR_OCSP_CERT_REVOKED;
    break;
  case V_OCSP_CERTSTATUS_UNKNOWN: // cert status unknown
    err = ERR_OCSP_CERT_UNKNOWN;
    break;
  default: // should never happen?
    err = ERR_OCSP_RESP_STATUS;
  }
  return err;
}

int calcNotaryDigest(SignedDoc* pSigDoc, NotaryInfo* pNotary)
{
  int err = ERR_OK, l1;
  char buf1[25];

  RETURN_IF_NULL_PARAM(pNotary);
  RETURN_IF_NULL_PARAM(pSigDoc);
  l1 = sizeof(buf1);
  err = calculateNotaryInfoDigest(pSigDoc, pNotary, (byte*)buf1, &l1);
  //err = calculateOcspBasicResponseDigest(br, buf1, &l1);
  if(!err) {
    setString((char**)&(pNotary->szOcspDigest), buf1, l1); 
    pNotary->nOcspDigestLen = l1;
  }
  return err;
}

//============================================================
// Initializes NotaryInfo object with data from OCSP object
// pNotary - NotaryInfo object to be initialized
// resp - OCSP response object
// notCert - Notary cert object
// return error code
//============================================================
int initializeNotaryInfoWithOCSP(SignedDoc *pSigDoc, NotaryInfo *pNotary, 
				OCSP_RESPONSE *resp, X509 *notCert, int initDigest)
{
	int n, err = ERR_OK, l2;
	char buf[500];
	OCSP_RESPBYTES *rb = resp->responseBytes;
	OCSP_BASICRESP *br = NULL;
	OCSP_RESPDATA  *rd = NULL;
	OCSP_RESPID *rid = NULL;
	// OCSP_CERTSTATUS *cst = NULL;
	OCSP_SINGLERESP *single = NULL;
	OCSP_CERTID *cid = NULL;
	X509_EXTENSION *nonce;
	

	RETURN_IF_NULL_PARAM(pNotary);
	RETURN_IF_NULL_PARAM(resp);
	//RETURN_IF_NULL_PARAM(notCert);
	//assert(notCert);
	// check the OCSP Response validity
	switch(OCSP_response_status(resp)) {
	case OCSP_RESPONSE_STATUS_SUCCESSFUL: // OK
		break;
	case OCSP_RESPONSE_STATUS_MALFORMEDREQUEST:
		SET_LAST_ERROR_RETURN_CODE(ERR_OCSP_MALFORMED);
	case OCSP_RESPONSE_STATUS_INTERNALERROR:
		SET_LAST_ERROR_RETURN_CODE(ERR_OCSP_INTERNALERR);
	case OCSP_RESPONSE_STATUS_TRYLATER:
		SET_LAST_ERROR_RETURN_CODE(ERR_OCSP_TRYLATER);
	case OCSP_RESPONSE_STATUS_SIGREQUIRED:
		SET_LAST_ERROR_RETURN_CODE(ERR_OCSP_SIGREQUIRED);
	case OCSP_RESPONSE_STATUS_UNAUTHORIZED:
		SET_LAST_ERROR_RETURN_CODE(ERR_OCSP_UNAUTHORIZED);
	default:
		SET_LAST_ERROR_RETURN_CODE(ERR_OCSP_UNSUCCESSFUL);
	}
	if(OBJ_obj2nid(rb->responseType) != NID_id_pkix_OCSP_basic) 
		SET_LAST_ERROR_RETURN_CODE(ERR_OCSP_UNKNOWN_TYPE);
	if((br = OCSP_response_get1_basic(resp)) == NULL) 
		SET_LAST_ERROR_RETURN_CODE(ERR_OCSP_NO_BASIC_RESP);
	rd = br->tbsResponseData;
	if(ASN1_INTEGER_get(rd->version) != 0) 
		SET_LAST_ERROR_RETURN_CODE(ERR_OCSP_WRONG_VERSION);
	rid =  rd->responderId;
	if(rid->type != V_OCSP_RESPID_NAME && 
	   rid->type != V_OCSP_RESPID_KEY) 
		SET_LAST_ERROR_RETURN_CODE(ERR_OCSP_WRONG_RESPID);
	n = sk_OCSP_SINGLERESP_num(rd->responses);
	if(n != 1) 
		SET_LAST_ERROR_RETURN_CODE(ERR_OCSP_ONE_RESPONSE);
	single = sk_OCSP_SINGLERESP_value(rd->responses, 0);
	RETURN_IF_NULL(single);
	cid = single->certId;
	RETURN_IF_NULL(cid);
	ddocDebug(4, "initializeNotaryInfoWithOCSP", "CertStatus-type: %d", single->certStatus->type);
	//printf("TYPE: %d\n", single->certStatus->type);
	if(single->certStatus->type != 0) {
	  ddocDebug(4, "initializeNotaryInfoWithOCSP", "errcode: %d", handleOCSPCertStatus(single->certStatus->type));
	  SET_LAST_ERROR_RETURN_CODE(handleOCSPCertStatus(single->certStatus->type));
	}
	//Removed 31.10.2003
	//if(single->singleExtensions) 
	//	SET_LAST_ERROR_RETURN_CODE(ERR_OCSP_NO_SINGLE_EXT);
	if(!rd->responseExtensions ||
		(sk_X509_EXTENSION_num(rd->responseExtensions) != 1) ||
		((nonce = sk_X509_EXTENSION_value(rd->responseExtensions, 0)) == NULL)) 
			SET_LAST_ERROR_RETURN_CODE(ERR_OCSP_NO_NONCE);
	i2t_ASN1_OBJECT(buf,sizeof(buf),nonce->object);
	if(strcmp(buf, OCSP_NONCE_NAME)) 
		SET_LAST_ERROR_RETURN_CODE(ERR_OCSP_NO_NONCE);
	// producedAt
	err = asn1time2str(pSigDoc, rd->producedAt, buf);
	setString(&(pNotary->timeProduced), buf, -1);
	n = sizeof(buf);
	if(rid->type == V_OCSP_RESPID_NAME) {
		setString(&(pNotary->szRespIdType), RESPID_NAME_VALUE, -1);
		X509_NAME_oneline(rid->value.byName,buf,n);
		setString(&(pNotary->szRespId), buf, -1);
		pNotary->nRespIdLen = strlen(buf);
	}
	if(rid->type == V_OCSP_RESPID_KEY) {
		setString(&(pNotary->szRespIdType), RESPID_KEY_VALUE, -1);
		setString(&(pNotary->szRespId), (const char*)rid->value.byKey->data, rid->value.byKey->length);
		pNotary->nRespIdLen = rid->value.byKey->length;
	}
	// signature time
	err = asn1time2str(pSigDoc, single->thisUpdate, buf);	
	setString(&(pNotary->thisUpdate), buf, -1);
	// next update time
	if(single->nextUpdate) {
		err = asn1time2str(pSigDoc, single->nextUpdate, buf);
		setString(&(pNotary->nextUpdate), buf, -1);
	}
	// serial number
	sprintf(buf, "%ld", ASN1_INTEGER_get(cid->serialNumber)); 
	setString(&(pNotary->szIssuerSerial), buf, -1);
	// digest type
	i2t_ASN1_OBJECT(buf,sizeof(buf),cid->hashAlgorithm->algorithm);
	setString(&(pNotary->szDigestType), buf, -1);
	// issuer name hash
	setString((char**)&(pNotary->szIssuerNameHash), 
		(const char*)cid->issuerNameHash->data, cid->issuerNameHash->length);
	pNotary->nIssuerNameHashLen = cid->issuerNameHash->length;
	// issuer key hash
	setString((char**)&(pNotary->szIssuerKeyHash), 
		(const char*)cid->issuerKeyHash->data, cid->issuerKeyHash->length);
	pNotary->nIssuerKeyHashLen = cid->issuerKeyHash->length;
	// nonce == customers Signature hash
	if(nonce->value->length > DIGEST_LEN)
	  setString((char**)&(pNotary->szUserSign), 
		    (const char*)(nonce->value->data + (nonce->value->length - DIGEST_LEN)), DIGEST_LEN);
	else
	  setString((char**)&(pNotary->szUserSign), 
		    (const char*)nonce->value->data, nonce->value->length);
	pNotary->nUserSignLen = DIGEST_LEN; // fixed after consulting with Urmo
	// signature algorithm
	i2t_ASN1_OBJECT(buf,sizeof(buf),br->signatureAlgorithm->algorithm);
	setString(&(pNotary->szSigType), buf, -1);
	// signature value
	setString((char**)&(pNotary->szSigValue), (const char*)br->signature->data, br->signature->length);
	pNotary->nSigLen = (short)br->signature->length;
	// notary cert
	if(notCert && !err)
		err = addNotaryInfoCert(pNotary, notCert);

	// save the response in memory
	pNotary->pOcspResponse = resp;

	// get the digest from original OCSP data
	if(initDigest && notCert) {
	  err = calcNotaryDigest(pSigDoc, pNotary);
	}
	if (err != ERR_OK) SET_LAST_ERROR(err);
	return err;
}


//============================================================
// Adds a new Notary SignedInfo element to a SignedDoc 
//   element and initializes it
// newNotaryInfo - newly created structure
// pSigDoc - signed doc data
// pSigInfo - signature object to be verified by this notary
// ocspRespFile - OCSP response file name
// notaryCertFile - Notary cert file name
// returns error code or ERR_OK if no error.
//============================================================
// FIXME : What to do if initializeNotaryInfoWithOCSP fails?
EXP_OPTION int NotaryInfo_new_file(NotaryInfo **newNotaryInfo, SignedDoc *pSigDoc,
					 const SignatureInfo *pSigInfo, const char *ocspRespFile,
					 const char *notaryCertFile)

{
	OCSP_RESPONSE* resp;
	X509* notCert;
	NotaryInfo* pNotInf = NULL;
	int err = ERR_OK;
	
	RETURN_IF_NULL_PARAM(pSigDoc);
	RETURN_IF_NULL_PARAM(pSigInfo);
	RETURN_IF_NULL_PARAM(ocspRespFile);
	RETURN_IF_NULL_PARAM(notaryCertFile);

	err = ReadOCSPResponse(&resp, ocspRespFile);
	RETURN_IF_NOT(err == ERR_OK, ERR_FILE_READ);
	err = ReadCertificate(&notCert, notaryCertFile);
	RETURN_IF_NOT(err == ERR_OK, ERR_CERT_READ);
	err = NotaryInfo_new(&pNotInf, pSigDoc, pSigInfo);
	RETURN_IF_NOT(err == ERR_OK, err);
	*newNotaryInfo = pNotInf;
	err = initializeNotaryInfoWithOCSP(pSigDoc, pNotInf, resp, notCert, 1);
	RETURN_IF_NOT(err == ERR_OK, err);
	
	return ERR_OK;
}




//--------------------------------------------------
// Helper function that escapes XML special chars
// src - input data
// srclen - length of input data. Use -1 for 0 terminated strings
// dest - address of output buffer. Caller is responsible for deallocating it!
// returns error code or ERR_OK
//--------------------------------------------------
int escapeXMLSymbols(const char* src, int srclen, char** dest)
{
  int j, i, n = srclen;
  char *p = 0;

  RETURN_IF_NULL_PARAM(src);
  RETURN_IF_NULL_PARAM(dest);
  *dest = 0;
  if(n < 0)
    n = strlen(src);
  if(n > 0 && ConfigItem_lookup_int("DEBUG_LEVEL", 1) >= 5) {
    p = (char*)malloc(n+1);
    if(p) {
      memset(p, 0, n+1);
      memcpy(p, src, n);
      ddocDebug(5, "escapeXMLSymbols", "src: \"%s\" len: %d", p, n);
      free(p);
    }
  }
  else
    ddocDebug(5, "escapeXMLSymbols", "src: \"%s\" len: %d", src, n);
  // count the amount of memory needed for conversion
  for(i = j = 0; i < n; i++) {
     switch(src[i]) {
     case '<': j += 4; break;
     case '>': j += 4; break;
     case '&': j += 5; break;
     case '\r': j += 5; break;
     case '\'': j += 6; break;
     case '\"': j += 6; break;
     default: j ++; break;
     }
  }
  // count the last terminator char
  j++;
  ddocDebug(5, "escapeXMLSymbols", "allocating: %d bytes", j);
  *dest = (char*)malloc(j);
  memset(*dest, 0, j);
  // now convert the data
  for(i = j = 0; i < n; i++) {
     switch(src[i]) {
     case '<': strcat(*dest, "&lt;"); j += 4; break;
     case '>': strcat(*dest, "&gt;"); j += 4; break;
     case '&': strcat(*dest, "&amp;"); j += 5; break;
     case '\r': strcat(*dest, "&#xD;"); j += 5; break;
     case '\'': strcat(*dest, "&apos;"); j += 6; break;
     case '\"': strcat(*dest, "&quot;"); j += 6; break;
     default: (*dest)[j] = src[i]; j++; break;
     }
  }
  ddocDebug(5, "escapeXMLSymbols", "Converted: \'%s\' len: %d", *dest, j);
  return ERR_OK;
}



//===================< XML generating function >=========================

const char* g_sdoc_hdr0 = "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n";
// VS: changed 16.02.2004
const char* g_sdoc_hdr1_1 = "<SignedDoc format=\"%s\" version=\"%s\">\n";
const char* g_sdoc_hdr1_2 = "<SignedDoc format=\"%s\" version=\"%s\" xmlns=\"http://www.sk.ee/DigiDoc/v1.3.0#\">\n";
const char* g_sdoc_hdr2_1 = "<DataFile ContentType=\"%s\" Filename=\"%s\" Id=\"%s\" MimeType=\"%s\" Size=\"%ld\" DigestType=\"%s\" DigestValue=\"%s\"";
const char* g_sdoc_hdr2_2 = " %s=\"%s\"";
const char* g_sdoc_hdr2_3 = ">";
const char* g_sdoc_hdr2_4 = "<DataFile ContentType=\"%s\" Filename=\"%s\" Id=\"%s\" MimeType=\"%s\" Size=\"%ld\"";
const char* g_sdoc_hdr2_5 = "<DataFile ContentType=\"%s\" Filename=\"%s\" Id=\"%s\" MimeType=\"%s\" Size=\"%ld\" DigestType=\"%s\" DigestValue=\"%s\"";
const char* g_sdoc_hdr2_6 = " xmlns=\"http://www.sk.ee/DigiDoc/v1.3.0#\"";

const char* g_sdoc_hdr3 = "</DataFile>";

const char* g_sdoc_hdr4 = "<Signature Id=\"%s\" xmlns=\"http://www.w3.org/2000/09/xmldsig#\">\n";
const char* g_sdoc_hdr8 = "<DigestValue>%s</DigestValue>\n";

const char* g_sdoc_hdr11 = "<SignatureValue Id=\"%s-SIG\">\n";
const char* g_sdoc_hdr12 = "</SignatureValue>\n";

const char* g_sdoc_hdr13_1 = "<KeyInfo><X509Data><X509Certificate Id=\"%s-CERT\">\n";


const char* g_sdoc_hdr13_2 = "<KeyInfo>\n";
const char* g_sdoc_hdr13_3 = "<KeyValue>\n<RSAKeyValue>\n";
const char* g_sdoc_hdr13_4 = "<Modulus>%s</Modulus>\n";
const char* g_sdoc_hdr13_5 = "<Exponent>%s</Exponent>\n";
const char* g_sdoc_hdr13_6 = "</RSAKeyValue>\n</KeyValue>\n";

const char* g_sdoc_hdr13_10 = "<X509Data><X509Certificate>\n";

const char* g_sdoc_hdr14 = "</X509Certificate></X509Data></KeyInfo>\n";

const char* g_sdoc_hdr15 = "</Signature>\n";

const char* g_sdoc_hdr22 = "<Object><QualifyingProperties>\n";

const char* g_sdoc_hdr221 = "<Object><QualifyingProperties xmlns=\"http://uri.etsi.org/01903/v1.1.1#\" Target=\"#%s\">\n";

const char* g_sdoc_hdr23 = "</QualifyingProperties></Object>\n";



const char* g_sdoc_hdr30 = "</SignedDoc>";




//============================================================
// Creates a timestamp string
// buf - output buffer
// returns number of output bytes written
//============================================================
int createTimestamp(const SignedDoc* pSigDoc, char* buf)
{
  time_t t;
  struct tm tm1;
  Timestamp *pTimestamp;
  int dmz=0;

  RETURN_OBJ_IF_NULL(buf, 0);
  _tzset();
  time(&t);
  // in version 1.0 we use format CCYY.MM.DDTHH:MM:SS-TZ
  if(pSigDoc && !strcmp(pSigDoc->szFormatVer, "1.0")) {
    ddocLocalTime(&t, &tm1, 1);
    if(_daylight != 0) {
      /*if(_timezone<0){*/
      dmz=(_timezone - (_daylight*3600))/ 3600;
      /*}else{
	dmz=(_timezone + (_daylight*3600))/ 3600;
	}*/
    }else{
      dmz = _timezone / 3600;
    }		
  } else { // in version 1.1 we use UTC time
    ddocLocalTime(&t, &tm1, 0);
  }
  (void)Timestamp_new(&pTimestamp, tm1.tm_year + 1900, tm1.tm_mon + 1, tm1.tm_mday,
		      tm1.tm_hour, tm1.tm_min, tm1.tm_sec, dmz);
  (void)convertTimestampToString(pSigDoc, pTimestamp, buf);
  Timestamp_free(pTimestamp);
  return strlen(buf);
}


//============================================================
// Canonicalizes XML
// source - input data
// len - input length
// returns a newly allocated buffer with canonicalized XML
// Caller must free() the result.
//============================================================
char* canonicalizeXML(char* source, int len)
{
  xmlDocPtr doc = NULL;
  xmlChar* pBuf = NULL;
  int rc, n;
  char* dest = NULL;
	
  ddocDebug(5, "canonicalizeXML", "Canonicalizing: %d bytes", len);
  if((doc = xmlParseMemory(source, len)) != NULL) {
    ddocDebug(5, "canonicalizeXML", "Canonicalizing parse: %s", (doc ? "OK" : "ERROR"));
    rc = xmlC14NDocDumpMemory(doc, NULL, 0, NULL, 0, &pBuf); 
    ddocDebug(5, "canonicalizeXML", "Canonicalizing RC: %d: BUF: %s", rc, (pBuf ? "OK" : "ERROR"));
    if(pBuf) {
      n = strlen((char*)pBuf);
      dest = (char*)malloc(n + 1);
      if(dest)
	strcpy(dest, (char*)pBuf);
      else
	SET_LAST_ERROR_IF_NOT(dest, ERR_BAD_ALLOC);
      xmlFree(pBuf);
    }
    xmlFreeDoc(doc);
  }
  return dest;
}



//============================================================
// Creates a <SignedProperties> XML block
// pSigDoc - signed document pointer
// pSigInfo - signature info data
// bWithEscapes - 1=escape xml smbols, 0=don't escape
// returns new <SignedProperties> node
//============================================================
char* createXMLSignedProperties(const SignedDoc* pSigDoc, const SignatureInfo* pSigInfo, int bWithEscapes)
{
  char buf1[300], *pRet = 0, *p1, *p2;
  int len1, i, err, l1;
  xmlNodePtr pSigProp, pSigSigProp, pSigCert, pN1, pN2, pN3;
  static xmlChar nl[] = "\n";
  xmlDocPtr doc;
  xmlChar* pBuf;
	
  // XML doc
  doc = xmlNewDoc((const xmlChar*)"1.0");
  doc->children = xmlNewDocNode(doc, NULL, (const xmlChar*)"", NULL);
  // <SignedProperties>
  pSigProp = xmlNewChild(doc->children, NULL, (const xmlChar*)"SignedProperties", nl);
  // Ver 1.76 - in format 1.3 xmlns atribute is not used
  // in ver 1.1 we need this namespace def
  if(!strcmp(pSigDoc->szFormatVer, DIGIDOC_XML_1_1_VER) ||
     !strcmp(pSigDoc->szFormatVer, DIGIDOC_XML_1_2_VER)) {
    xmlNewNs(pSigProp, (const xmlChar*)"http://www.w3.org/2000/09/xmldsig#", NULL);
  } else { // in 1.3 we use the correct etsi namespace
    xmlNewNs(pSigProp, (const xmlChar*)"http://uri.etsi.org/01903/v1.1.1#", NULL);
  }
  // in 1.0 we had this buggy URI
  if(!strcmp(pSigDoc->szFormatVer, SK_XML_1_VER))
    sprintf(buf1, "#%s-SignedProperties", pSigInfo->szId);
  else // current version is 1.1
    sprintf(buf1, "%s-SignedProperties", pSigInfo->szId);
  xmlSetProp(pSigProp, (const xmlChar*)"Id", (const xmlChar*)buf1);
  
  // Ver 1.76 - in format 1.3 Target atribute is not used
  if(strcmp(pSigDoc->szFormatVer, DIGIDOC_XML_1_3_VER)) {
    sprintf(buf1, "#%s", pSigInfo->szId);
    xmlSetProp(pSigProp, (const xmlChar*)"Target", (const xmlChar*)buf1);
  }
  // <SignedSignatureProperties>
  pSigSigProp = xmlNewChild(pSigProp, NULL, (const xmlChar*)"SignedSignatureProperties", nl);
  // <SigningTime>
  pN1 = xmlNewChild(pSigSigProp, NULL, (const xmlChar*)"SigningTime", (const xmlChar*)pSigInfo->szTimeStamp);
  xmlNodeAddContent(pSigSigProp, nl);
  // <SigningCertificate>
  pSigCert = xmlNewChild(pSigSigProp, NULL, (const xmlChar*)"SigningCertificate", nl);
  // <Cert>	
  pN1 = xmlNewChild(pSigCert, NULL, (const xmlChar*)"Cert", nl);
  // Ver 1.76 - in format 1.3 Id atribute is not used
  if(strcmp(pSigDoc->szFormatVer, DIGIDOC_XML_1_3_VER)) {
    // in 1.0 we had this buggy URI
    if(!strcmp(pSigDoc->szFormatVer, SK_XML_1_VER))
      sprintf(buf1, "#%s-CERTINFO", pSigInfo->szId);
    else
      sprintf(buf1, "%s-CERTINFO", pSigInfo->szId);
    xmlSetProp(pN1, (const xmlChar*)"Id", (const xmlChar*)buf1);
  }
  // <CertDigest>
  pN2 = xmlNewChild(pN1, NULL, (const xmlChar*)"CertDigest", nl);
  // <DigestMethod>
  pN3 = xmlNewChild(pN2, NULL, (const xmlChar*)"DigestMethod", nl);
  xmlNodeAddContent(pN2, nl);
  xmlSetProp(pN3, (const xmlChar*)"Algorithm", (const xmlChar*)"http://www.w3.org/2000/09/xmldsig#sha1");
  // <DigestValue>
  len1 = sizeof(buf1);
  encode(pSigInfo->szCertDigest, pSigInfo->nCertDigestLen, (byte*)buf1, &len1);
  pN3 = xmlNewChild(pN2, NULL, (const xmlChar*)"DigestValue", (const xmlChar*)buf1);
  xmlNodeAddContent(pN2, nl);
  // <IssuerSerial>
  xmlNodeAddContent(pN1, nl);
  // Ver 1.76 - in format 1.3 Id atribute is not used
  if(!strcmp(pSigDoc->szFormatVer, DIGIDOC_XML_1_3_VER)) {
    pN2 = xmlNewChild(pN1, NULL, (const xmlChar*)"IssuerSerial", nl);
    l1 = sizeof(buf1);
    memset(buf1,0,l1);
    X509_NAME_oneline(X509_get_issuer_name((X509*)pSigInfo->pX509Cert), buf1, l1);
    pN3 = xmlNewChild(pN2, NULL, (const xmlChar*)"X509IssuerName", (const xmlChar*)buf1);
    xmlSetProp(pN3, (const xmlChar*)"xmlns", (const xmlChar*)"http://www.w3.org/2000/09/xmldsig#");
    xmlNodeAddContent(pN2, nl);
    pN3 = xmlNewChild(pN2, NULL, (const xmlChar*)"X509SerialNumber", (const xmlChar*)pSigInfo->szIssuerSerial);
    xmlSetProp(pN3, (const xmlChar*)"xmlns", (const xmlChar*)"http://www.w3.org/2000/09/xmldsig#");
    xmlNodeAddContent(pN2, nl);
  } else {
    pN2 = xmlNewChild(pN1, NULL, (const xmlChar*)"IssuerSerial", (const xmlChar*)pSigInfo->szIssuerSerial);
  }
  // <SignaturePolicyIdentifier>
  xmlNodeAddContent(pSigSigProp, nl);
  pN1 = xmlNewChild(pSigSigProp, NULL, (const xmlChar*)"SignaturePolicyIdentifier", nl);
  // <SignaturePolicyImplied>
  pN2 = xmlNewChild(pN1, NULL, (const xmlChar*)"SignaturePolicyImplied", nl);
  xmlNodeAddContent(pN1, nl);
  // <SignatureProductionPlace>
  if(pSigInfo->sigProdPlace.szCity || pSigInfo->sigProdPlace.szStateOrProvince ||
     pSigInfo->sigProdPlace.szPostalCode || pSigInfo->sigProdPlace.szCountryName) {
    xmlNodeAddContent(pSigSigProp, nl);
    pN1 = xmlNewChild(pSigSigProp, NULL, (const xmlChar*)"SignatureProductionPlace", nl);
    if(pSigInfo->sigProdPlace.szCity) {
      if(bWithEscapes) {
	escapeXMLSymbols(pSigInfo->sigProdPlace.szCity, -1, &p2);
	pN2 = xmlNewChild(pN1, NULL, (const xmlChar*)"City", (const xmlChar*)p2);
	free(p2);
      }
      else
	pN2 = xmlNewChild(pN1, NULL, (const xmlChar*)"City", (const xmlChar*)pSigInfo->sigProdPlace.szCity);
      xmlNodeAddContent(pN1, nl);
    }
    if(pSigInfo->sigProdPlace.szStateOrProvince) {
      if(bWithEscapes) {
	escapeXMLSymbols(pSigInfo->sigProdPlace.szStateOrProvince, -1, &p2);
	pN2 = xmlNewChild(pN1, NULL, (const xmlChar*)"StateOrProvince", (const xmlChar*)p2);
	free(p2);
      }
      else
	pN2 = xmlNewChild(pN1, NULL, (const xmlChar*)"StateOrProvince", (const xmlChar*)pSigInfo->sigProdPlace.szStateOrProvince);
      xmlNodeAddContent(pN1, nl);
    }
    if(pSigInfo->sigProdPlace.szPostalCode) {
      if(bWithEscapes) {
	escapeXMLSymbols(pSigInfo->sigProdPlace.szPostalCode, -1, &p2);
	pN2 = xmlNewChild(pN1, NULL, (const xmlChar*)"PostalCode", (const xmlChar*)p2);
	free(p2);
      }
      else
	pN2 = xmlNewChild(pN1, NULL, (const xmlChar*)"PostalCode", (const xmlChar*)pSigInfo->sigProdPlace.szPostalCode);
      xmlNodeAddContent(pN1, nl);
    }
    if(pSigInfo->sigProdPlace.szCountryName) {
      if(bWithEscapes) {
	escapeXMLSymbols(pSigInfo->sigProdPlace.szCountryName, -1, &p2);
	pN2 = xmlNewChild(pN1, NULL, (const xmlChar*)"CountryName", (const xmlChar*)p2);
	free(p2);
      }
      else
	pN2 = xmlNewChild(pN1, NULL, (const xmlChar*)"CountryName", (const xmlChar*)pSigInfo->sigProdPlace.szCountryName);
      xmlNodeAddContent(pN1, nl);
    }
    xmlNodeAddContent(pSigSigProp, nl);
  }
  // <SignerRole>
  if(pSigInfo->signerRole.nClaimedRoles ||
     pSigInfo->signerRole.nCertifiedRoles) {
    pN1 = xmlNewChild(pSigSigProp, NULL, (const xmlChar*)"SignerRole", nl);
    // <ClaimedRoles>
    if(pSigInfo->signerRole.nClaimedRoles) {
      pN2 = xmlNewChild(pN1, NULL, (const xmlChar*)"ClaimedRoles", nl);
      for(i = 0; i < pSigInfo->signerRole.nClaimedRoles; i++) {
	if(bWithEscapes) {
	  escapeXMLSymbols(pSigInfo->signerRole.pClaimedRoles[i], -1, &p2);
	  ddocDebug(4, "createXMLSignedProperties", "file: %s --> %s", pSigInfo->signerRole.pClaimedRoles[i], p2);
	  pN3 = xmlNewChild(pN2, NULL, (const xmlChar*)"ClaimedRole", (const xmlChar*)p2);
	  free(p2);
	} else 
	  pN3 = xmlNewChild(pN2, NULL, (const xmlChar*)"ClaimedRole", (const xmlChar*)pSigInfo->signerRole.pClaimedRoles[i]);
	xmlNodeAddContent(pN2, nl);
      }
      xmlNodeAddContent(pN1, nl);
    }
    // <CertifiedRoles>
    if(pSigInfo->signerRole.nCertifiedRoles) {
      pN2 = xmlNewChild(pN1, NULL, (const xmlChar*)"CertifiedRoles", nl);
      for(i = 0; i < pSigInfo->signerRole.nClaimedRoles; i++) {
	escapeXMLSymbols(pSigInfo->signerRole.pCertifiedRoles[i], -1, &p2);
	pN3 = xmlNewChild(pN2, NULL, (const xmlChar*)"CertifiedRole", (const xmlChar*)p2);
	free(p2);
	xmlNodeAddContent(pN2, nl);
      }
      xmlNodeAddContent(pN1, nl);
    }
    xmlNodeAddContent(pSigSigProp, nl);
  }
  // <SignedDataObjectProperties>
  xmlNodeAddContent(pSigProp, nl);
  pN1 = xmlNewChild(pSigProp, NULL, (const xmlChar*)"SignedDataObjectProperties", nl);
  xmlNodeAddContent(pSigProp, nl);
  // convert to string
  pBuf = NULL;
  err = xmlC14NDocDumpMemory(doc, NULL, 0, NULL, 0, &pBuf);	
  p1 = strstr((const char*)pBuf, "<SignedProperties");
  if(p1) {
    p2 = strstr((const char*)pBuf, "</SignedProperties>");
    if(p2) {
      p2[strlen("</SignedProperties>")] = 0;
      pRet = strdup(p1);
    }
  }
  xmlFree(pBuf);
  xmlFreeDoc(doc);
  // return <SignedProperties> node
  return pRet;
}


//============================================================
// Creates a <SignedInfo> XML block for a signature
// This is the actual data to be signed
// pSigInfo - signature info data
// buf - output buffer
// returns number of output bytes written
//============================================================
EXP_OPTION char* createXMLSignedInfo(const SignedDoc* pSigDoc, const SignatureInfo* pSigInfo)
{
	char buf1[100], *pRet = 0, *p1, *p2;
	int i, err, l1;
	xmlNodePtr pnSigInfo, pN1, pN2, pN3;
	static xmlChar nl[] = "\n";
	xmlDocPtr doc;
	xmlChar* pBuf;
	//FILE* hFile;
	DataFile* pDF;

	RETURN_OBJ_IF_NULL(pSigInfo, 0);
	// XML doc
	doc = xmlNewDoc((const xmlChar*)"1.0");
	doc->children = xmlNewDocNode(doc, NULL, (const xmlChar*)"", NULL);
	// <SignedInfo>
	pnSigInfo = xmlNewChild(doc->children, NULL, (const xmlChar*)"SignedInfo", nl);
	// in ver 1.1 we need this namespace def
	xmlNewNs(pnSigInfo, (const xmlChar*)"http://www.w3.org/2000/09/xmldsig#", NULL);
	// <CanonicalizationMethod>
	pN1 = xmlNewChild(pnSigInfo, NULL, (const xmlChar*)"CanonicalizationMethod", nl);
	xmlSetProp(pN1, (const xmlChar*)"Algorithm", (const xmlChar*)"http://www.w3.org/TR/2001/REC-xml-c14n-20010315");
	xmlNodeAddContent(pnSigInfo, nl);
	// <SignatureMethod>
	pN1 = xmlNewChild(pnSigInfo, NULL, (const xmlChar*)"SignatureMethod", nl);
	xmlSetProp(pN1, (const xmlChar*)"Algorithm", (const xmlChar*)"http://www.w3.org/2000/09/xmldsig#rsa-sha1");
	xmlNodeAddContent(pnSigInfo, nl);
	// <Reference>
	for(i = 0; i < pSigInfo->nDocs; i++) {
		// documents digest
		pN1 = xmlNewChild(pnSigInfo, NULL, (const xmlChar*)"Reference", nl);
		sprintf(buf1, "#%s", pSigInfo->pDocs[i]->szDocId);
		xmlSetProp(pN1, (const xmlChar*)"URI", (const xmlChar*)buf1);
		pDF = getDataFileWithId(pSigDoc, pSigInfo->pDocs[i]->szDocId);
		// in ver 1.1 we use Transforms only for detached documents
		if((!strcmp(pSigDoc->szFormatVer, DIGIDOC_XML_1_1_VER) ||
			!strcmp(pSigDoc->szFormatVer, DIGIDOC_XML_1_2_VER)) &&
			pDF && !strcmp(pDF->szContentType, CONTENT_DETATCHED)) {
			pN2 = xmlNewChild(pN1, NULL, (const xmlChar*)"Transforms", NULL);
			pN3 = xmlNewChild(pN2, NULL, (const xmlChar*)"Transform", NULL);
			xmlSetProp(pN3, (const xmlChar*)"Algorithm", (const xmlChar*)"http://www.sk.ee/2002/10/digidoc#detatched-document-signature");
			xmlNodeAddContent(pN1, nl);
		}
		pN2 = xmlNewChild(pN1, NULL, (const xmlChar*)"DigestMethod", nl);
		xmlSetProp(pN2, (const xmlChar*)"Algorithm", (const xmlChar*)"http://www.w3.org/2000/09/xmldsig#sha1");
		xmlNodeAddContent(pN1, nl);
		// in ver 1.0 we use Transforms
		if(!strcmp(pSigDoc->szFormatVer, SK_XML_1_VER)) {
			pN2 = xmlNewChild(pN1, NULL, (const xmlChar*)"Transforms", NULL);
			pN3 = xmlNewChild(pN2, NULL, (const xmlChar*)"Transform", NULL);
			xmlSetProp(pN3, (const xmlChar*)"Algorithm", (const xmlChar*)"http://www.w3.org/2000/09/xmldsig#enveloped-signature");
			xmlNodeAddContent(pN1, nl);
		}
		l1 = sizeof(buf1);
		encode(pSigInfo->pDocs[i]->szDigest, 
			pSigInfo->pDocs[i]->nDigestLen, (byte*)buf1, &l1);
		pN2 = xmlNewChild(pN1, NULL, (const xmlChar*)"DigestValue", (const xmlChar*)buf1);
		xmlNodeAddContent(pN1, nl);
		xmlNodeAddContent(pnSigInfo, nl);
		// in ver 1.1 we don't use mime digest Reference blocks
		if(!strcmp(pSigDoc->szFormatVer, SK_XML_1_VER)) {
			// document mime type digest
			pN1 = xmlNewChild(pnSigInfo, NULL, (const xmlChar*)"Reference", nl);
			if(!strcmp(pSigDoc->szFormatVer, "1.0"))
				sprintf(buf1, "#%s-MIME", pSigInfo->pDocs[i]->szDocId);
			else // current version is 1.1
				sprintf(buf1, "#%s@MimeType", pSigInfo->pDocs[i]->szDocId);
			xmlSetProp(pN1, (const xmlChar*)"URI", (const xmlChar*)buf1);
			pN2 = xmlNewChild(pN1, NULL, (const xmlChar*)"DigestMethod", nl);
			xmlSetProp(pN2, (const xmlChar*)"Algorithm", (const xmlChar*)"http://www.w3.org/2000/09/xmldsig#sha1");
			xmlNodeAddContent(pN1, nl);
			pN2 = xmlNewChild(pN1, NULL, (const xmlChar*)"Transforms", NULL);
			pN3 = xmlNewChild(pN2, NULL, (const xmlChar*)"Transform", NULL);
			xmlSetProp(pN3, (const xmlChar*)"Algorithm", (const xmlChar*)"http://www.w3.org/2000/09/xmldsig#enveloped-signature");
			xmlNodeAddContent(pN1, nl);
			l1 = sizeof(buf1);
			encode(pSigInfo->pDocs[i]->szMimeDigest, 
				pSigInfo->pDocs[i]->nMimeDigestLen, (byte*)buf1, &l1);
			pN2 = xmlNewChild(pN1, NULL, (const xmlChar*)"DigestValue", (const xmlChar*)buf1);
			xmlNodeAddContent(pN1, nl);
		}
	}
	// signed properties digest
	pN1 = xmlNewChild(pnSigInfo, NULL, (const xmlChar*)"Reference", nl);
	sprintf(buf1, "#%s-SignedProperties", pSigInfo->szId);
	xmlSetProp(pN1, (const xmlChar*)"URI", (const xmlChar*)buf1);
	// in version 1.2 we ise the Type atribute
	// for References that contain <SignedProperties> digest
	if(!strcmp(pSigDoc->szFormatVer, DIGIDOC_XML_1_2_VER) ||
       !strcmp(pSigDoc->szFormatVer, DIGIDOC_XML_1_3_VER)) {
		xmlSetProp(pN1, (const xmlChar*)"Type", (const xmlChar*)"http://uri.etsi.org/01903/v1.1.1#SignedProperties");
	}

	pN2 = xmlNewChild(pN1, NULL, (const xmlChar*)"DigestMethod", nl);
	xmlSetProp(pN2, (const xmlChar*)"Algorithm", (const xmlChar*)"http://www.w3.org/2000/09/xmldsig#sha1");
	xmlNodeAddContent(pN1, nl);
	// in 1.0 we used transforms here
	if(!strcmp(pSigDoc->szFormatVer, SK_XML_1_VER)) {
		pN2 = xmlNewChild(pN1, NULL, (const xmlChar*)"Transforms", NULL);
		pN3 = xmlNewChild(pN2, NULL, (const xmlChar*)"Transform", NULL);
		xmlSetProp(pN3, (const xmlChar*)"Algorithm", (const xmlChar*)"http://www.w3.org/2000/09/xmldsig#enveloped-signature");
		xmlNodeAddContent(pN1, nl);
	}
	l1 = sizeof(buf1);
	encode(pSigInfo->szSigPropDigest, pSigInfo->nSigPropDigestLen, (byte*)buf1, &l1);
	pN2 = xmlNewChild(pN1, NULL, (const xmlChar*)"DigestValue", (const xmlChar*)buf1);
	xmlNodeAddContent(pN1, nl);
	xmlNodeAddContent(pnSigInfo, nl);
	// convert to string
	pBuf = NULL;
	err = xmlC14NDocDumpMemory(doc, NULL, 0, NULL, 0, &pBuf);	
	p1 = strstr((const char*)pBuf, "<SignedInfo");
	if(p1) {
		p2 = strstr((const char*)pBuf, "</SignedInfo>");
		if(p2) {
			p2[strlen("</SignedInfo>")] = 0;
			pRet = strdup(p1);
		}
	}
	xmlFree(pBuf);
	xmlFreeDoc(doc);
	return pRet;
}



const char* g_sdoc_hdr35_1 = "<UnsignedProperties Target=\"#%s\">\n<UnsignedSignatureProperties>";
const char* g_sdoc_hdr35_2 = "<UnsignedProperties>\n<UnsignedSignatureProperties>";
const char* g_sdoc_hdr36 = "</UnsignedSignatureProperties>\n</UnsignedProperties>";

const char* g_sdoc_hdr37_1 = "<CertRefs>";
const char* g_sdoc_hdr37_2 = "</CertRefs>";

const char* g_sdoc_hdr37 = "<CompleteCertificateRefs>";
const char* g_sdoc_hdr38 = "</CompleteCertificateRefs>";
const char* g_sdoc_hdr39 = "<Cert Id=\"%s%s\">\n";
const char* g_sdoc_hdr39_1 = "<Cert>\n";
const char* g_sdoc_hdr40 = "</Cert>\n";
const char* g_sdoc_hdr41 = "<CertDigest>\n<DigestMethod Algorithm=\"http://www.w3.org/2000/09/xmldsig#sha1\"/>\n";
const char* g_sdoc_hdr42 = "</CertDigest>\n";
const char* g_sdoc_hdr43 = "<IssuerSerial>%s</IssuerSerial>\n";
//const char* g_sdoc_hdr43_1 = "";

const char* g_sdoc_hdr44 = "<CertificateValues>\n<EncapsulatedX509Certificate Id=\"%s%s\">\n";
const char* g_sdoc_hdr45 = "</EncapsulatedX509Certificate>\n</CertificateValues>\n";
const char* g_sdoc_hdr46 = "<RevocationValues>";
const char* g_sdoc_hdr47 = "</RevocationValues>";

const char* g_sdoc_hdr46_1 = "<OCSPValues>";
const char* g_sdoc_hdr46_2 = "</OCSPValues>";

static char ocsp_hdr31[] = "<EncapsulatedOCSPValue Id=\"%s\">\n";
static char ocsp_hdr32[] = "</EncapsulatedOCSPValue>\n";
static char ocsp_hdr33[] = "<CompleteRevocationRefs>\n<OCSPRefs>\n<OCSPRef>\n";
static char ocsp_hdr34[] = "</OCSPRef>\n</OCSPRefs>\n</CompleteRevocationRefs>\n";
static char ocsp_hdr35[] = "<OCSPIdentifier URI=\"%s%s\">";
static char ocsp_hdr36[] = "<ResponderID>%s</ResponderID>\n";
static char ocsp_hdr37[] = "<ProducedAt>%s</ProducedAt>\n";
static char ocsp_hdr38[] = "</OCSPIdentifier>\n";
static char ocsp_hdr39[] = "<DigestAlgAndValue>\n<DigestMethod Algorithm=\"http://www.w3.org/2000/09/xmldsig#sha1\"></DigestMethod>\n<DigestValue>%s</DigestValue>\n";
static char ocsp_hdr40[] = "</DigestAlgAndValue>";




//============================================================
// Calculates the digest of OCSP_RESPONSE
// pResp - OCSP_RESPONSE data 
// digBuf - signature buffer
// digLen - signature buffer length
// returns error code
//============================================================
int calculateOcspBasicResponseDigest(OCSP_BASICRESP* pBsResp, byte* digBuf, int* digLen)
{
	int err = ERR_OK, l1;
	byte *buf, *p;
	//BIO* b1;

	RETURN_IF_NULL_PARAM(pBsResp);
	l1 = i2d_OCSP_BASICRESP(pBsResp, NULL);
	buf = (byte*)malloc(l1+10);
	RETURN_IF_BAD_ALLOC(buf);
	p = buf;
	i2d_OCSP_BASICRESP(pBsResp, &p);
	err = calculateDigest(buf, l1, DIGEST_SHA1, digBuf, digLen);
	free(buf);
	if (err != ERR_OK) SET_LAST_ERROR(err);
	return err;
}

//============================================================
// Calculates the digest of NotaryInfo
// pSigDoc - signed document pointer
// pNotInfo - notary signature info 
// digBuf - signature buffer
// digLen - signature buffer length
// returns error code
//============================================================
EXP_OPTION int calculateNotaryInfoDigest(const SignedDoc* pSigDoc, 
	const NotaryInfo* pNotInfo, byte* digBuf, int* digLen)
{
	OCSP_BASICRESP* pBsResp;
	int err = ERR_OK, l1;
	byte *buf, *p;
	
	// in version 1.0 we calculated digest over tbsResponseData
	if(!strcmp(pSigDoc->szFormatVer, SK_XML_1_VER)) {
		err = notary2ocspBasResp(pSigDoc, pNotInfo, (X509*)pNotInfo->pX509Cert, &pBsResp);
		if(err == ERR_OK) {
			err = calculateOcspBasicResponseDigest(pBsResp, digBuf, digLen);
			OCSP_BASICRESP_free(pBsResp);
		}
	} else { // in version 1.1  we calculate digest over the whole OCSP_RESPONSE
		l1 = i2d_OCSP_RESPONSE((OCSP_RESPONSE*)pNotInfo->pOcspResponse, NULL);
		buf = (byte*)malloc(l1+10);
		RETURN_IF_BAD_ALLOC(buf);
		p = buf;
		i2d_OCSP_RESPONSE((OCSP_RESPONSE*)pNotInfo->pOcspResponse, &p);
		err = calculateDigest(buf, l1, DIGEST_SHA1, digBuf, digLen);
		free(buf);
	}
	if (err != ERR_OK) SET_LAST_ERROR(err);
	return err;
}


//============================================================
// Writes OCSP_RESPONSE to file without the usual PEM headers
// bout - output file
// pResp - OCSP_RESPONSE
// returns error code
//============================================================
int writeOcspToXMLFile(BIO* bout, OCSP_RESPONSE* pResp)
{
	int l1, l2;
	char *p1, *p2;

	RETURN_IF_NULL_PARAM(bout);
	RETURN_IF_NULL_PARAM(pResp);
	l1 = i2d_OCSP_RESPONSE(pResp, NULL);
	p1 = (char*)malloc(l1+10);
	RETURN_IF_BAD_ALLOC(p1);
	p2 = p1;
	i2d_OCSP_RESPONSE(pResp, (unsigned char**)&p2);
	l2 = l1 * 2;
	p2 = (char*)malloc(l2);
	if(p2 == NULL) {
		free(p1);
                RETURN_IF_BAD_ALLOC(p2);
	}
	encode((const byte*)p1, l1, (byte*)p2, &l2);
	BIO_puts(bout, p2);
	free(p2);
	free(p1);
	return ERR_OK;
}

//============================================================
// Converts OCSP_RESPONSE to PEM form with or without the headers
// pResp - OCSP_RESPONSE
// bHeaders - 1= with headers, 0=no headers
// buf - output buffer newly allocated
// returns error code
//============================================================
EXP_OPTION int getOcspPEM(OCSP_RESPONSE* pResp, int bHeaders, char** buf)
{
	int l1, l2;
	char *p1, *p2;

	RETURN_IF_NULL_PARAM(buf);
	RETURN_IF_NULL_PARAM(pResp);
	l1 = i2d_OCSP_RESPONSE(pResp, NULL);
	p1 = (char*)malloc(l1+10);
	RETURN_IF_BAD_ALLOC(p1);
	p2 = p1;
	i2d_OCSP_RESPONSE(pResp, (unsigned char**)&p2);
	l2 = l1 * 2;
	*buf = (char*)malloc(l2+200);
	if(*buf == NULL) {
		free(p1);
                RETURN_IF_BAD_ALLOC(*buf);
	}
	memset(*buf, 0, l2+200);
	if(bHeaders)
	  strcpy(*buf, "-----BEGIN OCSP RESPONSE-----\n");
	encode((const byte*)p1, l1, (byte*)strchr(*buf, 0), &l2);
	if(bHeaders)
	  strcat(*buf, "\n-----END OCSP RESPONSE-----");
	free(p1);
	return ERR_OK;
}

//============================================================
// Adds a notary signature to signed document file
// bout - output file
// pSigDoc - signed doc pointer
// pNotInfo - notary signature info 
// returns error code
//============================================================
int addNotaryInfoXML(BIO* bout, const SignedDoc *pSigDoc, const NotaryInfo* pNotInfo)
{
	int l1, err = ERR_OK, l2;
	byte buf1[500], buf2[DIGEST_LEN+2];
	OCSP_RESPONSE* pResp;

	RETURN_IF_NULL_PARAM(pNotInfo);
	// unsigned prop begin
	// Ver 1.76 - in format 1.3 we don't use the Target atribute
	if(!strcmp(pSigDoc->szFormatVer, DIGIDOC_XML_1_3_VER))
		BIO_puts(bout, g_sdoc_hdr35_2);
	else
		BIO_printf(bout, g_sdoc_hdr35_1, pNotInfo->szSigId);

	// cert refs begin
	BIO_puts(bout, g_sdoc_hdr37);
	// Ver 1.76 - in format 1.3 we use here additionally element <CertRefs>
	if(!strcmp(pSigDoc->szFormatVer, DIGIDOC_XML_1_3_VER))
		BIO_puts(bout, g_sdoc_hdr37_1);
	// cert info
	RETURN_IF_NULL(pNotInfo->pX509Cert);
	//assert(pNotInfo->pX509Cert);
	// Ver 1.76 - in format 1.3 we don't ise the Id atribute
	if(!strcmp(pSigDoc->szFormatVer, DIGIDOC_XML_1_3_VER))
		BIO_puts(bout, g_sdoc_hdr39_1);
	else
		BIO_printf(bout, g_sdoc_hdr39, pNotInfo->szSigId, "-RESPONDER_CERTINFO");
	BIO_puts(bout, g_sdoc_hdr41);
	// cert digest
	l1 = sizeof(buf1);
	encode(pNotInfo->szCertDigest, pNotInfo->nCertDigestLen, buf1, &l1);
	BIO_printf(bout, g_sdoc_hdr8, buf1); // cert digest
	BIO_puts(bout, g_sdoc_hdr42);
	// issuer serial
	// Ver 1.76 - in format 1.3 we use the <X509IssuerName> and <X509SerialNumber>
	if(!strcmp(pSigDoc->szFormatVer, DIGIDOC_XML_1_3_VER)) {
		l1 = sizeof(buf1);
		memset(buf1,0,l1);
		X509_NAME_oneline(X509_get_issuer_name((X509*)pNotInfo->pX509Cert), (char*)buf1, l1);
		BIO_puts(bout, "<IssuerSerial>\n<X509IssuerName xmlns=\"http://www.w3.org/2000/09/xmldsig#\">");
		BIO_puts(bout, (const char*)buf1);
		BIO_puts(bout, "</X509IssuerName>\n");
		BIO_printf(bout, "<X509SerialNumber xmlns=\"http://www.w3.org/2000/09/xmldsig#\">%s</X509SerialNumber>\n", pNotInfo->szIssuerSerial);
		BIO_puts(bout, "</IssuerSerial>\n");
	} else
		BIO_printf(bout, g_sdoc_hdr43, pNotInfo->szIssuerSerial);	
	BIO_puts(bout, g_sdoc_hdr40);
	// Ver 1.76 - in format 1.3 we use here additionally element <CertRefs>
	if(!strcmp(pSigDoc->szFormatVer, DIGIDOC_XML_1_3_VER))
		BIO_puts(bout, g_sdoc_hdr37_2);
	// cert refs end
	BIO_puts(bout, g_sdoc_hdr38);
	// CompleteRevocationRefs 
	BIO_puts(bout, ocsp_hdr33);
	BIO_printf(bout, ocsp_hdr35, "#", pNotInfo->szId);
	BIO_printf(bout, ocsp_hdr36, pNotInfo->szRespId);
	BIO_printf(bout, ocsp_hdr37, pNotInfo->timeProduced);
	BIO_puts(bout, ocsp_hdr38);
	l2 = sizeof(buf2);
	err = calculateNotaryInfoDigest(pSigDoc, pNotInfo, buf2, &l2);
	l1 = sizeof(buf1);
	encode(buf2, l2, buf1, &l1);
	BIO_printf(bout, ocsp_hdr39, buf1);
	BIO_puts(bout, ocsp_hdr40);
	// end of CompleteRevocationRefs
	BIO_puts(bout, ocsp_hdr34);

	// responder cert
	BIO_printf(bout, g_sdoc_hdr44, pNotInfo->szSigId, "-RESPONDER_CERT");
	// TODO format cert without header
	writeCertToXMLFile(bout, (X509*)pNotInfo->pX509Cert);
	//PEM_write_bio_X509(bout, pNotInfo->pX509Cert);
	BIO_puts(bout, g_sdoc_hdr45);
	// revocation values
	BIO_puts(bout, g_sdoc_hdr46);
	// Ver 1.76 - in format 1.3 we use here additionally element <OCSPValues>
	if(!strcmp(pSigDoc->szFormatVer, DIGIDOC_XML_1_3_VER))
		BIO_puts(bout, g_sdoc_hdr46_1);

	// OCSP response
	BIO_printf(bout, ocsp_hdr31, pNotInfo->szId);	
        // Removed by AA (Veiko's description)
	//if(!strcmp(pSigDoc->szFormatVer, DIGIDOC_XML_1_2_VER) ||
	//   !strcmp(pSigDoc->szFormatVer, DIGIDOC_XML_1_1_VER))
	pResp = (OCSP_RESPONSE*)pNotInfo->pOcspResponse;
	//else
	//	err = notary2ocspresp(pSigDoc, pNotInfo, (X509*)pNotInfo->pX509Cert, &pResp);
	if (err == ERR_OK)
		err = writeOcspToXMLFile(bout, pResp);
		//PEM_write_bio_OCSP_RESPONSE(bout, pResp);
	BIO_puts(bout, ocsp_hdr32);

	// Ver 1.76 - in format 1.3 we use here additionally element <OCSPValues>
	if(!strcmp(pSigDoc->szFormatVer, DIGIDOC_XML_1_3_VER))
		BIO_puts(bout, g_sdoc_hdr46_2);
	BIO_puts(bout, g_sdoc_hdr47);
	// unsigned prop end
	BIO_puts(bout, g_sdoc_hdr36);
	if (err != ERR_OK) SET_LAST_ERROR(err);
	return err;
}

//============================================================
// Swaps bytes in this byte array. Length must be an even number
// used for big-endian <--> small-endian conversion
//============================================================
void swapBytes(byte* src, int len)
{
	byte b;
	int i;

	if(len % 2 == 0) {
		for(i = 0; i < (len / 2); i++) {
			b = src[i];
			src[i] = src[len - i - 1];
			src[len - i - 1] = b;
		}
	}
}

//============================================================
// Adds a signature to signed document file
// bout - output file
// pSigInfo - signature info 
// cert - signers certificate
// returns error code
//============================================================
int addSignatureInfoXML(BIO* bout, SignedDoc* pSigDoc, SignatureInfo* pSigInfo)
{
	int err = ERR_OK;
	unsigned char buf2[300], *buf1 = 0;
	int len2, len1;
	NotaryInfo* pNotInfo = 0;
	EVP_PKEY* pubKey;
	
	RETURN_IF_NULL_PARAM(bout);
	RETURN_IF_NULL_PARAM(pSigDoc);
	RETURN_IF_NULL_PARAM(pSigInfo);
	
	BIO_printf(bout, g_sdoc_hdr4, pSigInfo->szId);
	buf1 = (unsigned char*)createXMLSignedInfo(pSigDoc, pSigInfo);
	RETURN_IF_NULL(buf1);
	BIO_puts(bout, (const char*)buf1);
	free(buf1);
	len1 = 200;
	buf1 = (unsigned char*)malloc(len1);
	RETURN_IF_BAD_ALLOC(buf1);
	memset(buf1, 0, len1);

	BIO_printf(bout, g_sdoc_hdr11, pSigInfo->szId);
	if(pSigInfo->nSigLen > 0){
		len2 = sizeof(buf2);
		encode(pSigInfo->szSigValue, pSigInfo->nSigLen, buf2, &len2);	
		BIO_puts(bout, (const char*)buf2);
	}
	BIO_puts(bout, g_sdoc_hdr12);
	// cert data...
	if(!strcmp(pSigDoc->szFormatVer, SK_XML_1_VER)) {
		BIO_printf(bout, g_sdoc_hdr13_1, pSigInfo->szId);	

	} else {
		// RSA KEY value
		BIO_puts(bout, g_sdoc_hdr13_2);
		BIO_puts(bout, g_sdoc_hdr13_3);
		err = GetPublicKey(&pubKey, (const X509*)pSigInfo->pX509Cert);
		// FIXME
		// modulus
		len1 = BN_bn2bin(pubKey->pkey.rsa->n, buf1);
		// in version 1.1 we output modulus as it is
		// starting from 1.2 we convert it to big-endian
		/*len2 = sizeof(buf2);
		memset(buf2, 0, len2);
		encode(buf1, len1, buf2, &len2);
		printf("Old modulus: %s\n", buf2);
		if(!strcmp(pSigDoc->szFormatVer, DIGIDOC_XML_1_1_VER)) {
			swapBytes((byte*)buf1, len1);
		}*/
		len2 = sizeof(buf2);
		memset(buf2, 0, len2);
		encode(buf1, len1, buf2, &len2);
		//printf("New modulus: %s len: %d\n", buf2, len1);
		BIO_printf(bout, g_sdoc_hdr13_4, buf2);
		// exponent
		memset(buf1, 0, len1);
		len1 = BN_bn2bin(pubKey->pkey.rsa->e, buf1);
		len2 = sizeof(buf2);
		memset(buf2, 0, len2);
		encode(buf1, len1, buf2, &len2);
		BIO_printf(bout, g_sdoc_hdr13_5, buf2);
		EVP_PKEY_free(pubKey);
		BIO_puts(bout, g_sdoc_hdr13_6);
		// cert data
		BIO_puts(bout, g_sdoc_hdr13_10);
	}
	free(buf1);

	err = writeCertToXMLFile(bout, (X509*)pSigInfo->pX509Cert);
	//PEM_write_bio_X509(bout, pSigInfo->pX509Cert);
	BIO_puts(bout, g_sdoc_hdr14);

	// VS in releases prior to 1.76 we had incorrect <QualifyingProperties> atributes
	if(!strcmp(pSigDoc->szFormatVer, DIGIDOC_XML_1_3_VER)) {
		sprintf((char*)buf2, g_sdoc_hdr221, pSigInfo->szId);
		BIO_puts(bout, (const char*)buf2);
	} else {
		BIO_puts(bout, g_sdoc_hdr22);
	}
	// signed properties	
	buf1 = (unsigned char*)createXMLSignedProperties(pSigDoc, pSigInfo, 1);
	//dumpInFile("sigprop-file1.txt", buf1);
	RETURN_IF_NULL(buf1);
	BIO_puts(bout, (const char*)buf1);
	free(buf1);
	// unsigned properties
	pNotInfo = getNotaryWithSigId(pSigDoc, pSigInfo->szId);
	if(pNotInfo)
		addNotaryInfoXML(bout, pSigDoc, pNotInfo);

	// qualifying properties end
	BIO_puts(bout, g_sdoc_hdr23);

	// signature end
	BIO_puts(bout, g_sdoc_hdr15);

	if (err != ERR_OK) SET_LAST_ERROR(err);	
	return err;
}


//--------------------------------------------------
// Checks if this file exists
// szFileName - file name
// returns 1 if exists
// AA 2004/05/20 debuggeri all viga, kui null pikkus
//--------------------------------------------------
EXP_OPTION int checkFileExists(const char* szFileName)
{
  FILE* hFile;
  int exists = 0, err;
  char buf1[300];

  err = ddocConvertFileName(buf1, sizeof(buf1), szFileName);
  if(err) return 0; // return doesn't exist if error
  if (strlen(buf1) > 0) {
    if((hFile = fopen(buf1, "r")) != NULL) {
      exists = 1;
      fclose(hFile);
    }
  }
  return exists;
}

//============================================================
// Generates DataFile elements XML form and stores it in a file
// pSigDoc - signed document
// pDataFile - data file object to be converted
// szDataFile - input file name
// hOutFile - output file handle
//============================================================
int generateDataFileXML(SignedDoc* pSigDoc, DataFile* pDataFile, 
			const char* szDataFile, BIO* bOutFile)
{
  int err = ERR_OK, len1, len2, j, k;
  char buf1[2048], buf2[4096], fixedFileName[200], *p;
  char *name, *value, *fName;
  BIO* bin;
  EVP_ENCODE_CTX ectx;
	
  RETURN_IF_NULL_PARAM(pSigDoc);
  RETURN_IF_NULL_PARAM(pDataFile);
  RETURN_IF_NULL_PARAM(szDataFile);
  RETURN_IF_NULL_PARAM(bOutFile);

  len1 = sizeof(buf1);
  // replaces '&' with '&amp;'
  memset(fixedFileName, 0, sizeof(fixedFileName));
  fName = (char*)getSimpleFileName(pDataFile->szFileName);
  for(k = j = 0; k < (int)strlen(fName); k++) {
    if(fName[k] == '&') {
      strcat(fixedFileName, "&amp;");
      j += 5;
    } else {
      fixedFileName[j] = fName[k];
      j++;
    }
  }
  //in versions 1.0, 1.1 and 1.2 we used bad encoding 
  if(!strcmp(pSigDoc->szFormatVer, SK_XML_1_VER) ||
	!strcmp(pSigDoc->szFormatVer, DIGIDOC_XML_1_1_VER) ||
	!strcmp(pSigDoc->szFormatVer, DIGIDOC_XML_1_2_VER)) {
#ifdef WIN32
	utf82oem(fixedFileName, buf1, sizeof(buf1));
	len2 = sizeof(buf2);
	ascii2utf8(buf1, buf2, &len2);
#else
	convFNameToWin(fixedFileName, buf2);
#endif
	strncpy(fixedFileName, buf2, sizeof(fixedFileName));
  } 
  // in version 1.0 we use DigestType and DigestValue attributes
  if(!strcmp(pSigDoc->szFormatVer, SK_XML_1_VER) ||
     !strcmp(pDataFile->szContentType, "HASHCODE")) {
    len2 = sizeof(buf2);
    encode((const byte*)pDataFile->szDigest, pDataFile->nDigestLen, (byte*)buf2, &len2);
    BIO_printf(bOutFile, g_sdoc_hdr2_1, pDataFile->szContentType, 			 
	       fixedFileName, pDataFile->szId, pDataFile->szMimeType, 
	       pDataFile->nSize, pDataFile->szDigestType, buf2);
  } else { // in version 1.2 we don't use those attributes except for detached files
    if(!strcmp(pDataFile->szContentType, CONTENT_DETATCHED)) {
      len2 = sizeof(buf2);
      encode(pDataFile->szDetachedDigest, pDataFile->nDetachedDigestLen, (byte*)buf2, &len2);
      BIO_printf(bOutFile, g_sdoc_hdr2_5, pDataFile->szContentType,
		 fixedFileName, pDataFile->szId, pDataFile->szMimeType, pDataFile->nSize,
		 pDataFile->szDigestType, buf2);
    } else
      BIO_printf(bOutFile, g_sdoc_hdr2_4, pDataFile->szContentType,
		 fixedFileName, pDataFile->szId, pDataFile->szMimeType, pDataFile->nSize);
  }
  k = getCountOfDataFileAttributes(pDataFile);
  for(j = 0; j < k; j++) {
    getDataFileAttribute(pDataFile, j, &name, &value);
    escapeXMLSymbols(value, -1, &p);
    BIO_printf(bOutFile, g_sdoc_hdr2_2, name, p);
    free(p);
  }
  // VS - ver 1.80 - in format 1.3 we started using SignedDoc schema
  if(!strcmp(pSigDoc->szFormatVer, DIGIDOC_XML_1_3_VER))
    BIO_puts(bOutFile, g_sdoc_hdr2_6);
  BIO_puts(bOutFile, g_sdoc_hdr2_3);
  err = ddocConvertFileName(fixedFileName, sizeof(fixedFileName), pDataFile->szFileName);
  if(err) return err;
  // if this is our temp file not a real input file
  // then don't change anything in it.
  if(strcmp(fixedFileName, szDataFile) != 0) {
    if((bin = BIO_new_file(szDataFile, "r")) != NULL) {
      while((len1 = BIO_read(bin, buf1, sizeof(buf1))) > 0) {
	BIO_write(bOutFile, buf1, len1);
      }
      BIO_free(bin);
    }
  } else {
    // if the file must be embedded
    if(!strcmp(pDataFile->szContentType, CONTENT_EMBEDDED_BASE64) ||
       !strcmp(pDataFile->szContentType, CONTENT_EMBEDDED)) {
	  //printf("Opening embedded file: %s\n", fixedFileName);
      if((bin = BIO_new_file(fixedFileName, "r")) != NULL) {
		  //printf("Opened\n");
	// if encoded
	if(!strcmp(pDataFile->szContentType, CONTENT_EMBEDDED_BASE64)) {
	  EVP_EncodeInit(&ectx);
	  while((len1 = BIO_read(bin, buf1, sizeof(buf1))) > 0) {
	    len2 = sizeof(buf2);
	    EVP_EncodeUpdate(&ectx, (unsigned char*)buf2, &len2, (unsigned char*)buf1, len1);
	    BIO_write(bOutFile, buf2, len2);
	  }
	  EVP_EncodeFinal(&ectx, (unsigned char*)buf2, &len2);
	  BIO_write(bOutFile, buf2, len2);
	} else
	  if(!strcmp(pDataFile->szContentType, CONTENT_EMBEDDED)) {
	    while((len1 = BIO_read(bin, buf1, sizeof(buf1))) > 0) {
	      if(!strcmp(pDataFile->szCharset, CHARSET_UTF_8)) {
		BIO_write(bOutFile, buf1, len1);
	      } else 
		if(!strcmp(pDataFile->szCharset, CHARSET_ISO_8859_1)) {
		  len2 = sizeof(buf2);
		  memset(buf2, 0, len2);
		  isolat1ToUTF8((unsigned char*)buf2, &len2, 
				(const unsigned char*)buf1, &len1);
		  BIO_write(bOutFile, buf2, len2);
		} else
		  SET_LAST_ERROR(ERR_UNSUPPORTED_CHARSET);
	    }
	    if(!strcmp(pDataFile->szCharset, CHARSET_UTF_8)) {
	      BIO_write(bOutFile, buf1, len1);
	    } else 
	      if(!strcmp(pDataFile->szCharset, CHARSET_ISO_8859_1)) {
		len2 = sizeof(buf2);
		memset(buf2, 0, len2);
		isolat1ToUTF8((unsigned char*)buf2, &len2, 
			      (const unsigned char*)buf1, &len1);
		BIO_write(bOutFile, buf2, len2);
	      } else
		SET_LAST_ERROR(ERR_UNSUPPORTED_CHARSET);			
	  }
	BIO_free(bin);
      } else
	err = ERR_FILE_READ;
    }
  } // not temp file
  // print suffix-whitespace
  //if(pDataFile->szDataSuffix)
  //	BIO_puts(bOutFile, pDataFile->szDataSuffix);
  // else for DETATCHED do nothing!
  BIO_puts(bOutFile, g_sdoc_hdr3);
  if (err != ERR_OK) SET_LAST_ERROR(err);	
  return err;
}

//--------------------------------------------------
// Creates a new signed XML document
// pSigDoc - signed doc info
// szSigDocFile - output XML file name. If the file exists,
// then it will be used to read in embedded DataFile contents.
// returns error code or ERR_OK for success
//--------------------------------------------------
int createSignedXMLDoc(SignedDoc* pSigDoc, const char* szOldFile, const char* szSigDocFile)
{
  int err = ERR_OK, i, nFiles;
  BIO *bout;
  DataFile* pDf = NULL;
  char ** arrTempFiles = NULL;
  char buf1[200];

  RETURN_IF_NULL_PARAM(pSigDoc);
  RETURN_IF_NULL_PARAM(szSigDocFile);
	
  nFiles = getCountOfDataFiles(pSigDoc);
  ddocDebug(3, "createSignedXMLDoc", "Old file: %s new file: %s, DFs: %d", 
	    szOldFile, szSigDocFile, nFiles);
  // check if the file exists allready 
  // and extracts data files. They
  // will be used later to construct
  // a new document and then removed
  if(szOldFile && checkFileExists(szOldFile)) {
    arrTempFiles = (char**)malloc(nFiles * sizeof(void*));
    RETURN_IF_BAD_ALLOC(arrTempFiles);
    memset(arrTempFiles, 0, nFiles * sizeof(void*));
    for(i = 0; i < nFiles; i++) {
      pDf = getDataFile(pSigDoc, i);
      ddocDebug(3, "createSignedXMLDoc", "DataFile: %s - %s", pDf->szId, pDf->szFileName);
      arrTempFiles[i] = (char*)malloc(200); //FIXME
      arrTempFiles[i][0] = 0;
      // do not copy newly added files
      if(!strchr((const char*)pDf->szFileName, '/') &&
	 !strchr((const char*)pDf->szFileName, '\\')) {
	err = getTempFileName(arrTempFiles[i]); //FIXME
	// VS: test the new parser based on xmlReader interface
	//err = ddocXRdrCopyDataFile(pSigDoc, szOldFile, (const char*)arrTempFiles[i], pDf->szId, CHARSET_ISO_8859_1, CHARSET_ISO_8859_1);
	ddocDebug(3, "createSignedXMLDoc", "Store DataFile: %s to: %s", pDf->szId, (const char*)arrTempFiles[i]);
	err = ddocSaxExtractDataFile(pSigDoc, szOldFile, (const char*)arrTempFiles[i], pDf->szId, "NO-CHANGE");
      }
    }
  }
  err = ddocConvertFileName(buf1, sizeof(buf1), szSigDocFile);
  if(err) return err;
  // now create the new document
  if((bout = BIO_new_file(buf1, "w")) != NULL) {
    BIO_puts(bout, g_sdoc_hdr0);
    // VS: ver 1.80 - in version 1.3 we started using SignedDoc namespace
    if(!strcmp(pSigDoc->szFormatVer, DIGIDOC_XML_1_3_VER))
      BIO_printf(bout, g_sdoc_hdr1_2, pSigDoc->szFormat, pSigDoc->szFormatVer);
    else
      BIO_printf(bout, g_sdoc_hdr1_1, pSigDoc->szFormat, pSigDoc->szFormatVer);
    // DataFile objects
    for(i = 0; i < nFiles; i++) {
      pDf = getDataFile(pSigDoc, i);
      // if the file must be embedded
      if(arrTempFiles && arrTempFiles[i] && checkFileExists(arrTempFiles[i])) {
	ddocDebug(3, "createSignedXMLDoc", "Use temp file: %s", arrTempFiles[i]);
	err = generateDataFileXML(pSigDoc, pDf, (const char*)arrTempFiles[i], bout);
      } else {
	err = ddocConvertFileName(buf1, sizeof(buf1), pDf->szFileName);
	if(err) return err;
	ddocDebug(3, "createSignedXMLDoc", "Create new data file: %s", buf1);
	err = generateDataFileXML(pSigDoc, pDf, (const char*)buf1, bout);
      }
      BIO_puts(bout, "\n");
    }
    for(i = 0; i < pSigDoc->nSignatures; i++) {
      // VS: if Signature has been read from file then 
      // use the original content
      if(pSigDoc->pSignatures[i]->szOrigContent)
	BIO_write(bout, pSigDoc->pSignatures[i]->szOrigContent, pSigDoc->pSignatures[i]->nOrigContentLen);
      else
	err = addSignatureInfoXML(bout, pSigDoc, pSigDoc->pSignatures[i]);
    }			
    BIO_puts(bout, g_sdoc_hdr30);		
    BIO_free(bout);
    // delete temporary files we created when
    // extracting the data files from original XML signed doc
    // VS: fix the bug of deleting input files
    // This happened because prefix was empty and
    // thus doc names where the same as original
    // input file names
    if(szOldFile && arrTempFiles) { // check if temp files were created
      for(i = 0; i < nFiles; i++) {
	pDf = getDataFile(pSigDoc, i);
	// ignore not being able to delete temp file. It returns -1
	_unlink((const char*)arrTempFiles[i]);
	free(arrTempFiles[i]);
      }
      free(arrTempFiles);
    }
  } else
    err = ERR_FILE_WRITE;

  if (err != ERR_OK) SET_LAST_ERROR(err);
  return err;
}


//--------------------------------------------------
// Creates a new signed document
// pSigDoc - signed doc info
// returns error code or ERR_OK for success
//--------------------------------------------------
EXP_OPTION int createSignedDoc(SignedDoc* pSigDoc, const char* szOldFile, const char* szOutputFile)
{
  int err = ERR_OK;

  RETURN_IF_NULL_PARAM(pSigDoc);
  RETURN_IF_NULL_PARAM(szOutputFile);
  RETURN_IF_NULL_PARAM(pSigDoc->szFormat);
	
  clearErrors();	
  if(!strcmp(pSigDoc->szFormat, SK_XML_1_NAME) ||
     !strcmp(pSigDoc->szFormat, DIGIDOC_XML_1_1_NAME)) { 
    err = createSignedXMLDoc(pSigDoc, szOldFile, szOutputFile);
  }
  else
    err = ERR_UNSUPPORTED_FORMAT;

  if (err != ERR_OK) SET_LAST_ERROR(err);
  return err;
}


//================< certficate info functions> =================================

//--------------------------------------------------
// Returns the user signatures certificate data
// pSignInfo - signature object
// returns cert data. This is actually X509*
//--------------------------------------------------
EXP_OPTION X509* getSignCertData(const SignatureInfo* pSignInfo)
{
	RETURN_OBJ_IF_NULL(pSignInfo, 0);
	//assert(pSignInfo);
	RETURN_OBJ_IF_NULL(pSignInfo->pX509Cert, 0);
	//assert(pSignInfo->pX509Cert);
	return (X509*)pSignInfo->pX509Cert;
}


//--------------------------------------------------
// Returns the notary signatures certificate data
// pNotInfo - notary signature object
// returns cert data. This is actually X509*
//--------------------------------------------------
EXP_OPTION X509* getNotCertData(const NotaryInfo* pNotInfo)
{
	RETURN_OBJ_IF_NULL(pNotInfo, 0);
	//assert(pNotInfo);
	RETURN_OBJ_IF_NULL(pNotInfo->pX509Cert, 0);
	//assert(pNotInfo->pX509Cert);
	return (X509*)pNotInfo->pX509Cert;
}



//--------------------------------------------------
// Returns the user signatures certificates purpose count
// pSignDoc - signed document data
// nSign - signature index
// nr - pointer to number buffer
// returns error code or ERR_OK
//--------------------------------------------------
/*int getSignCertPurposeCount(SignedDoc* pSignDoc, int nSign, int* nr)
{
	int err = ERR_OK;

	if(nSign < pSignDoc->nSignatures && 
		pSignDoc->pSignatures[nSign]->pX509Cert) {
		*nr = X509_PURPOSE_get_count();		
	}
	else
		err = ERR_BAD_INDEX;
	return err;
}*/


//--------------------------------------------------
// Returns the user signatures certificates purpose id and name
// pSignDoc - signed document data
// nSign - signature index
// nPurpose - purpose index
// id - pointer to id buffer
// name - pointer to name buffer
// returns error code or ERR_OK
//--------------------------------------------------
/*int getSignCertPurpose(SignedDoc* pSignDoc, int nSign, int nPurpose, 
					   int* id, char* name)
{
	int err = ERR_OK;
	X509_PURPOSE *pt;

	if(nSign < pSignDoc->nSignatures && 
		pSignDoc->pSignatures[nSign]->pX509Cert) {
		pt = X509_PURPOSE_get0(nPurpose);
		*nr = X509_PURPOSE_get_id(pt);	
		strcpy(name, X509_PURPOSE_get0_name(pt));
	}
	else
		err = ERR_BAD_INDEX;
	return err;
}*/






//================< OCSP functions> =================================




//--------------------------------------------------
// Sets digest algorithm type object
// type - SHA1
//--------------------------------------------------
X509_ALGOR* setCIDAlgorithm(const EVP_MD * type)
{
	X509_ALGOR* alg = NULL;
	int nid;

	alg = X509_ALGOR_new();
	RETURN_OBJ_IF_NULL(alg, 0);
	if((alg->parameter == NULL) || 
		(alg->parameter->type != V_ASN1_NULL)) {
		ASN1_TYPE_free(alg->parameter);
		alg->parameter=ASN1_TYPE_new();
		RETURN_OBJ_IF_NULL(alg->parameter, NULL);
		alg->parameter->type=V_ASN1_NULL;
	}
	ASN1_OBJECT_free(alg->algorithm);
	if ((nid = EVP_MD_type(type)) != NID_undef) {
		alg->algorithm=OBJ_nid2obj(nid);
	}
	return alg;
}

//--------------------------------------------------
// Sets signature algorithm type object
// type - RSA+SHA1
//--------------------------------------------------
X509_ALGOR* setSignAlgorithm(const EVP_MD * type)
{
	X509_ALGOR* alg;
	//int nid;

	alg = X509_ALGOR_new();
	RETURN_OBJ_IF_NULL(alg, NULL);
	if((alg->parameter == NULL) || 
		(alg->parameter->type != V_ASN1_NULL)) {
		ASN1_TYPE_free(alg->parameter);
		alg->parameter=ASN1_TYPE_new();
		RETURN_OBJ_IF_NULL(alg->parameter, 0);
		alg->parameter->type=V_ASN1_NULL;
	}
	ASN1_OBJECT_free(alg->algorithm);
	/*if ((nid = EVP_MD_type(type)) != NID_undef) {
		alg->algorithm=OBJ_nid2obj(nid);
	}*/
	alg->algorithm = OBJ_nid2obj(type->pkey_type);
	return alg;
}



//============================================================
// Verifies the certificates signed attributes
// pNotInfo - notary info object
//============================================================
EXP_OPTION int verifyNotCert(const NotaryInfo* pNotInfo)
{
	int err = ERR_OK;
	int l1;
	char szOtherSerial[100];
	byte buf1[DIGEST_LEN+2];

	RETURN_IF_NULL_PARAM(pNotInfo);

	l1 = sizeof(buf1);
	RETURN_IF_NOT(X509_digest((X509*)pNotInfo->pX509Cert, EVP_sha1(), buf1, (unsigned int*)&l1), ERR_X509_DIGEST); 
	err = compareByteArrays(pNotInfo->szCertDigest, pNotInfo->nCertDigestLen, buf1, l1);
	RETURN_IF_NOT(err == ERR_OK, ERR_WRONG_CERT);
	err = ReadCertSerialNumber(szOtherSerial, sizeof(szOtherSerial), (X509*)pNotInfo->pX509Cert);
	RETURN_IF_NOT(err == ERR_OK, err);
	RETURN_IF_NOT(!strcmp(pNotInfo->szIssuerSerial, szOtherSerial), ERR_WRONG_CERT);
	return ERR_OK;
}


//--------------------------------------------------
// Helper function. Converts Notary info to an OCSP
// response structure. Used in verify and file writing
// functions
// pNotInfo - NotaryInfo object
// notCert - OCSP responder certificate
// pBasResp - pointer to a pointer of the new response structure
//--------------------------------------------------
int notary2ocspBasResp(const SignedDoc* pSigDoc, const NotaryInfo* pNotInfo, X509* notCert, OCSP_BASICRESP** pBasResp)
{
  OCSP_SINGLERESP * single = 0;
  // ASN1_GENERALIZEDTIME *tp = NULL;
  int err = ERR_OK;

  RETURN_IF_NULL_PARAM(notCert);
  // new basic response
  *pBasResp = OCSP_BASICRESP_new();
  RETURN_IF_NULL(*pBasResp);
  str2asn1time(pSigDoc, pNotInfo->timeProduced, (*pBasResp)->tbsResponseData->producedAt); 
  if(!strcmp(pNotInfo->szRespIdType, RESPID_NAME_VALUE)) {
    (*pBasResp)->tbsResponseData->responderId->type = V_OCSP_RESPID_NAME;
    (*pBasResp)->tbsResponseData->responderId->value.byName = 
      X509_NAME_dup(X509_get_subject_name(notCert /*(X509*)pNotInfo->pX509Cert*/));
  } else {
    (*pBasResp)->tbsResponseData->responderId->type = V_OCSP_RESPID_KEY;
    (*pBasResp)->tbsResponseData->responderId->value.byKey = 
      ASN1_OCTET_STRING_new();
    ASN1_OCTET_STRING_set((*pBasResp)->tbsResponseData->responderId->value.byKey,
			  (unsigned char*)pNotInfo->szRespId, pNotInfo->nRespIdLen);
  }
  // new single response
  single = OCSP_SINGLERESP_new();
  single->certStatus->type = V_OCSP_CERTSTATUS_GOOD;
  single->certStatus->value.good = ASN1_NULL_new();
  single->certId->hashAlgorithm = setCIDAlgorithm(EVP_sha1());
  ASN1_OCTET_STRING_set(single->certId->issuerNameHash, 
			pNotInfo->szIssuerNameHash, pNotInfo->nIssuerNameHashLen);
  ASN1_OCTET_STRING_set(single->certId->issuerKeyHash, 
			pNotInfo->szIssuerKeyHash, pNotInfo->nIssuerKeyHashLen);

  ASN1_INTEGER_set(single->certId->serialNumber, atol(pNotInfo->szIssuerSerial));
  str2asn1time(pSigDoc, pNotInfo->thisUpdate, single->thisUpdate);
  if(pNotInfo->nextUpdate) 
    str2asn1time(pSigDoc, pNotInfo->nextUpdate, single->nextUpdate);
  sk_OCSP_SINGLERESP_push((*pBasResp)->tbsResponseData->responses, single);
  // add nonce
  err = OCSP_basic_add1_nonce((*pBasResp), pNotInfo->szUserSign, pNotInfo->nUserSignLen);
  if (err == ERR_LIB_NONE){
    err = ERR_OK;
    // set signature 
    (*pBasResp)->signatureAlgorithm = setSignAlgorithm(EVP_sha1()); 
    ASN1_OCTET_STRING_set((*pBasResp)->signature, pNotInfo->szSigValue, pNotInfo->nSigLen);
  } else {
    OCSP_BASICRESP_free(*pBasResp);
    SET_LAST_ERROR_RETURN_CODE(ERR_OCSP_NO_NONCE);
  }
  //	checkErrors();
  return ERR_OK;
}

//--------------------------------------------------
// Helper function. Converts Notary info to an OCSP
// response structure. Used in verify and file writing
// functions
// pNotInfo - NotaryInfo object
// notCert - OCSP responder certificate
// pResp - pointer to a pointer of the new response structure
//--------------------------------------------------
int notary2ocspresp(const SignedDoc* pSigDoc, const NotaryInfo* pNotInfo, X509* notCert, OCSP_RESPONSE** pResp)
{
	int err = ERR_OK;
	OCSP_BASICRESP* bs;

	err = notary2ocspBasResp(pSigDoc, pNotInfo, notCert, &bs);
	if(err == ERR_OK) {
		*pResp = OCSP_response_create(OCSP_RESPONSE_STATUS_SUCCESSFUL, bs);
		OCSP_BASICRESP_free(bs);
	} else 
		SET_LAST_ERROR_RETURN_CODE(err);
	
	return err;
}


//--------------------------------------------------
// Verfies NotaryInfo signature
// pSigDoc - signed doc object
// pNotInfo - NotaryInfo object
// caFiles - array of CA file names terminated with NULL
// CApath - path to (directory) all certs
// notCertFile - Notary (e.g. OCSP responder) cert file 
//--------------------------------------------------
EXP_OPTION int verifyNotaryInfo(const SignedDoc* pSigDoc, 
				const NotaryInfo* pNotInfo, 
				// VS - ver 1.66
				const char** caFiles, const char *CApath, 
				const char* notCertFile)
{
	X509** caCerts;
	X509* notCert = NULL;
	int err = ERR_OK, l1, i;
	
	RETURN_IF_NULL_PARAM(caFiles);
	RETURN_IF_NULL_PARAM(CApath);
	RETURN_IF_NULL_PARAM(notCertFile);

	// find the chain length
	// VS - ver 1.67
	for(l1 = 0; caFiles && caFiles[l1]; l1++);
	caCerts = (X509**)malloc(sizeof(void*) * (l1 + 1));
	RETURN_IF_BAD_ALLOC(caCerts);
	memset(caCerts, 0, sizeof(void*) * (l1 + 1));
	for(i = 0; i < l1; i++) {
		err = ReadCertificate(&(caCerts[i]),caFiles[i]);
		if (err != ERR_OK) {
			err = ERR_CERT_READ;
			goto cleanup;
		}
	}
	err = ReadCertificate(&notCert, notCertFile);
	if (err != ERR_OK) {
		err = ERR_CERT_READ;
		goto cleanup;
        }
	err = verifyNotaryInfoCERT(pSigDoc, pNotInfo,
			   (const X509**)caCerts, CApath, notCert);
	if (err != ERR_OK) SET_LAST_ERROR(err);
	// cleanup
cleanup:
	if(notCert)
		X509_free(notCert);
	for(i = 0; i < l1; i++)
		if(caCerts[i])   
			X509_free(caCerts[i]);
	free(caCerts);
	return err;
}

//--------------------------------------------------
// Setup X509 store for verification purposes
// CApath - directory of all certs
// CA1file - highest root cert
// CA2file - actual parent cert
//--------------------------------------------------
int setup_verifyCERT(X509_STORE **newX509_STORE,
				const char *CApath, 
			     	// VS - ver 1.66
				const X509** certs)
{
	X509_STORE *store;
	X509_LOOKUP *lookup;
	int i;

	if((store = X509_STORE_new()) == NULL) goto end;
	lookup = X509_STORE_add_lookup(store,X509_LOOKUP_file());
	if (lookup == NULL) goto end;
	// VS - ver 1.66
	for(i = 0; certs && certs[i]; i++) 
		X509_STORE_add_cert(store, (X509*)certs[i]);
	 
		
	lookup=X509_STORE_add_lookup(store,X509_LOOKUP_hash_dir());
	if (lookup == NULL) goto end;
	if (CApath) {
		if(!X509_LOOKUP_add_dir(lookup,CApath,X509_FILETYPE_PEM)) {
			//BIO_printf(bp, "Error loading directory %s\n", CApath);
			printf("Error loading directory %s\n", CApath);
			goto end;
		}
	} else X509_LOOKUP_add_dir(lookup,NULL,X509_FILETYPE_DEFAULT);

	*newX509_STORE = store;

	// VS - ver 1.67
	ERR_clear_error();
	return ERR_OK;
end:
	if (store) X509_STORE_free(store);
	SET_LAST_ERROR_RETURN_CODE(ERR_CERT_STORE_READ);
}

//--------------------------------------------------
// Verfies NotaryInfo signature
// pSigDoc - signed doc object
// pNotInfo - NotaryInfo object
// caCerts - CA certificate pointer array terminated with NULL
// CApath - path to (directory) all certs
// notCertFile - Notary (e.g. OCSP responder) cert file 
//--------------------------------------------------
EXP_OPTION int verifyNotaryInfoCERT(const SignedDoc* pSigDoc, 
				    const NotaryInfo* pNotInfo,
				    const X509** caCerts, const char *CApath, 
				    const X509* notCert)
{
	X509_STORE *store;
	OCSP_BASICRESP* bs;
	STACK_OF(X509)* ver_certs = NULL;
	int err = ERR_OK, l1;
	SignatureInfo* pSigInfo = NULL;
	byte buf[DIGEST_LEN+2];
	X509* certNotaryDirectCA = 0;

	RETURN_IF_NULL_PARAM(pSigDoc);
	RETURN_IF_NULL_PARAM(pNotInfo);
	RETURN_IF_NULL_PARAM(notCert);
	RETURN_IF_NULL_PARAM(caCerts);

	// find the chain length
	for(l1 = 0; caCerts && caCerts[l1]; l1++);
	if(l1 < 1)
	  SET_LAST_ERROR_RETURN_CODE(ERR_CERT_INVALID);
	certNotaryDirectCA = (X509*)caCerts[l1-1];	
	// do we have a signature that this notary confirms?
	pSigInfo = getSignatureWithId(pSigDoc, pNotInfo->szSigId);
	if(!pSigInfo)
		SET_LAST_ERROR_RETURN_CODE(ERR_NOTARY_NO_SIGNATURE);
	// do the signature values match?
	l1 = sizeof(buf);
	calculateDigest(pSigInfo->szSigValue, pSigInfo->nSigLen, DIGEST_SHA1, buf, &l1);
	if(compareByteArrays(buf, l1, pNotInfo->szUserSign, pNotInfo->nUserSignLen))
		SET_LAST_ERROR_RETURN_CODE(ERR_NOTARY_SIG_MATCH);

	// now create an OCSP object and check its validity
	// VS - ver 1.66
	// debug
	//WriteOCSPResponse("test2.resp", pNotInfo->pOcspResponse);
	if((setup_verifyCERT(&store, CApath, caCerts)) == ERR_OK) {
		// new basic response
		// create OCSP basic response
		// in version 1.0 we calculated digest over tbsResponseData
		bs = OCSP_response_get1_basic((OCSP_RESPONSE*)pNotInfo->pOcspResponse);
		if (!bs) err = ERR_OCSP_WRONG_RESPID;
		if (err == ERR_OK) {
			ver_certs = sk_X509_new_null();
			if (ver_certs) {
				sk_X509_push(ver_certs, notCert);
				
				// VS - ver 1.66
				err = OCSP_basic_verify(bs, ver_certs, store, OCSP_TRUSTOTHER); //OCSP_NOCHAIN);
				//if(err <= 0) 
				//	err = OCSP_basic_verify(bs, NULL, store, 0);
				if(err == ERR_LIB_NONE) {
					err = ERR_OK;
				} else {
				  //checkErrors();
				  err = ERR_OCSP_CERT_NOTFOUND;
				  //err = ERR_OCSP_WRONG_SIGNATURE; //ERR_OCSP_WRONG_RESPID;
				  SET_LAST_ERROR(err);
				}
				// cleanup
				sk_X509_free(ver_certs);
			}
			OCSP_BASICRESP_free(bs);
		}
		X509_STORE_free(store);
	} else {
	 	err = ERR_CERT_STORE_READ;
		SET_LAST_ERROR(err);
	}
	if(err == ERR_OK) {
		notCert = getNotCertData(pNotInfo);
		if(notCert) { // VS: ver 1.66
			err = isCertValid((X509*)notCert, convertStringToTimeT(pSigDoc, pNotInfo->timeProduced));
			if (err != ERR_OK)
                                SET_LAST_ERROR(err);
		} else {
			err = ERR_CERT_INVALID;
			SET_LAST_ERROR(err);
		}
		if(err == ERR_OK) {
			err = isCertSignedByCERT(notCert, certNotaryDirectCA);
			if (err != ERR_OK) 
				SET_LAST_ERROR(err);
		}
		if(err == ERR_OK) {
			err = verifyNotCert(pNotInfo);
			if (err != ERR_OK)
                                SET_LAST_ERROR(err);
		}
	}
	return err;
}


//--------------------------------------------------
// Verfies NotaryInfo digest
// pNotInfo - NotaryInfo object
//--------------------------------------------------
EXP_OPTION int verifyNotaryDigest(const SignedDoc* pSigDoc, const NotaryInfo* pNotInfo)
{
	int err, l1;
	byte buf1[DIGEST_LEN+2];

	l1 = sizeof(buf1);
	err = calculateNotaryInfoDigest(pSigDoc, pNotInfo, buf1, &l1);
	RETURN_IF_NOT(err == ERR_OK, err);
	err = compareByteArrays(buf1, l1, pNotInfo->szOcspDigest, pNotInfo->nOcspDigestLen);
	RETURN_IF_NOT(err == ERR_OK, err);
	return ERR_OK;
}


//--------------------------------------------------
// teeb 00:0a:df stiilis hexprinditud stringist tagasi tavalise
//--------------------------------------------------
// Tanel - ver 1.66
unsigned char *decodeHex(unsigned char *str)
{
	unsigned int i, j, k, len;
	unsigned char *ret;
	static unsigned char hex[] = { '0', '1', '2', '3', '4', '5', '6', '7', 
			'8', '9', 'A', 'B', 'C', 'D', 'E', 'F' };

	len = (int)(strlen((const char*)str) / 3) + 2;
	if((ret=(unsigned char*)malloc(len)) == NULL)
		return NULL;
	memset(ret, 0, len);
	for(i=0, j=0; i<strlen((const char*)str); i+=3) {
		for(k=0; k<16; k++)
			if(str[i] == hex[k])
				ret[j] = (unsigned char)(k<<4);
		for(k=0; k<16; k++)
			if(str[i+1] == hex[k])
			  ret[j++] += (unsigned char)k;
	}
	return(ret);
}


//--------------------------------------------------
// otsib X.509v3 laienduste seest Authority Key Identifieri vlja
//--------------------------------------------------
// Tanel - ver 1.66
unsigned char *get_authority_key(STACK_OF(X509_EXTENSION) *exts)
{
  int i, found=0;
  X509_EXTENSION *ex=0;
  ASN1_OBJECT *obj;
  X509V3_EXT_METHOD *met;
  void *st = NULL;
  unsigned char *p;
  STACK_OF(CONF_VALUE) *vals=NULL;
  CONF_VALUE *val;
  
  for(i=0; i<sk_X509_EXTENSION_num(exts); i++) {
    ex = sk_X509_EXTENSION_value(exts, i);
    obj = X509_EXTENSION_get_object(ex);
    if(OBJ_obj2nid(obj) == NID_authority_key_identifier) {
      found++;
      break;
    }
  }

  if(!found) {
    ddocDebug(4, "get_authority_key", "Extension not found");
    return(NULL);
  }
  
  met = X509V3_EXT_get(ex);
  p = ex->value->data;
#if OPENSSL_VERSION_NUMBER > 0x00908000
  st = ASN1_item_d2i(NULL, (const unsigned char**)&p, ex->value->length, ASN1_ITEM_ptr(met->it));
#else
  st = ASN1_item_d2i(NULL, &p, ex->value->length, ASN1_ITEM_ptr(met->it));
#endif
  vals = met->i2v(met, st, NULL);
  
  for(i=0; i<sk_CONF_VALUE_num(vals); i++) {
    val = sk_CONF_VALUE_value(vals, i);
    ddocDebug(4, "get_authority_key", "Extension %s - %s", val->name, val->value);
    if(val->name)
      if(strcmp(val->name, "keyid") == 0)
	return(decodeHex((unsigned char*)val->value));
  }
  
  return (NULL);
}



//--------------------------------------------------
// creates OCSP_CERTID without using the issuer cert
// cert - certificate for which we need certid
// returns OCSP_CERTID pointer
//--------------------------------------------------
// Tanel - ver 1.66
OCSP_CERTID* createOCSPCertid(X509 *cert, X509* pCACert)
{
  OCSP_CERTID *pId = NULL;
  X509_NAME *iname;
  unsigned char *ikey = NULL;
  ASN1_INTEGER *sno;
  const EVP_MD *dgst;
  X509_ALGOR *alg;
  unsigned char md[EVP_MAX_MD_SIZE];
  unsigned int len;

  if(cert != NULL) {
    // standard variant would be
    //pId = OCSP_cert_to_id(EVP_sha1(), cert, issuer);
    
    // issuer name hashi arvutamine 
    iname = X509_get_issuer_name(cert);
    dgst = EVP_sha1();
    if(X509_NAME_digest(iname, dgst, md, &len)) {
      // issuer key hashi lugemine
      ikey = get_authority_key(cert->cert_info->extensions);
      if(ikey != NULL) {
	// serial numbri lugemine
	sno = X509_get_serialNumber(cert);
	// OCSP certid koostamine
	if((pId = OCSP_CERTID_new()) != NULL) {
	  // replace default algorithm ???
	  alg = pId->hashAlgorithm;
	  if(alg->algorithm != NULL)
	    ASN1_OBJECT_free(alg->algorithm);
	  alg->algorithm = OBJ_nid2obj(EVP_MD_type(dgst));
	  if((alg->parameter = ASN1_TYPE_new()) != NULL) {
	    alg->parameter->type = V_ASN1_NULL;
	    ASN1_INTEGER_free(pId->serialNumber);
	    pId->serialNumber = ASN1_INTEGER_dup(sno);
	    if(!ASN1_OCTET_STRING_set(pId->issuerNameHash, md, len) ||
	       !ASN1_OCTET_STRING_set(pId->issuerKeyHash, ikey, strlen((const char*)ikey)) ||
	       !pId->serialNumber) {
	      fprintf(stderr, "Unable to fill in CID\n");
	      OCSP_CERTID_free(pId);
	      pId = NULL;
	    }
	  } // else - failed to create algorithm					
	}
	// cleanup ikey
	free(ikey);
      } // else failed to find Authority Key Identifier
      else {
	pId = OCSP_cert_to_id(NULL, cert, pCACert);

      }
    } // else - SHA1 failed	
  }
  return pId;
}

//--------------------------------------------------
// Helper function that handles OCSP nonce "the old way"
// e.g. by sending pure data instead of ASN.1 OCTET_STRING
// like in openssl 0.9.7d. This required to correctly 
// handle OCSP verification. This code was contributed
// by Sven Heiberg <sven@taru.cyber.ee>
//--------------------------------------------------
static int MY_ocsp_add1_nonce(STACK_OF(X509_EXTENSION) **exts, unsigned char *val, int len)
{  
  unsigned char *tmpval;  
  ASN1_OCTET_STRING os;  
  int ret = 0;  

  if (len <= 0) 
    len = OCSP_DEFAULT_NONCE_LENGTH;  
  if (val) 
    tmpval = val;  
  else {  
    tmpval = (unsigned char*)OPENSSL_malloc(len);
    if (!tmpval)
      goto err;    
    RAND_pseudo_bytes(tmpval, len);  
  }  
  os.data = tmpval;  
  os.length = len;  
  if(!X509V3_add1_i2d(exts, NID_id_pkix_OCSP_Nonce, &os, 0, X509V3_ADD_REPLACE))    
    goto err;  
  ret = 1; 
err:  
  if(!val) 
    OPENSSL_free(tmpval);  
  return ret;
}

int MY_OCSP_request_add1_nonce(OCSP_REQUEST *req, unsigned char *val, int len)
{  
  return MY_ocsp_add1_nonce(&req->tbsRequest->requestExtensions, val, len);
}

int MY_OCSP_basic_add1_nonce(OCSP_BASICRESP *resp, unsigned char *val, int len)
{  
  return MY_ocsp_add1_nonce(&resp->tbsResponseData->responseExtensions, val, len);
}


//--------------------------------------------------
// Creates an OCSP_REQUEST object
// pSigDoc - address of signed document. If not NULL then
// used to check if older openssl 0.9.6 style request must
// be constructed.
// req - buffer for storing the pointer of new object
// cert - client certificate to verify
// nonce - nonce value (e.g. client signature value RSA-128 bytes)
// nlen - nonce value length
// pkey - public key used to signe th request (not used yet)
//--------------------------------------------------
// VS - ver 1.66
int createOCSPRequest(SignedDoc* pSigDoc, OCSP_REQUEST **req, 
		      X509 *cert, X509* pCACert, byte* nonce, int nlen)
{
  int err = ERR_OK;
  OCSP_CERTID *id = 0;
	
  RETURN_IF_NULL_PARAM(req);
  RETURN_IF_NULL_PARAM(cert);
  RETURN_IF_NULL_PARAM(nonce);
  if((*req = OCSP_REQUEST_new()) != 0) {
    // VS - ver 1.66
    if((id = createOCSPCertid(cert, pCACert)) != 0) {
      if(OCSP_request_add0_id(*req, id)) {
	if(pSigDoc && 
	   (!strcmp(pSigDoc->szFormatVer, SK_XML_1_VER) ||
	    !strcmp(pSigDoc->szFormatVer, DIGIDOC_XML_1_1_VER) ||
	    !strcmp(pSigDoc->szFormatVer, DIGIDOC_XML_1_3_VER))) {
	     if((err = MY_OCSP_request_add1_nonce(*req, nonce, nlen)) != 0)
	       err = ERR_OK;
	   } else {
	     if((err = OCSP_request_add1_nonce(*req, nonce, nlen)) != 0)
	       err = ERR_OK;
	   }
      }
    }
  }
  return err;
}


void printErrors()
{
	int err = 0;
	char *str = 0;

	while((err = ERR_get_error()) != 0) {
		str = ERR_error_string(err, 0);
		if(str != 0)
			printf("ERROR : %s\n", str);
		else
			printf("ERROR (%d) : <no message>\n", err);
	}

}

//--------------------------------------------------
// sends an OCSP_REQUES object to remore server and
// retrieves the OCSP_RESPONSE object
// resp - buffer to store the new responses pointer
// req - request objects pointer
// url - OCSP responder URL
//--------------------------------------------------
int ddocPullUrl(const char* url, DigiDocMemBuf* pSendData, DigiDocMemBuf* pRecvData, 
		const char* proxyHost, const char* proxyPort)
{	
  BIO* cbio = 0, *sbio = 0;
  SSL_CTX *ctx = NULL;
  char *host = NULL, *port = NULL, *path = "/", buf[200];
  int err = ERR_OK, use_ssl = -1, rc;

  //RETURN_IF_NULL_PARAM(pSendData); // may be null if nothing to send?
  RETURN_IF_NULL_PARAM(pRecvData);
  RETURN_IF_NULL_PARAM(url);

  ddocDebug(3, "ddocPullUrl", "URL: %s, in: %d bytes", url, pSendData->nLen);
  //there is an HTTP proxy - connect to that instead of the target host
  if (proxyHost != 0 && *proxyHost != '\0') {
    host = (char*)proxyHost;
    if(proxyPort != 0 && *proxyPort != '\0')
      port = (char*)proxyPort;
    path = (char*)url;
  } else {
    if(OCSP_parse_url((char*)url, &host, &port, &path, &use_ssl) == 0) {
      ddocDebug(1, "ddocPullUrl", "Failed to parse the URL");
      return ERR_WRONG_URL_OR_PROXY; 
    }
  }
	
  if((cbio = BIO_new_connect(host)) != 0) {
    if(port != NULL) 
      BIO_set_conn_port(cbio, port);
    if(use_ssl == 1) {
      ctx = SSL_CTX_new(SSLv23_client_method());
      SSL_CTX_set_mode(ctx, SSL_MODE_AUTO_RETRY);
      sbio = BIO_new_ssl(ctx, 1);
      cbio = BIO_push(sbio, cbio);
    }
    //printf("Connecting\n");
    if ((rc = BIO_do_connect(cbio)) > 0) {
      ddocDebug(4, "ddocPullUrl", "Connected: %d", rc);
      //printf("Connected\n");
      if(pSendData && pSendData->nLen && pSendData->pMem) {
	rc = BIO_write(cbio, pSendData->pMem, pSendData->nLen);
	ddocDebug(4, "ddocPullUrl", "Sent: %d bytes, got: %d", pSendData->nLen, rc);
      }
      //printf("Sent: %d bytes, got: %d\n", pSendData->nLen, rc);
      do {
	memset(buf, 0, sizeof(buf));
	rc = BIO_read(cbio, buf, sizeof(buf)-1);
	ddocDebug(4, "ddocPullUrl", "Received: %d bytes\n", rc);
	//printf("Received: %d bytes\n", rc);
	if(rc > 0)
	  err = ddocMemAppendData(pRecvData, buf, rc);
      } while(rc > 0);
      ddocDebug(4, "ddocPullUrl", "Total received: %d bytes\n", pRecvData->nLen);
      //printf("Total received: %d bytes\n", pRecvData->nLen);
    } else {
      //if no connection
      if (host != NULL)
	err = ERR_WRONG_URL_OR_PROXY;
      else
	err = ERR_CONNECTION_FAILURE;
    }
    BIO_free_all(cbio);
    if (use_ssl != -1) {
      OPENSSL_free(host);
      OPENSSL_free(port);
      OPENSSL_free(path);
      SSL_CTX_free(ctx);
    }
  }
  else
    err = ERR_CONNECTION_FAILURE;
  return(err);
}

/* Quick and dirty HTTP OCSP request handler.
 * Could make this a bit cleverer by adding
 * support for non blocking BIOs and a few
 * other refinements.
 * Qick and dirty adaption of openssl -s 
 * OCSP_sendreq_bio() to add UserAgent HTTP header
 */

OCSP_RESPONSE *OCSP_sendreq_bio_withParams(BIO *b, char *path, 
				      OCSP_REQUEST *req, unsigned long ip_addr )
{
  BIO *mem = NULL;
  char tmpbuf[1024], adrhdr[100];
  OCSP_RESPONSE *resp = NULL;
  char *p, *q, *r;
  int len, retcode;
  static char req_txt[] =
"POST %s HTTP/1.0\r\n\
Content-Type: application/ocsp-request\r\n\
User-Agent: LIB %s/%s APP %s\r\n%s\
Content-Length: %d\r\n\r\n";

  adrhdr[0] = 0;
  if(ip_addr > 0)
    sprintf(adrhdr, "From: %d.%d.%d.%d\r\n", 
    (ip_addr>>24)&0xFF, (ip_addr>>16)&0xFF, (ip_addr>>8)&0xFF, ip_addr&0xFF); 
  len = i2d_OCSP_REQUEST(req, NULL);
  if(BIO_printf(b, req_txt, path, getLibName(), getLibVersion(), 
		getGUIVersion(), (ip_addr > 0 ? adrhdr : ""), len) < 0) {
    OCSPerr(OCSP_F_OCSP_SENDREQ_BIO,OCSP_R_SERVER_WRITE_ERROR);
    goto err;
  }
#if OPENSSL_VERSION_NUMBER > 0x00908000
  retcode = ASN1_i2d_bio((int (*)(void*, unsigned char**))i2d_OCSP_REQUEST, b, (unsigned char*)req);
#else
  retcode = ASN1_i2d_bio((int (*)())i2d_OCSP_REQUEST, b, (unsigned char*)req);
#endif
  if(retcode <= 0) {
    OCSPerr(OCSP_F_OCSP_SENDREQ_BIO,OCSP_R_SERVER_WRITE_ERROR);
    goto err;
  }
  mem = BIO_new(BIO_s_mem());
  if(!mem) goto err;
  /* Copy response to a memory BIO: socket bios can't do gets! */
  do {
    len = BIO_read(b, tmpbuf, sizeof tmpbuf);
    if(len < 0) {
      OCSPerr(OCSP_F_OCSP_SENDREQ_BIO,OCSP_R_SERVER_READ_ERROR);
      goto err;
    }
    BIO_write(mem, tmpbuf, len);
  } while(len > 0);
  if(BIO_gets(mem, tmpbuf, 512) <= 0) {
    OCSPerr(OCSP_F_OCSP_SENDREQ_BIO,OCSP_R_SERVER_RESPONSE_PARSE_ERROR);
    goto err;
  }
  /* Parse the HTTP response. This will look like this:
   * "HTTP/1.0 200 OK". We need to obtain the numeric code and
   * (optional) informational message.
   */
  
  /* Skip to first white space (passed protocol info) */
  for(p = tmpbuf; *p && !isspace((unsigned char)*p); p++) continue;
  if(!*p) {
    OCSPerr(OCSP_F_OCSP_SENDREQ_BIO,OCSP_R_SERVER_RESPONSE_PARSE_ERROR);
    goto err;
  }
  /* Skip past white space to start of response code */
  while(*p && isspace((unsigned char)*p)) p++;
  if(!*p) {
    OCSPerr(OCSP_F_OCSP_SENDREQ_BIO,OCSP_R_SERVER_RESPONSE_PARSE_ERROR);
    goto err;
  }
  /* Find end of response code: first whitespace after start of code */
  for(q = p; *q && !isspace((unsigned char)*q); q++) continue;
  if(!*q) {
    OCSPerr(OCSP_F_OCSP_SENDREQ_BIO,OCSP_R_SERVER_RESPONSE_PARSE_ERROR);
    goto err;
  }
  /* Set end of response code and start of message */ 
  *q++ = 0;
  /* Attempt to parse numeric code */
  retcode = strtoul(p, &r, 10);
  if(*r) goto err;
  /* Skip over any leading white space in message */
  while(*q && isspace((unsigned char)*q))  q++;
  if(*q) {
    /* Finally zap any trailing white space in message (include CRLF) */
    /* We know q has a non white space character so this is OK */
    for(r = q + strlen(q) - 1; isspace((unsigned char)*r); r--) *r = 0;
  }
  if(retcode != 200) {
    OCSPerr(OCSP_F_OCSP_SENDREQ_BIO,OCSP_R_SERVER_RESPONSE_ERROR);
    if(!*q) { 
      ERR_add_error_data(2, "Code=", p);
    }
    else {
      ERR_add_error_data(4, "Code=", p, ",Reason=", q);
    }
    goto err;
  }
  /* Find blank line marking beginning of content */	
  while(BIO_gets(mem, tmpbuf, 512) > 0)
  {
    for(p = tmpbuf; *p && isspace((unsigned char)*p); p++) continue;
    if(!*p) break;
  }
  if(*p) {
    OCSPerr(OCSP_F_OCSP_SENDREQ_BIO,OCSP_R_NO_CONTENT);
    goto err;
  }
  resp = d2i_OCSP_RESPONSE_bio(mem, NULL);
  if(!resp) {
    OCSPerr(OCSP_F_OCSP_SENDREQ_BIO,ERR_R_NESTED_ASN1_ERROR);
    goto err;
  }
 err:
  BIO_free(mem);
  return resp;
}


//--------------------------------------------------
// sends an OCSP_REQUES object to remore server and
// retrieves the OCSP_RESPONSE object
// resp - buffer to store the new responses pointer
// req - request objects pointer
// url - OCSP responder URL
// ip_addr - senders ip address if known or 0
//--------------------------------------------------
int sendOCSPRequest(OCSP_RESPONSE** resp, OCSP_REQUEST *req, 
		    char* url, char* proxyHost, char* proxyPort,
		    unsigned long ip_addr)
{	
  BIO* cbio = 0, *sbio = 0;
  SSL_CTX *ctx = NULL;
  char *host = NULL, *port = NULL, *path = "/";
  int err = ERR_OK, use_ssl = -1;
  int connResult = 0;

  RETURN_IF_NULL_PARAM(resp);
  RETURN_IF_NULL_PARAM(req);
  RETURN_IF_NULL_PARAM(url);

  //there is an HTTP proxy - connect to that instead of the target host
  if (proxyHost != 0 && *proxyHost != '\0') {
    host = proxyHost;
    if(proxyPort != 0 && *proxyPort != '\0')
      port = proxyPort;
    path = url;
  } else {
    if(OCSP_parse_url(url, &host, &port, &path, &use_ssl) == 0) {
      //printf("BIO_parse_url failed\n");
      return ERR_WRONG_URL_OR_PROXY; 
    }
  }
  if((cbio = BIO_new_connect(host)) != 0) {
    if (port != NULL) 
      BIO_set_conn_port(cbio, port);
    if (use_ssl == 1) {
      ctx = SSL_CTX_new(SSLv23_client_method());
      SSL_CTX_set_mode(ctx, SSL_MODE_AUTO_RETRY);
      sbio = BIO_new_ssl(ctx, 1);
      cbio = BIO_push(sbio, cbio);
    }
    if ((connResult = BIO_do_connect(cbio)) > 0) {
      //printf("BIO_do_connect returned %d\n", connResult);
      *resp = OCSP_sendreq_bio_withParams(cbio, path, req, ip_addr);
      //printf("OCSP_sendreq_bio answered %lX\n", *resp);
      err = (*resp == 0) ? ERR_OCSP_WRONG_URL : ERR_OK;
      //if (*resp == 0) 
      //  printErrors();
    } else {
      //printf("BIO_do_connect failed, rc = %d, shouldRetry = %d\n", connResult, BIO_should_retry(cbio));
      //printErrors();
      //if no connection
      if (host != NULL)
	err = ERR_WRONG_URL_OR_PROXY;
      else
	err = ERR_CONNECTION_FAILURE;
    }
    BIO_free_all(cbio);
    if (use_ssl != -1) {
      OPENSSL_free(host);
      OPENSSL_free(port);
      OPENSSL_free(path);
      SSL_CTX_free(ctx);
    }
  }
  else
    err = ERR_CONNECTION_FAILURE;
  return(err);
}


//--------------------------------------------------
// Creates and writes an OCSP_REQUEST object
// to disk
// pSigDoc - signedDoc address
// signerCertFile - cert file to verify
// issuertCertFile - this certs direct CA cert
// nonce - nonce (signature value)
// nlen - nonce length
// szOutputFile - output filename
//--------------------------------------------------
 EXP_OPTION int writeOCSPRequest(SignedDoc* pSigDoc, 
				 const char* signerCertFile, const char* issuertCertFile,
				 byte* nonce, int nlen, const char* szOutputFile) 

{
  OCSP_REQUEST *req = 0;
  X509 *cert = 0, *issuer = 0;
  int err = ERR_OK, l1;
  //EVP_PKEY* pkey; 
  byte buf1[DIGEST_LEN+2];

  RETURN_IF_NULL_PARAM(signerCertFile);
  RETURN_IF_NULL_PARAM(issuertCertFile);
  RETURN_IF_NULL_PARAM(nonce);
  RETURN_IF_NULL_PARAM(szOutputFile);

  if((err = ReadCertificate(&cert, signerCertFile)) == ERR_OK) {
    //pkey = ReadPublicKey(signerCertFile);
    if((err = ReadCertificate(&issuer, issuertCertFile)) == ERR_OK) {
      l1 = sizeof(buf1);
      calculateDigest(nonce, nlen, DIGEST_SHA1, buf1, &l1);
      err = createOCSPRequest(pSigDoc, &req, cert, issuer, buf1, l1);
      //WriteOCSPRequest(szOutputFile, req);
      X509_free(issuer);
    }
    X509_free(cert);
  }
  return err;
}

//--------------------------------------------------
// Signs an OCSP_REQUEST using PKCS#12 conteiner
// req - OCSP_REQUEST
// filename - PKCS#12 conteiner file
// passwd - key decryption passwd
//--------------------------------------------------
EXP_OPTION int signOCSPRequestPKCS12(OCSP_REQUEST *req, const char* filename, const char* passwd)
{
	
	EVP_PKEY *pkey;
	int err = ERR_OK;
	time_t tNow;

	STACK_OF(X509)* certs = NULL;
	X509* x509=0;

	RETURN_IF_NULL_PARAM(filename);
	err = ReadCertificateByPKCS12(&x509, filename, passwd, &pkey);
	RETURN_IF_NOT(err == ERR_OK, err);

	// VS: ver 1.66
	time(&tNow);
	err = isCertValid(x509, tNow);
	if (err != ERR_OK)
		X509_free(x509);
	RETURN_IF_NOT(err == ERR_OK, ERR_PKCS12_EXPIRED);
	certs = sk_X509_new_null();
	RETURN_IF_NULL(certs);

	//sk_X509_push(certs, x509);
	if (! OCSP_request_sign(req,x509,pkey,EVP_sha1(),certs,0)) {
		EVP_PKEY_free(pkey);
		err = ERR_OCSP_SIGN;
		SET_LAST_ERROR(err);
	}
	X509_free(x509);
	EVP_PKEY_free(pkey);
	return err;
}

//--------------------------------------------------
// Signs an OCSP_REQUEST using X509 cert and separate keyfile
// req - OCSP_REQUEST
// certFile - signers certificate file
// keyfile - signer's key file
// passwd - key decryption passwd
//--------------------------------------------------
EXP_OPTION int signOCSPRequest(OCSP_REQUEST *req,const char* certFile,const char* keyfile,const char* passwd){
	
	EVP_PKEY *pkey;
	int err = ERR_OK;
	STACK_OF(X509)* certs = NULL;
	X509* x509 = NULL;

	certs = sk_X509_new_null();
	RETURN_IF_NULL_PARAM(certs);

	if((err = ReadCertificate(&x509, certFile)) != ERR_OK) {
		SET_LAST_ERROR_RETURN_CODE(ERR_PKCS_CERT_LOC);
	}
	sk_X509_push(certs, x509);
	if((err = ReadPrivateKey(&pkey, keyfile, passwd, FILE_FORMAT_PEM)) == ERR_OK) {
		//ASN1_item_sign(ASN1_ITEM_rptr(OCSP_REQINFO),req->optionalSignature->signatureAlgorithm,NULL,req->optionalSignature->signature,req->tbsRequest,pkey,setSignAlgorithm(EVP_sha1()));
		//OCSP_request_sign_internal(req, x509,pkey, NULL);
		if(! OCSP_request_sign(req,x509,pkey,EVP_sha1(),certs,0)){
			EVP_PKEY_free(pkey);
			SET_LAST_ERROR_RETURN_CODE(ERR_OCSP_SIGN);
		}
		//printf("OCSP_request_sign()=%d \n",r);
		EVP_PKEY_free(pkey);
	}else{
		SET_LAST_ERROR_RETURN_CODE(ERR_PRIVKEY_READ);
	}
	return err;
}

//--------------------------------------------------
// Creates and sends an OCSP_REQUEST object
// to the notary server, receives the response
// and uses it to create a confirmation object.
// pSigDoc - signed doc info
// pSigInfo - signature info
// caCerts - responder CA certs chain
// notaryCert - notarys cert search
// pkcs12FileName -  
// pkcs12Password - 
// notaryURL - notarys URL
// proxyHost - proxy host if needed
// proxyPort - proxy port if needed
//--------------------------------------------------
EXP_OPTION int getConfirmation(SignedDoc* pSigDoc, SignatureInfo* pSigInfo, 
			       const X509** caCerts, const X509* pNotCert,
			       char* pkcs12FileName, char* pkcs12Password,
			       char* notaryURL, char* proxyHost, char* proxyPort) 

{
  // default way to invoke it is without callers ip.
  return getConfirmationWithIp(pSigDoc, pSigInfo, caCerts, pNotCert,
			       pkcs12FileName, pkcs12Password,
			       notaryURL, proxyHost, proxyPort, 0);
}

//--------------------------------------------------
// Creates and sends an OCSP_REQUEST object
// to the notary server, receives the response
// and uses it to create a confirmation object.
// pSigDoc - signed doc info
// pSigInfo - signature info
// caCerts - responder CA certs chain
// notaryCert - notarys cert search
// pkcs12FileName -  
// pkcs12Password - 
// notaryURL - notarys URL
// proxyHost - proxy host if needed
// proxyPort - proxy port if needed
// ip - callers ip address if known
//--------------------------------------------------
EXP_OPTION int getConfirmationWithIp(SignedDoc* pSigDoc, SignatureInfo* pSigInfo, 
				     const X509** caCerts, const X509* pNotCert,
				     char* pkcs12FileName, char* pkcs12Password,
				     char* notaryURL, char* proxyHost, char* proxyPort,
				     unsigned long ip) 

{
  OCSP_REQUEST *req = 0;
  OCSP_RESPONSE *resp = 0;
  X509 *cert = 0, *pCA = 0;
  int err = ERR_OK, l1, i;
  byte buf1[DIGEST_LEN+2];
  NotaryInfo* pNotInf = NULL;
  
  RETURN_IF_NULL_PARAM(pSigDoc);
  RETURN_IF_NULL_PARAM(pSigInfo);
  cert = (X509*)pSigInfo->pX509Cert;
  RETURN_IF_NULL(cert);	
  RETURN_IF_NULL_PARAM(notaryURL);
  
  clearErrors();
	
  l1 = sizeof(buf1);
  err = calculateDigest(pSigInfo->szSigValue, pSigInfo->nSigLen, DIGEST_SHA1, buf1, &l1);
  RETURN_IF_NOT(err == ERR_OK, err);
  
  // find lowest CA cert
  for(i = 0; (caCerts != NULL) && (caCerts[i] != NULL); i++)
    pCA = (X509*)caCerts[i];
  err = createOCSPRequest(pSigDoc, &req, cert, pCA, buf1, l1);
  RETURN_IF_NOT(err == ERR_OK, err);

  // if both are NULL then this means don't sign OCSP requests
  if(pkcs12FileName /*&& pkcs12Password*/) {
    ddocDebug(3, "getConfirmationWithIp", "Sign OCSP request with: %s", pkcs12FileName);
    err=signOCSPRequestPKCS12(req, pkcs12FileName, pkcs12Password);
  }
  RETURN_IF_NOT(err == ERR_OK, err);
  
  ddocDebug(3, "getConfirmationWithIp", "Send OCSP to: %s over: %s:%s", notaryURL,
	    (proxyHost ? proxyHost : ""), (proxyPort ? proxyPort : ""));
  err = sendOCSPRequest(&resp, req, notaryURL, proxyHost, proxyPort, ip);
  RETURN_IF_NOT(err == ERR_OK, err);

  err = NotaryInfo_new(&pNotInf, pSigDoc, pSigInfo);
  RETURN_IF_NOT(err == ERR_OK, err);

  err = initializeNotaryInfoWithOCSP(pSigDoc, pNotInf, resp, NULL, 1);
  RETURN_IF_NOT(err == ERR_OK, err);

  if(caCerts && pNotCert) {
    err = finalizeAndVerifyNotary(pSigDoc, pNotInf, caCerts, pNotCert);
    RETURN_IF_NOT(err == ERR_OK, err);
  }
  return ERR_OK;
}


//--------------------------------------------------
// Adds responder certificate to notary, searches it's
// CA chain and then verifies notary
// pSigDoc - signed doc info
// pSigInfo - signature info
// caCertSearches - responder CA certs chain
// notaryCert - notarys cert search
// returns error code
//--------------------------------------------------
int EXP_OPTION finalizeAndVerifyNotary(SignedDoc* pSigDoc, NotaryInfo* pNotInf,
				       const X509** caCerts, const X509* pNotCert)
{
  int err = ERR_OK;

  RETURN_IF_NULL_PARAM(pNotCert);
  RETURN_IF_NULL_PARAM(caCerts);
  ddocDebug(3, "finalizeAndVerifyNotary", "Notary: %s cert: %s", pNotInf->szId, (pNotCert ? "OK" : "NULL"));
  err = addNotaryInfoCert(pNotInf, (X509*)pNotCert);
  RETURN_IF_NOT(err == ERR_OK, err);
  err = calcNotaryDigest(pSigDoc, pNotInf);
  RETURN_IF_NOT(err == ERR_OK, err);
  err = verifyNotaryInfoCERT(pSigDoc, pNotInf, caCerts, NULL, pNotCert);
  RETURN_IF_NOT(err == ERR_OK, err);
  ddocDebug(3, "finalizeAndVerifyNotary", "rc: %d cert: %s cert2: %s", err, (pNotCert ? "OK" : "NULL"), (pNotInf->pX509Cert ? "OK" : "NULL"));
  return ERR_OK;
}


//--------------------------------------------------
// Verfies OCSP_RESPONSE signature
// pResp - signed OCSP response
// caCerts - CA certificate pointer array terminated with NULL
// CApath - path to (directory) all certs
// notCertFile - Notary (e.g. OCSP responder) cert file 
//--------------------------------------------------
int verifyOCSPResponse(OCSP_RESPONSE* pResp, 
				    const X509** caCerts, const char *CApath, 
				    const X509* notCert)
{
	X509_STORE *store;
	OCSP_BASICRESP* bs;
	STACK_OF(X509)* ver_certs = NULL;
	int err = ERR_OK;

	RETURN_IF_NULL_PARAM(pResp);
	RETURN_IF_NOT(ASN1_ENUMERATED_get(pResp->responseStatus) == 0, ERR_OCSP_UNSUCCESSFUL);
	RETURN_IF_NOT(OBJ_obj2nid(pResp->responseBytes->responseType) == NID_id_pkix_OCSP_basic, ERR_OCSP_UNKNOWN_TYPE);
	RETURN_IF_NOT((bs = OCSP_response_get1_basic(pResp)) != NULL, ERR_OCSP_NO_BASIC_RESP);
	
	// now create an OCSP object and check its validity
	if((setup_verifyCERT(&store, CApath, caCerts)) == ERR_OK) {
		// new basic response
		// create OCSP basic response
		ver_certs = sk_X509_new_null();
		if(ver_certs) {
			sk_X509_push(ver_certs, notCert);
			err = OCSP_basic_verify(bs, ver_certs, store, OCSP_TRUSTOTHER);
			if(err == ERR_LIB_NONE) {
				err = ERR_OK;
			} else {
				//checkErrors();
				SET_LAST_ERROR(ERR_OCSP_WRONG_RESPID);
				err = ERR_OCSP_WRONG_RESPID;
			}
			// cleanup
			sk_X509_free(ver_certs);
		}
		X509_STORE_free(store);
	}
	if(bs)
		OCSP_BASICRESP_free(bs);
	return err;
}

int checkNonceAndCertbyOCSP(OCSP_RESPONSE* resp, X509* cert, byte* nonce1, int nonceLen)
{
	int err = ERR_OK, n;
	char buf[100];
	// OCSP_RESPBYTES *rb = resp->responseBytes;
	OCSP_BASICRESP *br = NULL;
	OCSP_RESPDATA  *rd = NULL;
	// OCSP_RESPID *rid = NULL;
	// OCSP_CERTSTATUS *cst = NULL;
	OCSP_SINGLERESP *single = NULL;
	OCSP_CERTID *cid = NULL;
	X509_EXTENSION *nonce;
	long certNr1;
	X509_NAME *iname;
	unsigned char *ikey;
	
	
	RETURN_IF_NULL_PARAM(resp);
	RETURN_IF_NULL_PARAM(cert);
	
	if((br = OCSP_response_get1_basic(resp)) == NULL) 
		SET_LAST_ERROR_RETURN_CODE(ERR_OCSP_NO_BASIC_RESP);
	rd = br->tbsResponseData;
	n = sk_OCSP_SINGLERESP_num(rd->responses);
	RETURN_IF_NOT(n == 1, ERR_OCSP_ONE_RESPONSE);
	single = sk_OCSP_SINGLERESP_value(rd->responses, 0);
	RETURN_IF_NOT(single, ERR_OCSP_ONE_RESPONSE);
	cid = single->certId;
	RETURN_IF_NULL(cid);
	err = handleOCSPCertStatus(single->certStatus->type);
	if(err)
	  SET_LAST_ERROR_RETURN_CODE(err);
	if(single->singleExtensions) 
		SET_LAST_ERROR_RETURN_CODE(ERR_OCSP_NO_SINGLE_EXT);
	if(!rd->responseExtensions ||
		(sk_X509_EXTENSION_num(rd->responseExtensions) != 1) ||
		((nonce = sk_X509_EXTENSION_value(rd->responseExtensions, 0)) == NULL)) 
		SET_LAST_ERROR_RETURN_CODE(ERR_OCSP_NO_NONCE);
	i2t_ASN1_OBJECT(buf, sizeof(buf), nonce->object);
	if(strcmp(buf, OCSP_NONCE_NAME)) 
		SET_LAST_ERROR_RETURN_CODE(ERR_OCSP_NO_NONCE);
	
	// check serial number
	certNr1 = ASN1_INTEGER_get(cid->serialNumber);
	if(certNr1 != ASN1_INTEGER_get(X509_get_serialNumber(cert)))
		SET_LAST_ERROR_RETURN_CODE(ERR_WRONG_CERT);
	
	// check issuer name hash
	iname = X509_get_issuer_name(cert);
	n = sizeof(buf);
	X509_NAME_digest(iname, EVP_sha1(), (byte*)buf, (unsigned int*)&n);
	err = compareByteArrays((byte*)buf, (unsigned int)n, cid->issuerNameHash->data, cid->issuerNameHash->length);
	RETURN_IF_NOT(err == ERR_OK, err);
	
	// check issuer key hash
	if((ikey = get_authority_key(cert->cert_info->extensions)) != NULL) {
		err = compareByteArrays(ikey, strlen((const char*)ikey), 
			cid->issuerKeyHash->data, cid->issuerKeyHash->length);
		// cleanup ikey
		free(ikey);
	} 
	
	// verify nonce value
	err = compareByteArrays(nonce->value->data, nonce->value->length, nonce1, nonceLen);
	if (err != ERR_OK) SET_LAST_ERROR(err);
	return err;
}

//--------------------------------------------------
// Checks if this element tag contains the
// required attributes with the given values
// data - input data, XML tags data (not content but the attributes)
// nAttrs - number of attributes to check
// attNames - array of attribute names
// attValues - array of attribute values
// returns 0 if OK (all atributes found or none desired)
//--------------------------------------------------
int checkAttrs(const char* data, int nAttrs,
				const char** attNames, const char** attValues)
{
  int remains = 0, i;
  char *pTmp1 = 0, *pTmp2 = 0;

  RETURN_IF_NULL_PARAM(data);

  if(nAttrs) {
    RETURN_IF_NULL_PARAM(attNames);
    RETURN_IF_NULL_PARAM(attValues);
    remains = nAttrs; // must find nAttrs values
    for(i = 0; i < nAttrs; i++) {
      RETURN_IF_NULL(attNames[i]);
      RETURN_IF_NULL(attValues[i]);
      if((pTmp1 = strstr(data, attNames[i])) != 0) {
	if((pTmp2 = strstr(pTmp1, "\"")) != 0) {
	  if(!strncmp(pTmp2+1, attValues[i], strlen(attValues[i])))
	    remains--; // found one
	}
      }
    }
  }
  if (remains == 0) 
    return ERR_OK;
  else 
    return remains;
}

char* findString(char* mainBuf, char* search)
{
	char* pTmp = NULL;
	// first find in the latest 2KB
	pTmp = strstr(mainBuf+2048, search);
	// if not found check the previous buffer
	// as well because the tag could have been broken
	// between two buffer borders
	if(!pTmp)
		pTmp = strstr(mainBuf, search);
	return pTmp;
}





//============================================================
// Verifies the whole document, but returns on first error
// Use the functions defined earlier to verify all contents
// step by step.
// pSigDoc - signed doc data
// signerCA - direct signer CA certs filename
// szDateFile - name of the digidoc file
// bUseCA - use CA certs or not 1/0
// pPaths - file name / path pairs for detatched files
// nPaths - number of file name / path pairs
//============================================================
EXP_OPTION int verifySigDocCERT(const SignedDoc* pSigDoc, 
								// VS - ver 1.66
				const void* signerCA, const X509** caCerts,
				 const char* caPath, const X509* notCert,
				 const char* szDataFile, int bUseCA, FilePath* pPaths, int nPaths)
{
	SignatureInfo* pSigInfo;
	NotaryInfo* pNotInfo;
	int i, d, err = ERR_OK;

	RETURN_IF_NULL_PARAM(pSigDoc);
	//assert(pSigDoc);
	d = getCountOfSignatures(pSigDoc);
	for(i = 0; i < d; i++) {
		pSigInfo = getSignature(pSigDoc, i);
		err = verifySignatureInfoCERT(pSigDoc, pSigInfo, signerCA, 
			szDataFile, bUseCA, pPaths, nPaths);
		RETURN_IF_NOT(err == ERR_OK, err);
	}
	d = getCountOfNotaryInfos(pSigDoc);
	for(i = 0; (!err) && (i < d); i++) {
		pNotInfo = getNotaryInfo(pSigDoc, i);
		// VS - ver 1.66
		err = verifyNotaryInfoCERT(pSigDoc, pNotInfo, caCerts, caPath, notCert);
		RETURN_IF_NOT(err == ERR_OK, err);
	}
	return err;
}

//============================================================
// Verifies this signature
// pSigDoc - signed doc data
// pSigInfo - signature info object
// signerCA - direct signer CA certs filename
// szDataFile - provide to read <SignedInfo> and <SignedProperties>
// from original file and use it for hash function.
// This is usefull if the file has been generated by
// another library and possibly formats these elements
// differently.
// bUseCA - use CA certs or not 1/0
// pPaths - file name / path pairs for detatched files
// nPaths - number of file name / path pairs
//============================================================
EXP_OPTION int verifySignatureInfoCERT(const SignedDoc* pSigDoc, const SignatureInfo* pSigInfo, 
					const void* signerCACert, const char* szDataFile, int bUseCA,
					FilePath* pPaths, int nPaths)
{
	int err = ERR_OK;
	int j, k, i;
	char* pFileName = NULL;
	X509* cert;
	DocInfo* pDocInfo;
	DataFile* pDf = NULL;

	RETURN_IF_NULL_PARAM(pSigInfo);

	err = verifyEstIDSignature((const byte*)pSigInfo->szSigInfoRealDigest, 
				pSigInfo->nSigInfoRealDigestLen, DIGEST_SHA1,
				pSigInfo->szSigValue, pSigInfo->nSigLen, (X509*)pSigInfo->pX509Cert);
	RETURN_IF_NOT(err == ERR_OK, ERR_SIGNATURE);
	k = getCountOfDocInfos(pSigInfo);
	ddocDebug(4, "verifySignatureInfoCERT", "DFs: %d", k);
	for(j = 0; (err == ERR_OK) && (j < k); j++) {
		pDocInfo = getDocInfo(pSigInfo, j);
		ddocDebug(4, "verifySignatureInfoCERT", "Verify doc: %d - \'%s\'", j, pDocInfo->szDocId);
		RETURN_IF_NULL(pDocInfo);
		pDf = getDataFileWithId(pSigDoc, pDocInfo->szDocId);
		RETURN_IF_NULL(pDf);
		// find data file path
		for(i = 0; i < nPaths; i++) {
			if(!strcmp(pDf->szFileName, pPaths->szFileName)) {
				pFileName = pPaths->szPath;
				break;
			}
		}
		if(!pFileName)
		  pFileName = pDf->szFileName;
		err = verifySigDocDigest(pSigDoc, pSigInfo, pDocInfo, pFileName, szDataFile);
		ddocDebug(4, "verifySignatureInfoCERT", "Verify doc: %s - %d", pDocInfo->szDocId, err);
		RETURN_IF_NOT(err == ERR_OK, err);
		err = verifySigDocMimeDigest(pSigDoc, pSigInfo, pDocInfo, NULL);
		RETURN_IF_NOT(err == ERR_OK, err);
	}
		
	err = verifySigDocSigPropDigest(pSigDoc, pSigInfo, szDataFile);
	RETURN_IF_NOT(err == ERR_OK, err);
	err = verifySigCert(pSigInfo);
	RETURN_IF_NOT(err == ERR_OK, err);

	cert = getSignCertData(pSigInfo);
	// VS: ver 2.2.4 - removed this check as OCSP check is sufficient
	//err = isCertValid(cert, convertStringToTimeT(pSigDoc, pSigInfo->szTimeStamp));
	RETURN_IF_NOT(err == ERR_OK, err);
	if(bUseCA)
	  err = isCertSignedByCERT((const X509*)cert, (const X509*)signerCACert);
	RETURN_IF_NOT(err == ERR_OK, err);
	return ERR_OK;
}


//--------------------------------------------------
// Finds all detatched files and returns their
// names. The user can now set the path portion.
// Allocaes new memory that caller must free 
// with FilePaths_free()
// pSigDoc - signed document
// pPaths - pointer to FilePaths array
// nPaths - pointer to number of paths
//--------------------------------------------------
EXP_OPTION int findDetatchedDataFiles(SignedDoc* pSigDoc, FilePath** pPaths, int* nPaths)
{
	FilePath *pPath, *pTmp;
	int err = ERR_OK, d;

	RETURN_IF_NULL_PARAM(pSigDoc);
	RETURN_IF_NULL_PARAM(pPaths);
	RETURN_IF_NULL_PARAM(nPaths);
	*pPaths = NULL;
	*nPaths = 0;
	for(d = 0; d < pSigDoc->nDataFiles; d++) {
		DataFile* pDf = pSigDoc->pDataFiles[d];
		if (!strcmp(pDf->szContentType, CONTENT_DETATCHED)) {
			if (*pPaths && *nPaths) {
				pTmp = (FilePath*)realloc(*pPaths, sizeof(FilePath) * (*nPaths+1));
				if (pTmp != NULL) { //malloc failure!
					*pPaths	= pTmp;
					*nPaths = *nPaths + 1;	
					pPath = (*pPaths) + ((*nPaths) - 1);
				} else 
					SET_LAST_ERROR_RETURN_CODE(ERR_BAD_ALLOC);
			} else {
				pTmp = (FilePath*)malloc(sizeof(FilePath));
				if (pTmp != NULL) { //malloc failure!
					*pPaths = pTmp;
					pPath = *pPaths;
					*nPaths = 1;
				} else 
					 SET_LAST_ERROR_RETURN_CODE(ERR_BAD_ALLOC);
			}
			pPath->szFileName = pPath->szPath = NULL;
			pPath->szFileName = strdup(pDf->szFileName);
		}
	}
	return err;
}


//--------------------------------------------------
// Frees file paths array
// pPaths - FilePaths array
// nPaths - number of paths
//--------------------------------------------------
EXP_OPTION void FilePaths_free(FilePath* pPaths, int nPaths)
{
	int i;
	for(i = 0; i < nPaths; i++) {
		free(pPaths[i].szFileName);
		free(pPaths[i].szPath);
	}
	free(pPaths);
}

//--------------------------------------------------
// Finds the contents of a given XML tag
// in the given file.
// data - buffer for tag content data (caller must deallocate)
// tagName - tag name to search
// nAttrs - number of attributes to check
// attNames - array of attribute names
// attValues - array of attribute values
// withTags - 1 if include tags themselves, else 0
// returns 0 if tag was found and data read.
//--------------------------------------------------
int readTagContents(char** data, const char* fileName, 
					const char* tagName, int nAttrs,
					const char** attNames, const char** attValues,
					int withTags)
{
  int err = ERR_OK, status, len, level;
  FILE *hFile = 0;
  char *pTmp1 = 0, *pTmp2 = 0, *pTmp3 = 0, *pBegin = 0, *pData = NULL;
  char buf1[4097], buf2[100];

  RETURN_IF_NULL_PARAM(data);
  RETURN_IF_NULL_PARAM(fileName);
  RETURN_IF_NULL_PARAM(tagName);
  RETURN_IF_NULL_PARAM(attNames);
  RETURN_IF_NULL_PARAM(attValues);

  if((hFile = fopen(fileName, "rb")) != 0) {
    status = 0; // nothing found yet
    level = 0;
    memset(buf1, 0, sizeof(buf1));
    // allways load the second half of the buffer 
    // warning - assignment in conditional expression -> yes but the code seems clearer this way!
    while((len = fread(buf1+2048, 1, 2048, hFile)) && status < 2) {
      switch(status) {
      case 0:
	// find <tagName
	sprintf(buf2, "<%s ", tagName);	
	pTmp1 = findString(buf1, buf2);
	while(pTmp1 && (status == 0) && ((int)(pTmp1-buf1) < (int)sizeof(buf1))) {
	  pTmp2 = strstr(pTmp1, ">");
	  if(pTmp2) {
	    *pTmp2 = 0;
	    err = checkAttrs(pTmp1, nAttrs, attNames, attValues);
	    *pTmp2 = '>';
	    if(!err) {
	      // mark the found tag
	      // in order not to later mistake this 
	      // for a new level. Take also buffer moving
	      // in account
	      pBegin = pTmp1-2048;
	      status = 1; // now search for...
	      if(withTags) {								
		sprintf(buf2, "</%s>", tagName);				
		if((pTmp3 = strstr(pTmp1, buf2)) != 0) 
		  *(pTmp3+strlen(buf2)) = 0;
		len = strlen(pTmp1)+1;
		pData = (char*)malloc(len);
		memset(pData, 0, len);
		RETURN_IF_BAD_ALLOC(pData);
		strcpy(pData, pTmp1);
		if(pTmp3) {
		  *data = pData;
		  status = 2;
		}
	      } else {
		pTmp2++; // first byte of content data
		// find </tagName>
		sprintf(buf2, "</%s>", tagName);				
		if((pTmp3 = strstr(pTmp1, buf2)) != 0) 
		  *pTmp3 = 0;								
		len = strlen(pTmp2);
		pData = (char*)malloc(len+1);
		RETURN_IF_BAD_ALLOC(pData);
		strcpy(pData, pTmp2);
		if(pTmp3) {
		  *data = pData;
		  status = 2;
		}
	      } // else
	    } // if(!err)
	    else
	      pTmp1 = strstr(pTmp2, buf2);
	  } // if(pTmp2)
	  else
	    pTmp1++;
	} // if(pTmp1)
	break;
      case 1:
	sprintf(buf2, "</%s>", tagName);
	pTmp3 = findString(buf1, buf2);
	// if the found end-tag is fully in the 
	// previous buffer then if cannot be the right 
	// one because I would have noticed it in
	// the last step
	if((pTmp3+strlen(buf2)) < (buf1+2048))
	  pTmp3 = NULL;
	sprintf(buf2, "<%s ", tagName);				
	pTmp1 = findString(buf1, buf2);
	if(pTmp1 && pTmp1 > pBegin && !pTmp3) 
	  level++;
	if(pTmp3 && !level) {
	  if(withTags) {
	    sprintf(buf2, "</%s>", tagName);
	    *(pTmp3 + strlen(buf2)) = 0;
	  } else
	    *pTmp3 = 0;
	  *data = pData;
	  status = 2;
	}
	if(pTmp3 && level > 0)
	  level--;
	len = strlen(buf1+2048);
	if(len) {
	  RETURN_IF_NULL(pData);
	  pData = (char*)realloc(pData, strlen(pData)+len+1);
	  strcpy(strchr(pData, 0), buf1+2048);
	  *data = pData;
	}
	break;
	
      default:
	break;
      }
      memcpy(buf1, buf1+2048, 2048);
      memset(buf1+2048, 0, 2049);
    } // while
    fclose(hFile);
  } // if(hFile
  else
    err = ERR_FILE_READ;
  /*if(hFile = fopen("c:\\veiko\\work\\sk\\test1.xml", "wb")) {
    fputs(pData, hFile);
    fclose(hFile);
    }*/
  
  return (pData == NULL);
}

// ASN1 structure prefix - RSA-SHA1 signature with 20 bytes follows
char g_sigPrefix[] = { 48, 33, 48, 9, 6, 5, 43, 14, 3, 2, 26, 5, 0, 4, 20 };

//--------------------------------------------------
// Prepares a new signature for signing and calculates
// the final hash value to sign.
// pSigDoc - signed document object
// ppSigInfo - pointer for address of newly allocated signature
// manifest - manifest or role
// city - signers address , city
// state - signers address , state or province
// zip - signers address , postal code
// country - signers address , country name
// id - id for new signature. Optional, use NULL for default
// return returns error code or ERR_OK
//--------------------------------------------------
EXP_OPTION int ddocPrepareSignature(SignedDoc* pSigDoc, SignatureInfo** ppSigInfo,
			 const char* manifest, const char* city, 
			 const char* state, const char* zip, 
			 const char* country, X509* pCert, const char* id)
{
  int err = ERR_OK, l1;
  char buf1[50];

  ddocDebug(3, "ddocPrepareSignature", "Preparing signature manifest: %s country: %s, state: %s, city: %s, zip: %s, cert: %s, id: %s", 
	    (manifest ? manifest : "NULL"), (country ? country : "NULL"),
	    (state ? state : "NULL"), (city ? city : "NULL"), (zip ? zip : "NULL"),
	    (pCert ? "OK" : "ERROR"), (id ? id : "NULL"));
  // check mandator fields
  RETURN_IF_NULL_PARAM(pSigDoc);
  RETURN_IF_NULL_PARAM(ppSigInfo);
  RETURN_IF_NULL_PARAM(pCert);
  clearErrors();
  // add new signature
  err = SignatureInfo_new(ppSigInfo, pSigDoc, id);
  RETURN_IF_NOT(err == ERR_OK, err);
  // automatically calculate doc-info elements for this signature
  addAllDocInfos(pSigDoc, *ppSigInfo);
  // add signature production place
  if (city || state || zip || country)
    err = setSignatureProductionPlace(*ppSigInfo, city, state, zip, country);
  // add user roles/manifests
  if (manifest)
    err = addSignerRole(*ppSigInfo, 0, manifest, -1, 0);
  RETURN_IF_NOT(err == ERR_OK, err);
  // add signers certificate
  err = setSignatureCert(*ppSigInfo, pCert);
  RETURN_IF_NOT(err == ERR_OK, err); 
  // timestamp
  createTimestamp(pSigDoc, (char*)buf1);
  setString((char**)&((*ppSigInfo)->szTimeStamp), (const char*)buf1, -1);
  // now calculate signed properties digest
  err = calculateSignedPropertiesDigest(pSigDoc, *ppSigInfo);
  setString((char**)&((*ppSigInfo)->szSigPropRealDigest), 
	    (const char*)(*ppSigInfo)->szSigPropDigest, 
	    (*ppSigInfo)->nSigPropDigestLen);
  (*ppSigInfo)->nSigPropRealDigestLen = (*ppSigInfo)->nSigPropDigestLen;
  // signed-info digest
  setString(&((*ppSigInfo)->szSigType), SIGN_RSA_NAME, -1);
  // calc signed-info digest
  l1 = sizeof(buf1);
  err = calculateSignedInfoDigest(pSigDoc, *ppSigInfo, (byte*)buf1, &l1);
  setString((char**)&((*ppSigInfo)->szSigInfoRealDigest), (const char*)buf1, l1);
  (*ppSigInfo)->nSigInfoRealDigestLen = l1;
  // debug output - final hash to sign
  l1 = sizeof(buf1);
  encode((const byte*)(*ppSigInfo)->szSigInfoRealDigest, 
	 (*ppSigInfo)->nSigInfoRealDigestLen, (byte*)buf1, &l1);
  ddocDebug(3, "ddocPrepareSignature", "signing hash %s len: %d b64len: %d", 
	    buf1, (*ppSigInfo)->nSigInfoRealDigestLen, l1);
  
  return err;
}

//--------------------------------------------------
// Returns the hash to be signed with or without ASN1
// prefix and with or without base64 encoding
// pSigInfo - signature address
// pBuf - buffer for hash value with or without prefix
// pBufLen - pointer to buffer length
// b64 - return base64 encode 1/0
// bWithAsn1Prefix - return with or without ASN1 prefix 1/0
// return returns error code or ERR_OK
//--------------------------------------------------
EXP_OPTION  int ddocGetSignedHash(SignatureInfo* pSigInfo, char* pBuf, int* pBufLen, int b64, int bWithAsn1Prefix)
{
  int err = ERR_OK, l1;
  char buf1[50];

  RETURN_IF_NULL_PARAM(pSigInfo);
  RETURN_IF_NULL_PARAM(pBuf);
  //RETURN_IF_NOT(err == ERR_OK, err);
  // TODO: check buflen
  if(b64) {
    if(bWithAsn1Prefix) {
      memcpy(buf1, g_sigPrefix, sizeof(g_sigPrefix));
      memcpy(buf1 + sizeof(g_sigPrefix), pSigInfo->szSigInfoRealDigest, 
	     pSigInfo->nSigInfoRealDigestLen);
      l1 = pSigInfo->nSigInfoRealDigestLen + sizeof(g_sigPrefix);
    }
    else {
      memcpy(buf1, pSigInfo->szSigInfoRealDigest, pSigInfo->nSigInfoRealDigestLen);
      l1 = pSigInfo->nSigInfoRealDigestLen;
    }
    encode((const byte*)buf1, l1, (byte*)pBuf, pBufLen);
  } else {
    if(bWithAsn1Prefix) {
      memcpy(pBuf, g_sigPrefix, sizeof(g_sigPrefix));
      memcpy(pBuf + sizeof(g_sigPrefix), pSigInfo->szSigInfoRealDigest, 
	     pSigInfo->nSigInfoRealDigestLen);
      *pBufLen = pSigInfo->nSigInfoRealDigestLen  + sizeof(g_sigPrefix);
    }
    else {
      *pBufLen = pSigInfo->nSigInfoRealDigestLen;
      memcpy(pBuf, pSigInfo->szSigInfoRealDigest, pSigInfo->nSigInfoRealDigestLen);
    }
  }
  return err;
}


//================< deprecated functions> =================================

#ifdef WITH_DEPRECATED_FUNCTIONS

// get signers id-code
EXP_OPTION int getSignerCode(const SignatureInfo* pSigInfo, char* buf)
{
  int err, l1;
  X509* cert;
  char buf1[500], *p;
  
  RETURN_IF_NULL_PARAM(pSigInfo);
  RETURN_IF_NULL_PARAM(buf);
  cert = getSignCertData(pSigInfo);
  if (!cert) SET_LAST_ERROR_RETURN_CODE(ERR_CERT_INVALID);
  l1 = sizeof(buf1);
  err = getCertSubjectName(cert, buf1, &l1);
  if (err != ERR_OK) SET_LAST_ERROR_RETURN_CODE(err);
  err = ERR_CERT_READ; 
  p = strstr(buf1, "CN=");
  if (p) {
    p = strchr(p, ',');
    if(p) {
      p = strchr(p+1, ',');
      if(p) {
	strncpy(buf, p+1, 11);
	buf[11] = 0;
	err = ERR_OK;
      } 
    } else { // no comma -> no id-code !
      buf[0] = 0; 
      // is this really an error ?
      err = ERR_WRONG_CERT;
    }  
  }
  if (err != ERR_OK) SET_LAST_ERROR(err);
  return err;
}

// get signers first name
EXP_OPTION int getSignerFirstName(const SignatureInfo* pSigInfo, char* buf)
{
  int err = ERR_OK, l1;
  X509* cert;
  char buf1[500], *p, *p2, *p1;

  RETURN_IF_NULL_PARAM(pSigInfo);
  RETURN_IF_NULL_PARAM(buf);
  cert = getSignCertData(pSigInfo);
  if (!cert) SET_LAST_ERROR_RETURN_CODE(ERR_CERT_INVALID);
  l1 = sizeof(buf1);
  err = getCertSubjectName(cert, buf1, &l1);
  if (err != ERR_OK) SET_LAST_ERROR_RETURN_CODE(err);
  p = strstr(buf1, "CN=");
  if (p) {
    p1 = strchr(p, ',');
    if(!p1)
      p1 = strchr(p, '/');
    if (p1) {
      p1 += 1;
      p2 = strchr(p1, ',');
      if(!p2)
	p2 = strchr(p1, '/');
      if(p2) {
	strncpy(buf, p1, p2-p1);
	buf[p2-p1] = 0;
	err = ERR_OK;
      }	
    }
  }
  if (err != ERR_OK) SET_LAST_ERROR(err);
  return err;
}

// get signers last name
EXP_OPTION int getSignerLastName(const SignatureInfo* pSigInfo, char* buf)
{
  int err = ERR_OK, l1;
  X509* cert;
  char buf1[500], *p, *p2;
  
  RETURN_IF_NULL_PARAM(pSigInfo);
  RETURN_IF_NULL_PARAM(buf);
  cert = getSignCertData(pSigInfo);
  if (!cert) SET_LAST_ERROR_RETURN_CODE(ERR_CERT_INVALID);
  l1 = sizeof(buf1);
  err = getCertSubjectName(cert, buf1, &l1);
  if (err != ERR_OK) SET_LAST_ERROR_RETURN_CODE(err);
  p = strstr(buf1, "CN=");
  if(p) {
    p += 3;
    p2 = strchr(p, ',');
    if(!p2)
      p2 = strchr(p, '/');
    if(p2) {
      strncpy(buf, p, p2-p);
      buf[p2-p] = 0;
      err = ERR_OK;
    } else {
      strcpy(buf, p);
    }
  }
  if (err != ERR_OK) SET_LAST_ERROR(err);
  return err;
}

EXP_OPTION int getSignerCN(const SignatureInfo* pSigInfo, char* buf, int bUTF8)
{
  int err = ERR_OK, l1;
  X509* cert;
  char buf1[500], *p, *p2, buf2[500];
  
  RETURN_IF_NULL_PARAM(pSigInfo);
  RETURN_IF_NULL_PARAM(buf);
  cert = getSignCertData(pSigInfo);
  if (!cert) SET_LAST_ERROR_RETURN_CODE(ERR_CERT_INVALID);
  l1 = sizeof(buf1);
  err = getCertSubjectName(cert, buf1, &l1);
  if (err != ERR_OK) SET_LAST_ERROR_RETURN_CODE(err);
  if(!bUTF8) {
    l1 = sizeof(buf2);
    utf82ascii(buf1, buf2, &l1);
    strcpy(buf1, buf2);
  }
  err = ERR_CERT_READ;
  p = strstr(buf1, "CN=");
  if (p) {
    p += 3;
    p2 = strchr(p, '/');
    if(!p2)
      p2 = p + strlen(p);
    if (p2) {
      strncpy(buf, p, p2-p);
      buf[p2-p] = 0;
      err = ERR_OK;	
    } 
  }
  
  if (err != ERR_OK) SET_LAST_ERROR(err);	
  return err;
}

#endif  // WITH_DEPRECATED_FUNCTIONS


